/* ----------------------------------------------------------------------- */
/*    Fernando Berzal Galiano                ETS Ingeniera Informtica    */
/* ----------------------------------------------------------------------- */
/*                       Procesamiento de imgenes                         */
/* ----------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <math.h>

#include "IMG.H"

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*                              TDA Imagen                                 */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

Imagen CreaImagen ( WORD anchura, WORD altura )
{
  WORD   linea ;
  Imagen img   = (Imagen) malloc (sizeof(struct InfoImagen));

  if (img) {

     img->Anchura = anchura;
     img->Altura  = altura;
     img->Paleta  = NULL;

     img->Lineas  = (BYTE**) calloc (altura, sizeof(BYTE*) );

     if (img->Lineas) {

	for (linea=0; linea<altura; linea++) {
	    img->Lineas[linea] = (BYTE*)calloc(anchura, sizeof(BYTE));
	    memset ( img->Lineas[linea], 0, anchura );
	}

     } else {

	free (img);
	img = NULL;
     }

  }

  return img;
}

/* ----------------------------------------------------------------------- */
/*  Crea una paleta especial para la imagen                                */

int CreaPaletaImagen (Imagen img)
{
  img->Paleta = (RGB *) calloc ( 256, sizeof(RGB));

  return (img->Paleta!=NULL);
}

/* ----------------------------------------------------------------------- */
/*  Copia la paleta de una imagen                                          */

void CopiaPaleta (Imagen destino, Imagen fuente)
{
  if (fuente->Paleta) {

     if (destino->Paleta==NULL)
	CreaPaletaImagen(destino);

     memcpy (destino->Paleta, fuente->Paleta, 256*sizeof(RGB) );

  } else if (destino->Paleta) {

     free (destino->Paleta);
     destino->Paleta = NULL;

  }
}

/* ----------------------------------------------------------------------- */

void DestruyeImagen (Imagen *img)
{
  WORD linea;

  if (*img) {

     if ((*img)->Paleta)
	free ((*img)->Paleta);

     for (linea=0; linea<ALTURA(*img); linea++)
	 free((*img)->Lineas[linea]);

     free ((*img)->Lineas);
     free (*img);
     *img = NULL;
  }
}

/* ----------------------------------------------------------------------- */

BYTE GetPixel (Imagen img, WORD x, WORD y)
{
  return PIXEL(img,x,y);
}

void SetPixel (Imagen img, WORD x, WORD y, BYTE c)
{
  PIXEL(img,x,y) = c;
}

WORD Anchura (Imagen img)
{
  return ANCHURA(img);
}

WORD Altura (Imagen img)
{
  return ALTURA(img);
}

/* ----------------------------------------------------------------------- */
/*  Devuelve la paleta de la imagen                                        */
/* ----------------------------------------------------------------------- */

RGB *PaletaImagen (Imagen img)
{
  return img->Paleta;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*  Formato de imagen FBGVG                                                */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

#define FBG_HEADER 0x46424700 /* "FBG" */

typedef struct FBGHeader {
			   QUAD Identificador;
			   WORD Ancho;
			   WORD Alto;
			   BYTE Formato;
			 } FBGHeader;

#define FBG_STANDARD  0                            /* Cada pixel en 1 byte */
#define FBG_GRAYLEVEL 1                            /* Cada pixel en 6 bits */
#define FBG_BW        2                            /* Cada pixel en 1 bit  */

/* ----------------------------------------------------------------------- */

void GrayToStandard ( WORD longitud, BYTE *linea )
{
  int   x;
  BYTE *ptr = linea + longitud - longitud/4 - 1;

  for (x=longitud-1; x>=0; x--) {

      switch (x%4) {

	case 0: linea[x] = (*ptr)&0xFC;
		ptr--;
		break;

	case 1: linea[x] = ((*ptr)&0xF0)>>2;
		ptr--;
		linea[x] |= ((*ptr)&0x03)<<6;
		break;

	case 2: linea[x] = ((*ptr)&0xC0)>>4;
		ptr--;
		linea[x] |= ((*ptr)&0x0F)<<4;
		break;

	case 3: linea[x] = ((*ptr)&0x3F)<<2;
		break;
      }
  }
}

/* --- */

void StandardToGray ( WORD longitud, BYTE *linea )
{
  WORD  x;
  BYTE *ptr = linea;

  for (x=0; x<longitud; x++) {

      switch (x%4) {

	case 0: *ptr  = (linea[x]&0xFC);
		break;

	case 1: *ptr |= (linea[x]&0xC0)>>6;
		ptr++;
		*ptr = (linea[x]&0x3C)<<2;
		break;

	case 2: *ptr |= (linea[x]&0xF0)>>4;
		ptr++;
		*ptr = (linea[x]&0x0C)<<4;
		break;

	case 3: *ptr |= (linea[x]&0xFC)>>2;
		ptr++;
		break;
      }
  }
}

/* ----------------------------------------------------------------------- */

void BWToStandard ( WORD longitud, BYTE *linea )
{
  int   x;
  BYTE *ptr = linea + (longitud+7)/8 - 1;
  BYTE  val;

  for (x=longitud-1; x>=0; x--) {

      val = 0x01 << (x%8);
      linea[x] = (*ptr & val) ? 0xFF: 0x00;

      if (x%8==0)
	 ptr --;
  }
}

/* --- */

void StandardToBW ( WORD longitud, BYTE *linea )
{
  WORD  x;
  BYTE *ptr = linea - 1;
  BYTE  val;

  for (x=0; x<longitud; x++) {

      val = (linea[x]!=0) ? 0x01: 0x00;

      if (x%8==0) {
	 ptr++;
	 *ptr = val;
      } else {
	 val <<= (x%8);
	 *ptr |= val;
      }
  }
}

/* ----------------------------------------------------------------------- */
/*  Cargar Imagen                                                          */
/* ----------------------------------------------------------------------- */

Imagen CargaImagen (char *nombre)
{
  FILE     *fich;
  Imagen    img;
  WORD      linea;
  FBGHeader hdr;

  fich = fopen (nombre, "rb");

  if (fich) {

     fread (&hdr, 9, 1, fich);

     if (hdr.Identificador==FBG_HEADER) {

	img = CreaImagen (hdr.Ancho, hdr.Alto);

	if (img) {

	   if ( hdr.Formato==FBG_STANDARD ) {

	      for (linea=0; linea<hdr.Alto; linea++)
		  fread ( LINEA(img,linea), sizeof(BYTE), hdr.Ancho, fich );

	   } else if ( hdr.Formato==FBG_GRAYLEVEL ) {

	      for (linea=0; linea<hdr.Alto; linea++) {
		  fread ( LINEA(img,linea), sizeof(BYTE), hdr.Ancho-hdr.Ancho/4, fich );
		  GrayToStandard (hdr.Ancho, LINEA(img,linea));
	      }

	   } else if ( hdr.Formato==FBG_BW ) {

	      for (linea=0; linea<hdr.Alto; linea++) {
		  fread ( LINEA(img,linea), sizeof(BYTE), (hdr.Ancho+7)/8, fich );
		  BWToStandard (hdr.Ancho, LINEA(img,linea));
	      }
	   }
	}
     }

     fclose (fich);
  }

  return img;
}

/* ----------------------------------------------------------------------- */
/*  Guardar Imagen                                                         */
/* ----------------------------------------------------------------------- */

void GuardaImagen ( Imagen img, char *nombre )
{
  FILE     *fich;
  WORD      linea;
  FBGHeader hdr;

  hdr.Identificador = FBG_HEADER;
  hdr.Ancho         = ANCHURA(img);
  hdr.Alto          = ALTURA(img);
  hdr.Formato       = FBG_STANDARD;

  fich = fopen (nombre, "wb");

  if (fich) {

     fwrite (&hdr, 9, 1, fich);

     for (linea=0; linea<ALTURA(img); linea++)
	 fwrite ( LINEA(img,linea), sizeof(BYTE), ANCHURA(img), fich );

     fclose (fich);
  }
}

/* ----------------------------------------------------------------------- */

void GuardaImagenGRAYLEVEL ( Imagen img, char *nombre )
{
  FILE     *fich;
  WORD      linea;
  FBGHeader hdr;

  hdr.Identificador = FBG_HEADER;
  hdr.Ancho         = ANCHURA(img);
  hdr.Alto          = ALTURA(img);
  hdr.Formato       = FBG_GRAYLEVEL;

  fich = fopen (nombre, "wb");

  if (fich) {

     fwrite (&hdr, 9, 1, fich);

     for (linea=0; linea<ALTURA(img); linea++) {
	 StandardToGray ( ANCHURA(img), LINEA(img,linea) );
	 fwrite ( LINEA(img,linea), sizeof(BYTE), ANCHURA(img)-ANCHURA(img)/4, fich );
	 GrayToStandard ( ANCHURA(img), LINEA(img,linea) );
     }

     fclose (fich);
  }
}

/* ----------------------------------------------------------------------- */

void GuardaImagenBW ( Imagen img, char *nombre )
{
  FILE     *fich;
  WORD      linea;
  FBGHeader hdr;

  hdr.Identificador = FBG_HEADER;
  hdr.Ancho         = ANCHURA(img);
  hdr.Alto          = ALTURA(img);
  hdr.Formato       = FBG_BW;

  fich = fopen (nombre, "wb");

  if (fich) {

     fwrite (&hdr, 9, 1, fich);

     for (linea=0; linea<ALTURA(img); linea++) {
	 StandardToBW ( ANCHURA(img), LINEA(img,linea) );
	 fwrite ( LINEA(img,linea), sizeof(BYTE), (ANCHURA(img)+7)/8, fich );
	 BWToStandard ( ANCHURA(img), LINEA(img,linea) );
     }

     fclose (fich);
  }
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*  Ficheros BMP (256 colores)                                             */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

#define CAB_BMP 0x4D42  /* Cabecera del fichero: BM */

#define BMP_RGB  0
#define BMP_RLE8 1
#define BMP_RLE4 2

typedef struct BMPFileHeader {
			      WORD Tipo;
			      long Size;
			      WORD Reserved1;
			      WORD Reserved2;
			      long Offbits;
			     } BMPFileHeader;

typedef struct BMPInfoHeader {
			      long Size;
			      long Anchura;
			      long Altura;
			      WORD Planos;
			      WORD BitsPixel;
			      long Compresion;
			      long SizeImage;
			      long X_PixelsMetro;
			      long Y_PixelsMetro;
			      long ColoresUsados;
			      long ColoresImportantes;
			     } BMPInfoHeader;

typedef struct RGBQuad {
			BYTE Azul;
			BYTE Verde;
			BYTE Rojo;
			BYTE Reservado;
		       } RGBQuad;

typedef RGBQuad RGBQpaleta[256];

/* --- */

Imagen CargaBMP (char *nombre)
{
  Imagen  img;
  FILE   *fichero;

  BMPFileHeader BMPFile;
  BMPInfoHeader BMPInfo;
  RGBQpaleta    BMPPaleta;
  BYTE         *linea;

  int           scanline, i,x;
  int           color;

  if ((fichero=fopen(nombre,"rb"))!=NULL) {

     fread(&BMPFile, 14, 1, fichero);
     fread(&BMPInfo, 40, 1, fichero);

     img = CreaImagen (BMPInfo.Anchura,BMPInfo.Altura);

     if ( img && (BMPFile.Tipo==CAB_BMP) &&
	  (BMPInfo.BitsPixel==8) && (BMPInfo.Compresion==0) ) {

        /* Paleta */

	fread (BMPPaleta, sizeof(BMPPaleta), 1, fichero);

	color = 0;

	for (i=0; i<256 && !color; i++)
	    if (  BMPPaleta[i].Rojo!=i
	       || BMPPaleta[i].Verde!=i
	       || BMPPaleta[i].Azul!=i )
               color = 1;

        if (color)
           if (CreaPaletaImagen(img))
	      for (i=0; i<256; i++) {
	          PALETA(img, i).R = BMPPaleta[i].Rojo;
	          PALETA(img, i).G = BMPPaleta[i].Verde;
	          PALETA(img, i).B = BMPPaleta[i].Azul;
	      }

        /* Pixels */

	scanline = BMPInfo.Anchura + (4-(BMPInfo.Anchura%4))%4;
	linea    = (BYTE *)malloc (scanline);

	for (i=BMPInfo.Altura-1; i>=0; i--) {
	    fread (linea, scanline, 1, fichero);

	    for (x=0; x<BMPInfo.Anchura; x++)
		PIXEL(img,x,i) = linea[x];
	}

	free (linea);
     }

     fclose (fichero);
  }

  return img;
}

/* ----------------------------------------------------------------------- */

void GuardaBMP (Imagen img, char *nombre)
{
  int     x,y;
  int     scanline;
  int     nivel;
  long    size;
  BYTE   *linea;
  FILE   *fichero;

  BMPFileHeader BMPFile;
  BMPInfoHeader BMPInfo;
  RGBQpaleta    BMPPaleta;

  if ((fichero=fopen(nombre,"wb"))!=NULL) {

     size = ANCHURA(img);

     while (size%4)
	   size++;

     size *= ALTURA(img);

     /* File header */

     BMPFile.Tipo      = CAB_BMP;
     BMPFile.Size      = 1078 + size;
     BMPFile.Reserved1 = 0;
     BMPFile.Reserved2 = 0;
     BMPFile.Offbits   = 1078;

     fwrite(&BMPFile, 14, 1, fichero);

     /* Info header */

     BMPInfo.Size               = 40;
     BMPInfo.Anchura            = ANCHURA(img);
     BMPInfo.Altura             = ALTURA(img);
     BMPInfo.Planos             = 1;
     BMPInfo.BitsPixel          = 8;
     BMPInfo.Compresion         = 0;
     BMPInfo.SizeImage          = size;
     BMPInfo.X_PixelsMetro      = 0;
     BMPInfo.Y_PixelsMetro      = 0;
     BMPInfo.ColoresUsados      = 0;
     BMPInfo.ColoresImportantes = 0;

     fwrite (&BMPInfo, 40, 1, fichero);

     /* Paleta */

     if (img->Paleta) {      /* Paleta especfica  */

	for (x=0; x<256; x++) {
	    BMPPaleta[x].Rojo      = PALETA(img,x).R;
	    BMPPaleta[x].Verde     = PALETA(img,x).G;
	    BMPPaleta[x].Azul      = PALETA(img,x).B;
	    BMPPaleta[x].Reservado = 0;
	}

     } else {                /* Paleta por defecto: Grises */

	for (nivel=0; nivel<256; nivel++) {
	    BMPPaleta[nivel].Rojo      = nivel;
	    BMPPaleta[nivel].Verde     = nivel;
	    BMPPaleta[nivel].Azul      = nivel;
	    BMPPaleta[nivel].Reservado = 0;
	}

     }

     fwrite (BMPPaleta, sizeof(BMPPaleta), 1, fichero);

     /* Imagen */

     scanline = BMPInfo.Anchura + (4-(BMPInfo.Anchura%4))%4;
     linea    = (BYTE*)malloc (scanline);

     for (x=0; x<scanline; x++)
	 linea[x] = 0;

     for (y=BMPInfo.Altura-1; y>=0; y--) {

	 for (x=0; x<BMPInfo.Anchura; x++)
	     linea[x] = PIXEL(img,x,y);

	 fwrite (linea, scanline, 1, fichero);
     }

     free (linea);

     fclose (fichero);
  }

}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*                        Tratamiento de imgenes                          */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

/* ----------------------------------------------------------------------- */
/*  Convolucin 2x2 y 3x3 (GRAYLEVEL)                                      */
/* ----------------------------------------------------------------------- */

Imagen Convolution (Imagen img, int t[3][3], BYTE div)
{
  WORD anchura = ANCHURA(img) - 2;
  WORD altura  = ALTURA(img)  - 2;
  WORD x,y;
  int  valor;

  Imagen conv = CreaImagen (anchura,altura);

  if (conv)
     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++) {
	     valor = t[0][0]*PIXEL(img,  x,  y) +
		     t[0][1]*PIXEL(img,x+1,  y) +
		     t[0][2]*PIXEL(img,x+2,  y) +
		     t[1][0]*PIXEL(img,  x,y+1) +
		     t[1][1]*PIXEL(img,x+1,y+1) +
		     t[1][2]*PIXEL(img,x+2,y+1) +
		     t[2][0]*PIXEL(img,  x,y+2) +
		     t[2][1]*PIXEL(img,x+1,y+2) +
		     t[2][2]*PIXEL(img,x+2,y+2) ;

	     valor /= div;

	     if (valor>0xFF)
		valor = 0xFF;
	     else if (valor<0)
		valor = 0;

	     PIXEL(conv,x,y) = valor;
	 }

  return conv;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

Imagen Convolution2 (Imagen img, int t[2][2], BYTE div)
{
  WORD anchura = ANCHURA(img) - 1;
  WORD altura  = ALTURA(img)  - 1;
  WORD x,y;
  int  valor;

  Imagen conv = CreaImagen (anchura,altura);

  if (conv)
     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++) {
	     valor = t[0][0]*PIXEL(img,  x,  y) +
		     t[0][1]*PIXEL(img,x+1,  y) +
		     t[1][0]*PIXEL(img,  x,y+1) +
		     t[1][1]*PIXEL(img,x+1,y+1) ;

	     valor /= div;

	     if (valor>0xFF)
		valor = 0xFF;
	     else if (valor<0)
		valor = 0;

	     PIXEL(conv,x,y) = valor;
	 }

  return conv;
}

/* ----------------------------------------------------------------------- */
/*  Moravec Operator                                                       */
/* ----------------------------------------------------------------------- */

#define MIN(a,b) (((a)<(b))?(a):(b))

Imagen Moravec ( Imagen origen, BYTE a )
{
  WORD anchura = ANCHURA(origen);
  WORD altura  = ALTURA(origen);
  WORD x,y;
  WORD Xfin, Yfin;
  BYTE min, var;

  Imagen aux1    = CreaImagen ( anchura - 2*a    , altura - 2*a     );
  Imagen aux2    = CreaImagen ( anchura - 2*a - 2, altura - 2*a - 2 );
  Imagen destino = CreaImagen ( anchura - 2*a - 4, altura - 2*a - 4 );

  if (!aux1 || !aux2 || !destino) {
     if (aux1)    DestruyeImagen (&aux1);
     if (aux2)    DestruyeImagen (&aux2);
     if (destino) DestruyeImagen (&destino);
     return destino;
  }

  /* --- */

  Xfin = anchura - 2*a;
  Yfin = altura  - 2*a;

  for (x=0; x<Xfin; x++)
      for (y=0; y<Yfin; y++) {

	  min  = PIXEL(origen,x+a,y+a);
	  var  = abs ((int)min-(int)PIXEL(origen,x,y+a));
	  var += abs ((int)min-(int)PIXEL(origen,x+a,y));
	  var += abs ((int)min-(int)PIXEL(origen,x+2*a,y+a));
	  var += abs ((int)min-(int)PIXEL(origen,x+a,y+2*a));

	  PIXEL (aux1,x,y) = var;
      }

  /* --- */

  Xfin = anchura - 2*a - 2;
  Yfin = altura  - 2*a - 2;

  for (x=0; x<Xfin; x++)
      for (y=0; y<Yfin; y++) {

	  min =            PIXEL(aux1,  x,  y)  ;
	  min = MIN ( min, PIXEL(aux1,x+1,  y) );
	  min = MIN ( min, PIXEL(aux1,x+2,  y) );
	  min = MIN ( min, PIXEL(aux1,  x,y+1) );
	  min = MIN ( min, PIXEL(aux1,x+1,y+1) );
	  min = MIN ( min, PIXEL(aux1,x+2,y+1) );
	  min = MIN ( min, PIXEL(aux1,  x,y+2) );
	  min = MIN ( min, PIXEL(aux1,x+1,y+2) );
	  min = MIN ( min, PIXEL(aux1,x+2,y+2) );

	  PIXEL(aux2,x,y)=min;
      }

  /* --- */

  Xfin = anchura - 2*a - 4;
  Yfin = altura  - 2*a - 4;

  for (x=0; x<Xfin; x++)
      for (y=0; y<Yfin; y++) {

	  min = PIXEL (aux2,x+1,y+1);

	  if (  min < PIXEL(aux2,  x,  y)
	     || min < PIXEL(aux2,x+1,  y)
	     || min < PIXEL(aux2,x+2,  y)
	     || min < PIXEL(aux2,  x,y+1)
	     || min < PIXEL(aux2,x+2,y+2)
	     || min < PIXEL(aux2,  x,y+2)
	     || min < PIXEL(aux2,x+1,y+2)
	     || min < PIXEL(aux2,x+2,y+2) )
	     PIXEL (destino,x,y) = 0x00;
	  else
	     PIXEL(destino,x,y) = PIXEL(origen,x+a+2,y+a+2);
      }

  DestruyeImagen(&aux1);
  DestruyeImagen(&aux2);

  return destino;
}
