//---------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "Main.h"

#include "ABOUT.H"
#include "CONTORNO.H"
#include "GAUSS.H"
#include "MAX.H"

//---------------------------------------------------------------------
#pragma resource "*.dfm"
TMainForm *MainForm;
//---------------------------------------------------------------------
__fastcall TMainForm::TMainForm(TComponent *Owner)
	: TForm(Owner)
{
}
//---------------------------------------------------------------------
void __fastcall TMainForm::FormCreate(TObject *Sender)
{
  Application->OnHint = ShowHint;
  Screen->OnActiveFormChange = UpdateMenuItems;

  NumContornos = 0;
}
//---------------------------------------------------------------------
void __fastcall TMainForm::ShowHint(TObject *Sender)
{
  StatusBar->SimpleText = Application->Hint;
}
//---------------------------------------------------------------------
void __fastcall TMainForm::CreateMDIChild(Imagen img)
{
  char   s[16];

  TMDIChild *Child;

  //--- create a new MDI child window ----

  Child = new TMDIChild(Application);

  NumContornos++;
  sprintf(s,"Contorno %d",NumContornos);
  Child->Caption = s;

  Child->ImagenContorno = img;
  Child->ActualizarVisor();
  Child->Extraer();
}
//---------------------------------------------------------------------
void __fastcall TMainForm::FileOpenItemClick(TObject *Sender)
{
  Imagen img;
  int    i;
  char   fichero[256];
  char   extension[4];

  if (OpenDialog->Execute()) {

      strcpy(fichero,OpenDialog->FileName.c_str());

      // Extensin del fichero

      for (i=0; i<3; i++)
          extension[i] = (char)toupper(fichero[strlen(fichero)-3+i]);

      extension[3] = '\0';

     // Carga la imagen

     if ( !strcmp(extension,"GRZ") ) {

        img = CargaImagen(fichero);

        if (img) {
           CreateMDIChild(img);
        } else {
           Application->MessageBox ("Imagen de formato incorrecto", "Error", MB_OK);
        }

     } else {
        Application->MessageBox ("Extensin incorrecta", "Error", MB_OK);
     }
  }

  UpdateMenuItems(Sender);
}
//---------------------------------------------------------------------
void __fastcall TMainForm::FileCloseItemClick(TObject *Sender)
{
  if (ActiveMDIChild)
     ActiveMDIChild->Close();

  UpdateMenuItems(Sender);
}
//---------------------------------------------------------------------
void __fastcall TMainForm::FileSaveItemClick(TObject *Sender)
{
  // Save current file (ActiveMDIChild points to the window)

  TMDIChild *Child = (TMDIChild *) ActiveMDIChild;

  int  i;
  char fichero[256];
  char extension[4];

  SaveDialog->FileName = "";

  if (SaveDialog->Execute()) {

     strcpy(fichero,SaveDialog->FileName.c_str());

     // Extensin del fichero

     for (i=0; i<3; i++)
         extension[i] = (char)toupper(fichero[strlen(fichero)-3+i]);

     extension[3] = '\0';

     // Guarda la imagen

     if (!strcmp(extension,"GRZ")) {

        GuardaImagenBW(Child->ImagenContorno,fichero);

     } else {

        Application->MessageBox ("Extensin no permitida", "Error", IDOK);

     }
  }
}
//---------------------------------------------------------------------
void __fastcall TMainForm::FileExitItemClick(TObject *Sender)
{
  Close();
}
//---------------------------------------------------------------------
void __fastcall TMainForm::WindowCascadeItemClick(TObject *Sender)
{
  Cascade();
}
//---------------------------------------------------------------------
void __fastcall TMainForm::WindowTileItemClick(TObject *Sender)
{
  Tile();
}
//---------------------------------------------------------------------
void __fastcall TMainForm::WindowArrangeItemClick(TObject *Sender)
{
  ArrangeIcons();
}
//---------------------------------------------------------------------
void __fastcall TMainForm::WindowMinimizeItemClick(TObject *Sender)
{
  int i;

  //---- Must be done backwards through the MDIChildren array ----

  for (i=MDIChildCount-1; i >= 0; i--)
      MDIChildren[i]->WindowState = wsMinimized;
}
//---------------------------------------------------------------------
void __fastcall TMainForm::UpdateMenuItems(TObject *Sender)
{
  FileSaveItem->Enabled = MDIChildCount > 0;
  SaveBtn->Enabled = MDIChildCount > 0;
  WindowCascadeItem->Enabled = MDIChildCount > 0;
  WindowTileItem->Enabled = MDIChildCount > 0;
  WindowArrangeItem->Enabled = MDIChildCount > 0;
  WindowMinimizeItem->Enabled = MDIChildCount > 0;

  MenuContorno->Visible = MDIChildCount > 0;
}
//---------------------------------------------------------------------
void __fastcall TMainForm::FormDestroy(TObject *Sender)
{
  Screen->OnActiveFormChange = NULL;
}
//---------------------------------------------------------------------
void __fastcall TMainForm::HelpAboutItemClick(TObject *Sender)
{
  Hide();
  FormAbout->ShowModal();
  Show();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SpeedButtonInfoClick(TObject *Sender)
{
  Hide();
  FormAbout->ShowModal();
  Show();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::MenuSuavizadoClick(TObject *Sender)
{
  TMDIChild *Child = (TMDIChild *) ActiveMDIChild;

  if (FormGauss->ShowModal()==mrOk) {
     Screen->Cursor = crHourGlass;
     Child->Suavizar(FormGauss->S, FormGauss->D);
     Screen->Cursor = crDefault;
  }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::MenuPoligonalClick(TObject *Sender)
{
  TMDIChild *Child = (TMDIChild *) ActiveMDIChild;
  Imagen     tmp;

  if (FormGauss->ShowModal()==mrOk) {
     Screen->Cursor = crHourGlass;

     tmp = AproximacionPoligonal ( ANCHURA(Child->ImagenContorno),
                                   ALTURA(Child->ImagenContorno),
                                   FormGauss->S,
                                   FormGauss->D,
                                   Child->Puntos, Child->X, Child->Y );

     CreateMDIChild (tmp);

     Screen->Cursor = crDefault;
  }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::MenuInfoClick(TObject *Sender)
{
  TMDIChild *Child = (TMDIChild *) ActiveMDIChild;

  FormContorno->X = Child->X;
  FormContorno->Y = Child->Y;
  FormContorno->N = Child->Puntos;

  FormContorno->ShowModal();

  // char s[40];
  // sprintf (s,"El contorno contiene %d puntos", Child->Puntos);
  // Application->MessageBox (s, "Informacin", MB_OK);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::MenuMaxClick(TObject *Sender)
{
  TMDIChild *Child = (TMDIChild *) ActiveMDIChild;

  FormMaxK->X = Child->X;
  FormMaxK->Y = Child->Y;
  FormMaxK->N = Child->Puntos;

  FormMaxK->ShowModal();
}
//---------------------------------------------------------------------------
