/* ----------------------------------------------------------------------- */
/*    Fernando Berzal Galiano                ETS Ingeniera Informtica    */
/* ----------------------------------------------------------------------- */
/*                       Procesamiento de imgenes                         */
/* ----------------------------------------------------------------------- */

#ifndef BYTE
#define BYTE unsigned char
#endif

#ifndef WORD
#define WORD unsigned short
#endif

#ifndef QUAD
#define QUAD unsigned long
#endif

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*                              TDA Imagen                                 */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

#ifndef __FB_IMAGEN__
#define __FB_IMAGEN__

typedef struct RGB {
		     BYTE R;  /* Rojo:  0..255 */
		     BYTE G;  /* Verde: 0..255 */
                     BYTE B;  /* Azul:  0..255 */
                   } RGB;


typedef struct InfoImagen {
			   WORD   Anchura;
			   WORD   Altura;
			   BYTE **Lineas;
			   RGB   *Paleta;
		          } InfoImagen;

typedef struct InfoImagen *Imagen;

/* ----------------------------------------------------------------------- */

#define ANCHURA(img)   ((img)->Anchura)

#define ALTURA(img)    ((img)->Altura)

#define PIXEL(img,x,y) ((img)->Lineas[y][x])

#define LINEA(img,y)   ((img)->Lineas[y])

#define PALETA(img,c)  ((img)->Paleta[c])

/* ----------------------------------------------------------------------- */

Imagen CreaImagen ( WORD anchura, WORD altura );

void DestruyeImagen (Imagen *img);

/* ----------------------------------------------------------------------- */

WORD Anchura (Imagen img);

WORD Altura  (Imagen img);

BYTE GetPixel (Imagen img, WORD x, WORD y);

void SetPixel (Imagen img, WORD x, WORD y, BYTE c);

RGB *PaletaImagen (Imagen img);



void CopiaPaleta (Imagen destino, Imagen fuente);


/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*  Formato de imagen FBGVG                                                */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

/* ----------------------------------------------------------------------- */
/*  Cargar Imagen                                                          */
/* ----------------------------------------------------------------------- */

Imagen CargaImagen (char *nombre);

/* ----------------------------------------------------------------------- */
/*  Guardar Imagen                                                         */
/* ----------------------------------------------------------------------- */

void GuardaImagen ( Imagen img, char *nombre );

/* ----------------------------------------------------------------------- */

void GuardaImagenGRAYLEVEL ( Imagen img, char *nombre );

/* ----------------------------------------------------------------------- */

void GuardaImagenBW ( Imagen img, char *nombre );

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*  Ficheros BMP (256 colores)                                             */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

Imagen CargaBMP (char *nombre);

/* ----------------------------------------------------------------------- */

void GuardaBMP (Imagen img, char *nombre);

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*                        Tratamiento de imgenes                          */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

/* ----------------------------------------------------------------------- */
/*  Convolucin 2x2 y 3x3 (GRAYLEVEL)                                      */
/* ----------------------------------------------------------------------- */

Imagen Convolution (Imagen img, int t[3][3], BYTE div);

/* ----------------------------------------------------------------------- */

Imagen Convolution2 (Imagen img, int t[2][2], BYTE div);

/* ----------------------------------------------------------------------- */
/*  Moravec Operator                                                       */
/* ----------------------------------------------------------------------- */

Imagen Moravec ( Imagen origen, BYTE a );

/* ----------------------------------------------------------------------- */

#endif
