/* ----------------------------------------------------------------------- */
/*    Fernando Berzal Galiano                ETS Ingeniera Informtica    */
/* ----------------------------------------------------------------------- */
/*                       Procesamiento de imgenes                         */
/* ----------------------------------------------------------------------- */
/*                   Transformaciones geomtricas en 2D                    */
/* ----------------------------------------------------------------------- */

#include <stdlib.h>
#include <math.h>

#include "IMG_2D.H"

/* ----------------------------------------------------------------------- */

Imagen ZoomImagen ( Imagen img, WORD anchura, WORD altura )
{
  WORD   x,y;
  float  dx   = (float)ANCHURA(img)/(float)anchura;
  float  dy   = (float)ALTURA(img)/(float)altura;
  Imagen zoom = CreaImagen (anchura,altura);

  if (zoom)
     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(zoom,x,y) = PIXEL(img, (WORD)(x*dx), (WORD)(y*dy));

  CopiaPaleta(zoom,img);

  return zoom;
}

/* ----------------------------------------------------------------------- */

Imagen ZoomInterpolacion ( Imagen img, int orden ) // Imagen en niveles de gris
{
  WORD   i,x,y;
  WORD   factor      = orden+1;
  WORD   anchura     = factor*ANCHURA(img)+orden;
  WORD   altura      = factor*ALTURA(img)+orden;

  int    mask [3][3] = { { 1, 2, 1}, { 2, 4, 2}, { 1, 2, 1} };

  Imagen zoom = CreaImagen (anchura,altura);
  Imagen conv;


  if (zoom) {

     // Entrelazado de ceros
   
     for (x=orden; x<anchura; x+=factor)
	 for (y=orden; y<altura; y+=factor)
	     PIXEL(zoom,x,y) = PIXEL(img, (x-orden)/factor, (y-orden)/factor);

     // Convolucin

     for (i=0; i<orden; i++) {
	 conv = Convolution (zoom,mask,4);
	 DestruyeImagen(&zoom);
	 zoom = conv;
     }
  }

  return zoom;
}

/* ----------------------------------------------------------------------- */

Imagen SubImagen ( Imagen img, WORD x1, WORD y1, WORD x2, WORD y2 )
{
  WORD   x,y;
  Imagen sub = CreaImagen (x2-x1+1,y2-y1+1);

  if (sub)
     for (x=x1; x<=x2; x++)
	 for (y=y1; y<=y2; y++)
	     PIXEL(sub,x-x1,y-y1) = PIXEL(img,x,y);

  CopiaPaleta(sub,img);

  return sub;
}

/* ----------------------------------------------------------------------- */

Imagen ReflejoVertical ( Imagen img )
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  Imagen reflejo = CreaImagen(anchura,altura);
  WORD   x,y;

  if (reflejo)
     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(reflejo,x,y) = PIXEL(img,x,altura-y-1);

  CopiaPaleta(reflejo,img);

  return reflejo;
}

/* --- */

Imagen ReflejoHorizontal ( Imagen img )
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  Imagen reflejo = CreaImagen(anchura,altura);
  WORD   x,y;

  if (reflejo)
     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(reflejo,x,y) = PIXEL(img,anchura-x-1,y);

  CopiaPaleta(reflejo,img);

  return reflejo;
}

/* ----------------------------------------------------------------------- */

Imagen RotacionClock (Imagen img, double angulo )
{
  float  coseno  = cos(angulo);
  float  seno    = sin(angulo);
  WORD   width   = ANCHURA(img);
  WORD   height  = ALTURA(img);
  WORD   anchura = ALTURA(img)*seno + ANCHURA(img)*coseno;
  WORD   altura  = ANCHURA(img)*seno + ALTURA(img)*coseno;
  int    incr_x  = ALTURA(img)*seno;
  int    x,y;
  int    ximg,yimg;

  Imagen rot = CreaImagen (anchura, altura);

  if (rot)
     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++) {

	     ximg = (int)((x-incr_x)*coseno+y*seno);
	     yimg = (int)(y*coseno-(x-incr_x)*seno);

	     if (ximg>=0 && ximg<width && yimg>=0 && yimg<height)
		PIXEL(rot,x,y) = PIXEL(img, (WORD)ximg, (WORD)yimg);
	 }

  CopiaPaleta(rot,img);

  return rot;
}

/* --- */

Imagen Rotacion90 ( Imagen img )
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  Imagen reflejo = CreaImagen(altura,anchura);
  WORD   x,y;

  if (reflejo)
     for (x=0; x<altura; x++)
	 for (y=0; y<anchura; y++)
	     PIXEL(reflejo,x,y) = PIXEL(img,anchura-y-1,x);

  CopiaPaleta(reflejo,img);

  return reflejo;
}

/* --- */

Imagen Rotacion180 ( Imagen img )
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  Imagen reflejo = CreaImagen(anchura,altura);
  WORD   x,y;

  if (reflejo)
     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(reflejo,x,y) = PIXEL(img,anchura-x-1,altura-y-1);

  CopiaPaleta(reflejo,img);

  return reflejo;
}

/* --- */

Imagen Rotacion270 ( Imagen img )
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  Imagen reflejo = CreaImagen(altura,anchura);
  WORD   x,y;

  if (reflejo)
     for (x=0; x<altura; x++)
	 for (y=0; y<anchura; y++)
	     PIXEL(reflejo,x,y) = PIXEL(img,y,altura-x-1);

  CopiaPaleta(reflejo,img);

  return reflejo;
}

/* --- */

#define PI 3.14159265359

Imagen Rotacion ( Imagen img, double angulo )
{
  Imagen aux = NULL;
  Imagen rot = NULL;

  if (angulo>0 && angulo<2*PI) {
     if (angulo<PI/2) {
	aux = Rotacion90(img);
	rot = RotacionClock(aux,PI/2-angulo);
	DestruyeImagen(&aux);
     } else if (angulo==PI/2) {
	rot = Rotacion90(img);
     } else if (angulo<PI) {
	aux = Rotacion180(img);
	rot = RotacionClock(aux,PI-angulo);
	DestruyeImagen(&aux);
     } else if (angulo==PI) {
	rot = Rotacion180(img);
     } else if (angulo<3*PI/2) {
	aux = Rotacion270(img);
	rot = RotacionClock(aux,3*PI/2-angulo);
	DestruyeImagen(&aux);
     } else if (angulo==3*PI/2) {
	rot = Rotacion270(img);
     } else {
	rot = RotacionClock(img,2*PI-angulo);
     }
  }

  return rot;
}

/* ----------------------------------------------------------------------- */
