/* ----------------------------------------------------------------------- */
/*    Fernando Berzal Galiano                ETS Ingeniera Informtica    */
/* ----------------------------------------------------------------------- */
/*                       Procesamiento de imgenes                         */
/* ----------------------------------------------------------------------- */
/*                         Deteccin de aristas                            */
/* ----------------------------------------------------------------------- */

#include <stdlib.h>
#include <math.h>

#include "IMG_2D.H"
#include "IMG_SEGM.H"
#include "IMG_TYPE.H"

#include "IMG_Edge.H"

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*  Edge detection                                                         */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

#define MAX(a,b) (((a)>(b))?(a):(b))

/* ----------------------------------------------------------------------- */

Imagen MaskEdgeDetector2x2 ( Imagen img )
{
  WORD   anchura = (WORD)(ANCHURA(img)-1);
  WORD   altura  = (WORD)(ALTURA(img)-1);

  int    ConvolutionVertical   [2][2] = { {  1, -1 } , {  1, -1 } };
  int    ConvolutionHorizontal [2][2] = { {  1,  1 } , { -1, -1 } };

  Imagen imgx    = Convolution2 ( img, ConvolutionHorizontal, 1 );
  Imagen imgy    = Convolution2 ( img, ConvolutionVertical,   1 );
  Imagen edges   = CreaImagen (anchura, altura);

  WORD   x,y;

  if (imgx && imgy && edges)
     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = MAX(PIXEL(imgx,x,y),PIXEL(imgy,x,y));

  DestruyeImagen (&imgx);
  DestruyeImagen (&imgy);

  return edges;
}

/* ----------------------------------------------------------------------- */

Imagen MaskEdgeDetector3x3 ( Imagen img )
{
  WORD   anchura = (WORD)(ANCHURA(img)-2);
  WORD   altura  = (WORD)(ALTURA(img)-2);

  int maskV[3][3] = { {-1, 0, 1}, {-1, 0, 1}, {-1, 0, 1} };
  int maskH[3][3] = { { 1, 1, 1}, { 0, 0, 0}, {-1,-1,-1} };

  Imagen imgx  = Convolution ( img, maskH, 1 );
  Imagen imgy  = Convolution ( img, maskV, 1 );
  Imagen edges = CreaImagen (anchura, altura);

  WORD   x,y;

  if (imgx && imgy && edges)
     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = MAX(PIXEL(imgx,x,y),PIXEL(imgy,x,y));

  DestruyeImagen (&imgx);
  DestruyeImagen (&imgy);

  return edges;
}

/* ----------------------------------------------------------------------- */

Imagen PhillipsMaskEdgeDetector ( Imagen img )
{
  int  mask[3][3] = { {-1, 0,-1}, { 0, 4, 0}, {-1, 0,-1} };

  return Convolution (img, mask, 1);
}

Imagen FalerMaskEdgeDetector ( Imagen img )
{
  int  mask[3][3] = { { 0, 1, 0}, {-1, 0, 1}, { 0,-1, 0} };

  return Convolution (img, mask, 1);
}


/* Filtros laplacianos */

Imagen Laplacian1EdgeDetector ( Imagen img )
{
  int  mask[3][3] = { { 0,-1, 0}, {-1, 4,-1}, { 0,-1, 0} };

  return Convolution (img, mask, 1);
}

Imagen Laplacian2EdgeDetector ( Imagen img )
{
  int  mask[3][3] = { {-1,-1,-1}, {-1, 8,-1}, {-1,-1,-1} };

  return Convolution (img, mask, 1);
}

Imagen Laplacian3EdgeDetector ( Imagen img )
{
  int  mask[3][3] = { { 1,-2, 1}, {-2, 4,-2}, { 1,-2, 1} };

  return Convolution (img, mask, 1);
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*  Sobel                                                                  */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

Imagen SobelEdgeDetector ( Imagen img )
{
  WORD   anchura = (WORD)(ANCHURA(img)-2);
  WORD   altura  = (WORD)(ALTURA(img)-2);

  int K0[3][3] = { { 1, 2, 1}, { 0, 0, 0}, {-1,-2,-1} };
  int K1[3][3] = { { 2, 1, 0}, { 1, 0,-1}, { 0,-1,-2} };
  int K2[3][3] = { { 1, 0,-1}, { 2, 0,-2}, { 1, 0,-1} };
  int K3[3][3] = { { 0,-1,-2}, { 1, 0,-1}, { 2, 1, 0} };
  int K4[3][3] = { {-1,-2,-1}, { 0, 0, 0}, { 1, 2, 1} };
  int K5[3][3] = { {-2,-1, 0}, {-1, 0, 1}, { 0, 1, 2} };
  int K6[3][3] = { {-1, 0, 1}, {-2, 0, 2}, {-1, 0, 1} };
  int K7[3][3] = { { 0, 1, 2}, {-1, 0, 1}, {-2,-1, 0} };

  Imagen aux   ;
  Imagen edges = CreaImagen (anchura, altura);

  WORD   x,y;

  if (edges) {

     aux = Convolution (img, K0, 1);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = PIXEL(aux,x,y);

     DestruyeImagen (&aux);

     aux = Convolution (img, K1, 1);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = MAX(PIXEL(aux,x,y),PIXEL(edges,x,y));

     DestruyeImagen (&aux);

     aux = Convolution (img, K2, 1);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = MAX(PIXEL(aux,x,y),PIXEL(edges,x,y));

     DestruyeImagen (&aux);

     aux = Convolution (img, K3, 1);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = MAX(PIXEL(aux,x,y),PIXEL(edges,x,y));

     DestruyeImagen (&aux);

     aux = Convolution (img, K4, 1);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = MAX(PIXEL(aux,x,y),PIXEL(edges,x,y));

     DestruyeImagen (&aux);

     aux = Convolution (img, K5, 1);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = MAX(PIXEL(aux,x,y),PIXEL(edges,x,y));

     DestruyeImagen (&aux);

     aux = Convolution (img, K6, 1);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = MAX(PIXEL(aux,x,y),PIXEL(edges,x,y));

     DestruyeImagen (&aux);

     aux = Convolution (img, K7, 1);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = MAX(PIXEL(aux,x,y),PIXEL(edges,x,y));

     DestruyeImagen (&aux);
  }

  return edges;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*  Kirsch                                                                 */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

Imagen KirschEdgeDetector ( Imagen img )
{
  WORD   anchura = (WORD)(ANCHURA(img)-2);
  WORD   altura  = (WORD)(ALTURA(img)-2);

  int K0[3][3] = { {-3,-3, 5}, {-3, 0, 5}, {-3,-3, 5} };
  int K1[3][3] = { {-3, 5, 5}, {-3, 0, 5}, {-3,-3,-3} };
  int K2[3][3] = { { 5, 5, 5}, {-3, 0,-3}, {-3,-3,-3} };
  int K3[3][3] = { { 5, 5,-3}, { 5, 0,-3}, {-3,-3,-3} };
  int K4[3][3] = { { 5,-3,-3}, { 5, 0,-3}, { 5,-3,-3} };
  int K5[3][3] = { {-3,-3,-3}, { 5, 0,-3}, { 5, 5,-3} };
  int K6[3][3] = { {-3,-3,-3}, {-3, 0,-3}, { 5, 5, 5} };
  int K7[3][3] = { {-3,-3,-3}, {-3, 0, 5}, {-3, 5, 5} };

  Imagen aux   ;
  Imagen edges = CreaImagen (anchura, altura);

  WORD   x,y;

  if (edges) {

     aux = Convolution (img, K0, 1);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = PIXEL(aux,x,y);

     DestruyeImagen (&aux);

     aux = Convolution (img, K1, 1);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = MAX(PIXEL(aux,x,y),PIXEL(edges,x,y));

     DestruyeImagen (&aux);

     aux = Convolution (img, K2, 1);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = MAX(PIXEL(aux,x,y),PIXEL(edges,x,y));

     DestruyeImagen (&aux);

     aux = Convolution (img, K3, 1);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = MAX(PIXEL(aux,x,y),PIXEL(edges,x,y));

     DestruyeImagen (&aux);

     aux = Convolution (img, K4, 1);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = MAX(PIXEL(aux,x,y),PIXEL(edges,x,y));

     DestruyeImagen (&aux);

     aux = Convolution (img, K5, 1);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = MAX(PIXEL(aux,x,y),PIXEL(edges,x,y));

     DestruyeImagen (&aux);

     aux = Convolution (img, K6, 1);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = MAX(PIXEL(aux,x,y),PIXEL(edges,x,y));

     DestruyeImagen (&aux);

     aux = Convolution (img, K7, 1);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = MAX(PIXEL(aux,x,y),PIXEL(edges,x,y));

     DestruyeImagen (&aux);
  }

  return edges;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*  Prewitt                                                                */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

Imagen PrewittEdgeDetector ( Imagen img )
{
  WORD   anchura = (WORD)(ANCHURA(img)-2);
  WORD   altura  = (WORD)(ALTURA(img)-2);

  int K0[3][3] = { { 1, 1, 1}, { 1,-2, 1}, {-1,-1,-1} };
  int K1[3][3] = { { 1, 1, 1}, { 1,-2,-1}, { 1,-1,-1} };
  int K2[3][3] = { { 1, 1,-1}, { 1,-2,-1}, { 1, 1,-1} };
  int K3[3][3] = { { 1,-1,-1}, { 1,-2,-1}, { 1, 1, 1} };
  int K4[3][3] = { {-1,-1,-1}, { 1,-2, 1}, { 1, 1, 1} };
  int K5[3][3] = { {-1,-1, 1}, {-1,-2, 1}, { 1, 1, 1} };
  int K6[3][3] = { {-1, 1, 1}, {-1,-2, 1}, {-1, 1, 1} };
  int K7[3][3] = { { 1, 1, 1}, {-1,-2, 1}, {-1,-1, 1} };

  Imagen aux   ;
  Imagen edges = CreaImagen (anchura, altura);

  WORD   x,y;

  if (edges) {

     aux = Convolution (img, K0, 1);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = PIXEL(aux,x,y);

     DestruyeImagen (&aux);

     aux = Convolution (img, K1, 1);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = MAX(PIXEL(aux,x,y),PIXEL(edges,x,y));

     DestruyeImagen (&aux);

     aux = Convolution (img, K2, 1);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = MAX(PIXEL(aux,x,y),PIXEL(edges,x,y));

     DestruyeImagen (&aux);

     aux = Convolution (img, K3, 1);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = MAX(PIXEL(aux,x,y),PIXEL(edges,x,y));

     DestruyeImagen (&aux);

     aux = Convolution (img, K4, 1);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = MAX(PIXEL(aux,x,y),PIXEL(edges,x,y));

     DestruyeImagen (&aux);

     aux = Convolution (img, K5, 1);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = MAX(PIXEL(aux,x,y),PIXEL(edges,x,y));

     DestruyeImagen (&aux);

     aux = Convolution (img, K6, 1);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = MAX(PIXEL(aux,x,y),PIXEL(edges,x,y));

     DestruyeImagen (&aux);

     aux = Convolution (img, K7, 1);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(edges,x,y) = MAX(PIXEL(aux,x,y),PIXEL(edges,x,y));

     DestruyeImagen (&aux);
  }

  return edges;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*  PseudoPyramid Edge Detector                                            */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

Imagen PseudoPyramidEdgeDetector ( Imagen origen )
{
  WORD   anchura = ANCHURA(origen)/2;
  WORD   altura  = ALTURA(origen)/2;

  int    Laplacian [3][3] = { {-2, 1,-2}, { 1, 4, 1}, {-2, 1,-2} };

  Imagen aux1 = CreaImagen (anchura, altura);
  Imagen aux2;
  Imagen destino;

  WORD   x,y;

  for (x=0; x<anchura; x++)
      for (y=0; y<altura; y++)
	  PIXEL(aux1,x,y) = ( (int)PIXEL(origen,2*x,2*y)
                            + (int)PIXEL(origen,2*x,2*y+1)
                            + (int)PIXEL(origen,2*x+1,2*y)
                            + (int)PIXEL(origen,2*x+1,2*y+1) ) / 4;

  aux2    = ZoomImagen  ( aux1, anchura, altura );
  destino = Convolution ( aux2, Laplacian, 1 );

  DestruyeImagen(&aux1);
  DestruyeImagen(&aux2);

  return destino;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*  Range Edge Detector                                                    */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

Imagen RangeEdgeDetector (Imagen img)
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  Imagen edges   = CreaImagen (anchura-2, altura-2);

  int x,y,i,j;
  int min, max;

  if (edges)
     for (x=1; x<anchura-1; x++)
	 for (y=1; y<altura-1; y++) {

	     min = 0xFF;
	     max = 0x00;

	     for (i=-1; i<=1; i++)
		 for (j=-1; j<=1; j++) {
		     if (PIXEL(img,x+i,y+j)<min) min = PIXEL(img,x+i,y+j);
		     if (PIXEL(img,x+i,y+j)>max) max = PIXEL(img,x+i,y+j);
             }

	     PIXEL(edges,x-1,y-1) = max-min;
	 }

  return edges;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*  Difference Edge Detector                                               */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

Imagen DifferenceEdgeDetector (Imagen img)
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  Imagen edges   = CreaImagen (anchura-2, altura-2);

  WORD   x,y;
  int    dif, max;

  if (edges)
     for (x=1; x<anchura-1; x++)
	 for (y=1; y<altura-1; y++) {
	     max = abs ( (int)PIXEL(img,x+1,y+1) - (int)PIXEL(img,x-1,y-1) );
	     dif = abs ( (int)PIXEL(img,x+1,  y) - (int)PIXEL(img,x-1,  y) );
             if (dif>max) max=dif;
	     dif = abs ( (int)PIXEL(img,  x,y+1) - (int)PIXEL(img,  x,y-1) );
	     if (dif>max) max=dif;
	     dif = abs ( (int)PIXEL(img,x+1,y-1) - (int)PIXEL(img,x-1,y+1) );
	     if (dif>max) max=dif;
	     PIXEL(edges,x-1,y-1) = max;
	 }

  return edges;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*  Homogeneity Edge Detector                                              */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

Imagen HomogeneityEdgeDetector (Imagen img)
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  Imagen edges   = CreaImagen (anchura-2, altura-2);

  WORD   x,y;
  int    dif, max;

  if (edges)
     for (x=1; x<anchura-1; x++)
	 for (y=1; y<altura-1; y++) {
	     max = abs ( (int)PIXEL(img,x,y) - (int)PIXEL(img,x-1,y-1) );
	     dif = abs ( (int)PIXEL(img,x,y) - (int)PIXEL(img,x-1,  y) );
             if (dif>max) max=dif;
	     dif = abs ( (int)PIXEL(img,x,y) - (int)PIXEL(img,x-1,y+1) );
	     if (dif>max) max=dif;
	     dif = abs ( (int)PIXEL(img,x,y) - (int)PIXEL(img,  x,y-1) );
	     if (dif>max) max=dif;
	     dif = abs ( (int)PIXEL(img,x,y) - (int)PIXEL(img,  x,y+1) );
	     if (dif>max) max=dif;
	     dif = abs ( (int)PIXEL(img,x,y) - (int)PIXEL(img,x+1,y-1) );
	     if (dif>max) max=dif;
	     dif = abs ( (int)PIXEL(img,x,y) - (int)PIXEL(img,x+1,  y) );
	     if (dif>max) max=dif;
	     dif = abs ( (int)PIXEL(img,x,y) - (int)PIXEL(img,x+1,y+1) );
	     if (dif>max) max=dif;
	     PIXEL(edges,x-1,y-1) = max;
	 }

  return edges;
}


/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*  Cracks (Relajacin)                                                    */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

Imagen CrackEdgeDetector ( Imagen imagen, BYTE n )
{
  WORD anchura = ANCHURA(imagen);
  WORD altura  = ALTURA(imagen);

  BYTE punto, abajo, derecha;
  WORD x,y,c,i,j;
  BYTE crk[3],crk2[3],max,max2,cracks,cracks2;
  WORD pmax;

  Imagen vertical   = CreaImagen (anchura-1, altura-1);
  Imagen horizontal = CreaImagen (anchura-1, altura-1);
  Imagen destino    = CreaImagen (anchura-2, altura-2);

  for (x=0; x<anchura-1; x++)
      for (y=0; y<altura-1; y++) {
	  punto   = PIXEL(imagen,x,y);
	  abajo   = PIXEL(imagen,x,y+1);
	  derecha = PIXEL(imagen,x+1,y);
	  PIXEL(vertical,x,y) = abs((int)punto-(int)derecha);
	  PIXEL(horizontal,x,y) = abs((int)punto-(int)abajo);
      }


  for (c=0; c<n; c++) {

      /* Cracks verticales */

      for (x=0; x<anchura-2; x++)
	  for (y=1; y<altura-2; y++) {

	      /* Cracks */

	      crk  [1] = PIXEL(horizontal,x  ,y-1);
	      crk  [2] = PIXEL(horizontal,x+1,y-1);
	      crk2 [1] = PIXEL(horizontal,x  ,y  );
	      crk2 [2] = PIXEL(horizontal,x+1,y  );
	      crk  [0] = PIXEL(vertical  ,x,  y-1);
	      crk2 [0] = PIXEL(vertical  ,x,  y+1);

	      /* Ordenacin */

	      for (i=1; i<3; i++)
		  for (j=i+1; j<3; j++) {
		      if (crk[i]>crk[j]) {
			 max    = crk[i];
			 crk[i] = crk[j];
			 crk[j] = max;
		      }
		      if (crk2[i]>crk2[j]) {
			 max     = crk2[i];
			 crk2[i] = crk2[j];
			 crk2[j] = max;
		      }
		  }

	      max  = MAX (crk[2], 256/10);
	      max2 = MAX (crk2[2],256/10);

	      /* Estimacin del tipo (cracks,cracks2) */

	      pmax   = crk[0]*crk[1]*crk[2];
	      cracks = 3;

	      if (pmax<crk[2]*crk[1]*(max-crk[0])) {
		 pmax = crk[2]*crk[1]*(max-crk[0]);
		 cracks--;

		 if (pmax<crk[2]*(max-crk[1])*(max-crk[0])) {
		    pmax = crk[2]*(max-crk[1])*(max-crk[0]);
		    cracks--;

		    if (pmax<(max-crk[2])*(max-crk[1])*(max-crk[0]))
		       cracks--;
		 }
	      }

	      pmax    = crk2[0]*crk2[1]*crk2[2];
	      cracks2 = 3;

	      if (pmax<crk2[2]*crk2[1]*(max2-crk2[0])) {
		 pmax = crk2[2]*crk2[1]*(max2-crk2[0]);
		 cracks2--;

		 if (pmax<crk2[2]*(max2-crk2[1])*(max2-crk2[0])) {
		    pmax = crk2[2]*(max2-crk2[1])*(max2-crk2[0]);
		    cracks2--;

		    if (pmax<(max2-crk2[2])*(max2-crk2[1])*(max2-crk2[0]))
		       cracks2--;
		 }
	      }

	      /* Variacin del crack */

	      if (!cracks || !cracks2)
		 PIXEL(vertical,x,y) = ( (PIXEL(vertical,x,y)>0) ? PIXEL(vertical,x,y)-1: 0 );
	      else
		 PIXEL(vertical,x,y) = PIXEL(vertical,x,y)+1;

	  }

      /* Cracks horizontales */

      for (x=1; x<anchura-2; x++)
	  for (y=0; y<altura-2; y++) {

	      /* Cracks */

	      crk  [1] = PIXEL(vertical,x-1,y  );
	      crk  [2] = PIXEL(vertical,x-1,y+1);
	      crk2 [1] = PIXEL(vertical,x  ,y  );
	      crk2 [2] = PIXEL(vertical,x  ,y+1);
	      crk  [0] = PIXEL(horizontal,x-1,y);
	      crk2 [0] = PIXEL(horizontal,x+1,y);

	      /* Ordenacin */

	      for (i=1; i<3; i++)
		  for (j=i+1; j<3; j++) {
		      if (crk[i]>crk[j]) {
			 max    = crk[i];
			 crk[i] = crk[j];
			 crk[j] = max;
		      }
		      if (crk2[i]>crk2[j]) {
			 max     = crk2[i];
			 crk2[i] = crk2[j];
			 crk2[j] = max;
		      }
		  }

	      max  = MAX (crk[2], 256/10);
	      max2 = MAX (crk2[2],256/10);

	      /* Estimacin del tipo (cracks,cracks2) */

	      pmax   = crk[0]*crk[1]*crk[2];
	      cracks = 3;

	      if (pmax<crk[2]*crk[1]*(max-crk[0])) {
		 pmax = crk[2]*crk[1]*(max-crk[0]);
		 cracks--;

		 if (pmax<crk[2]*(max-crk[1])*(max-crk[0])) {
		    pmax = crk[2]*(max-crk[1])*(max-crk[0]);
		    cracks--;

		    if (pmax<(max-crk[2])*(max-crk[1])*(max-crk[0]))
		       cracks--;
		 }
	      }

	      pmax    = crk2[0]*crk2[1]*crk2[2];
	      cracks2 = 3;

	      if (pmax<crk2[2]*crk2[1]*(max2-crk2[0])) {
		 pmax = crk2[2]*crk2[1]*(max2-crk2[0]);
		 cracks2--;

		 if (pmax<crk2[2]*(max2-crk2[1])*(max2-crk2[0])) {
		    pmax = crk2[2]*(max2-crk2[1])*(max2-crk2[0]);
		    cracks2--;

		    if (pmax<(max2-crk2[2])*(max2-crk2[1])*(max2-crk2[0]))
		       cracks2--;
		 }
	      }

	      /* Variacin del crack */

	      if (!cracks || !cracks2)
		 PIXEL(horizontal,x,y) = ((PIXEL(horizontal,x,y)>0)?PIXEL(horizontal,x,y)-1:0);
	      else
		 PIXEL(horizontal,x,y) = PIXEL(horizontal,x,y)+1 ;

	  }
  }

  /* Mezcla de las dos imgenes de cracks */

  for (x=0; x<ANCHURA(destino); x++)
      for (y=0; y<ALTURA(destino); y++)
	  PIXEL(destino,x,y) = PIXEL(vertical,x,y)+PIXEL(horizontal,x,y);

  DestruyeImagen (&horizontal);
  DestruyeImagen (&vertical);

  return destino;
}

/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */
/*  Gradiente                                                            */
/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */

/* G1 */

Imagen Gradient1EdgeDetector (Imagen x)
{
  int    i,j;
  double z, dx, dy;

  WORD   anchura = ANCHURA(x);
  WORD   altura  = ALTURA(x);
  Imagen img     = CreaImagen (anchura,altura);

  for (i=anchura-1; i>= 1; i--)
       for (j=altura-1; j>=1; j--) {

	   dx = (int)PIXEL(x,i,j)-(int)PIXEL(x,i-1,j);
	   dy = (int)PIXEL(x,i,j)-(int)PIXEL(x,i,j-1);
	   z = sqrt(dx*dx + dy*dy);

	   if (z > 255.0)
	      z = 255.0;
	   else if (z < 0.0)
	      z = 0.0;

	   PIXEL(img,i,j) = (BYTE)z;
       }

  return img;
}

/* G2 */

Imagen Gradient2EdgeDetector (Imagen x)
{
  int    i,j;
  double z, dx, dy;

  WORD   anchura = ANCHURA(x);
  WORD   altura  = ALTURA(x);
  Imagen y       = CreaImagen (anchura,altura);

  for (i=1; i<anchura-1; i++)
      for (j=1; j<altura-1; j++) {

	  dx = (int)PIXEL(x,i,j+1)-(int)PIXEL(x,i,j-1);
	  dy = (int)PIXEL(x,i+1,j)-(int)PIXEL(x,i-1,j);
	  z = sqrt(dx*dx + dy*dy);

	  if (z > 255.0)
	     z = 255.0;
	  else if (z < 0.0)
	     z = 0.0;

	  PIXEL(y,i,j) = (BYTE)z;
      }

  return y;
}

/* G3: Gradiente de Roberts */

Imagen Gradient3EdgeDetector (Imagen x)
{
  int    i,j;
  double z, dx, dy;

  WORD   anchura = ANCHURA(x);
  WORD   altura  = ALTURA(x);
  Imagen y       = CreaImagen (anchura,altura);

  for (i=0; i<anchura-1; i++)
      for (j=0; j<altura-1; j++) {

	  dx = (int)PIXEL(x,i,j)-(int)PIXEL(x,i+1,j+1);
	  dy = (int)PIXEL(x,i+1,j)-(int)PIXEL(x,i,j+1);
	  z = sqrt(dx*dx + dy*dy);

	  if (z > 255.0)
	     z = 255.0;
	  else if (z < 0.0)
	     z = 0.0;

	  PIXEL(y,i,j) = (BYTE)z;
      }

  return y;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

#define MAX_MASK_SIZE 24

#define PI 3.14159265359

/* --- */

static float norm (float x, float y)
{
  return (float) sqrt ( (double)(x*x + y*y) );
}

static float distance (float a, float b, float c, float d)
{
  return norm ( (a-c), (b-d) );
}

/* --- */

static float gauss(float x, float sigma)
{
  return (float)exp((double) ((-x*x)/(2*sigma*sigma)));
}

static float meanGauss (float x, float sigma)
{
  float z = (gauss(x,sigma)+gauss(x+0.5,sigma)+gauss(x-0.5,sigma))/3.0;

  return z/(PI*2.0*sigma*sigma);
}

static float dGauss (float x, float sigma)
{
  return -x/(sigma*sigma) * gauss(x, sigma);
}

static float LoG (float x, float sigma)
{
  return (x*x-2*sigma*sigma)/(sigma*sigma*sigma*sigma) * gauss (x, sigma);
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*  Marr/Hildreth edge Detector                                            */
/* ----------------------------------------------------------------------- */

static Imagen zero_cross (FloatImg lap)
{
  int    i,j;
  Imagen img = CreaImagen(lap->anchura,lap->altura);
                                       
  for (i=1; i<ANCHURA(img)-1; i++)
      for (j=1; j<ALTURA(img)-1; j++)
	  if (  (lap->datos[i-1][j]*lap->datos[i+1][j]<0)
	     || (lap->datos[i][j-1]*lap->datos[i][j+1]<0)
	     || (lap->datos[i+1][j-1]*lap->datos[i-1][j+1]<0)
	     || (lap->datos[i-1][j-1]*lap->datos[i+1][j+1]<0) )
	     PIXEL(img,i,j)=0xFF;
	  else
	     PIXEL(img,i,j)=0x00;

  return img;
}

/* --- */

/* Parmetro 's'  Desviacin estndar de la distribucin gaussiana (1.0) */

Imagen MarrHildreth (Imagen img, float s)
{
  int      i,j,k,l,m,n;
  int      dim,width,ofs;
  FloatImg smx;
  WORD     anchura = ANCHURA(img);
  WORD     altura  = ALTURA(img);
  Imagen   edges;
  float    lg[MAX_MASK_SIZE][MAX_MASK_SIZE];

  /* Convolucin */

  width = 3.35*s + 0.33;
  dim   = 2*width + 1;
  ofs   = dim/2;

  for (i=0; i<dim; i++)
      for (j=0; j<dim; j++)
	  lg[i][j] = LoG ( distance((float)i,(float)j,(float)width,(float)width), s);

  smx = CreaFloatImg (anchura,altura);

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++) {
	  smx->datos[i][j] = 0.0;
	  for (k=0; k<dim; k++) {
	      n = i - ofs + k;
	      if (n>=0 && n<anchura)
                 for (l=0; l<dim; l++) {
		     m = j - ofs + l;
		     if (m>=0 && m<altura) 
			smx->datos[i][j] += lg[k][l]*PIXEL(img,n,m);
		 }
	  }
      }

  // Zero crossings
 
  edges = zero_cross (smx);

  DestruyeFloatImg(&smx);

  // Elimina el borde de la imagen 

  for (i=0; i<anchura; i++) {
      for (j=0; j<=width; j++)
	  PIXEL(edges,i,j) = 0x00;
      for (j=altura-width-1; j<altura; j++)
	  PIXEL(edges,i,j) = 0x00;
  }

  for (j=0; j<altura; j++) {
      for (i=0; i<=width; i++)
	  PIXEL(edges,i,j) = 0x00;
      for (i=anchura-width-1; i<anchura; i++)
	  PIXEL(edges,i,j) = 0x00;
  }

  return edges;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*  Canny Edge Detector                                                    */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

#define MAG_SCALE   20.0

int WIDTH = 0;

/* ----------------------------------------------------------------------- */

static void separable_convolution ( Imagen img, float *g, int width,
				    FloatImg smx, FloatImg smy)
{
  int   i,j,k, I1, I2;
  WORD  anchura = ANCHURA(img);
  WORD  altura  = ALTURA(img);
  float x, y;

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++) {

	  x = g[0] * PIXEL(img,i,j);
	  y = g[0] * PIXEL(img,i,j);

	  for (k=1; k<width; k++) {
	      I1 = (i+k)%anchura;
	      I2 = (i-k+anchura)%anchura;
	      y += g[k]*PIXEL(img,I1,j) + g[k]*PIXEL(img,I2,j);
	      I1 = (j+k)%altura;
	      I2 = (j-k+altura)%altura;
	      x += g[k]*PIXEL(img,i,I1) + g[k]*PIXEL(img,i,I2);
	  }

	  smx->datos[i][j] = x;
	  smy->datos[i][j] = y;
      }
}

/* ----------------------------------------------------------------------- */

static FloatImg dxy_separable_convolution ( FloatImg im,
					    float *g, int width, int dir)
{
  int      i,j,k,I1,I2;
  float    x;
  WORD     anchura = im->anchura;
  WORD     altura  = im->altura;
  FloatImg sm      = CreaFloatImg(anchura,altura);

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++) {

	  x = 0.0;

	  for (k=1; k<width; k++) {
	      if (dir == 0) {
		I1 = (i+k)%anchura;
		I2 = (i-k+anchura)%anchura;
		x += -g[k]*im->datos[I1][j] + g[k]*im->datos[I2][j];
	      } else {
		I1 = (j+k)%altura;
		I2 = (j-k+altura)%altura;
		x += -g[k]*im->datos[i][I1] + g[k]*im->datos[i][I2];
	      }
	  }
	  sm->datos[i][j] = x;
      }

  return sm;
}

/* ----------------------------------------------------------------------- */
/*  Nonmaximum suppression                                                 */

static Imagen nonmax_suppress ( FloatImg dx, FloatImg dy)
{
  int    i,j;
  float  xx, yy, g2, g1, g3, g4, g, xc, yc;

  WORD   anchura = dx->anchura;
  WORD   altura  = dx->altura;
  Imagen mag     = CreaImagen (anchura, altura);

  for (i=1; i<anchura-1; i++) {
      for (j=1; j<altura-1; j++) {

	  xc = dx->datos[i][j];
	  yc = dy->datos[i][j];
	  if (fabs(xc)<0.01 && fabs(yc)<0.01) continue;
	  g  = norm (xc, yc);

	  /* Follow the gradient direction, as indicated by the direction of
	     the vector (xc, yc); retain pixels that are a local maximum. */

	  if (fabs(yc) > fabs(xc)) {

	      /* The Y component is biggest,
		 so gradient direction is basically UP/DOWN */

	      xx = fabs(xc)/fabs(yc);
	      yy = 1.0;

	      g2 = norm (dx->datos[i-1][j], dy->datos[i-1][j]);
	      g4 = norm (dx->datos[i+1][j], dy->datos[i+1][j]);

	      if (xc*yc > 0.0) {
		g3 = norm (dx->datos[i+1][j+1], dy->datos[i+1][j+1]);
		g1 = norm (dx->datos[i-1][j-1], dy->datos[i-1][j-1]);
	      } else {
		g3 = norm (dx->datos[i+1][j-1], dy->datos[i+1][j-1]);
		g1 = norm (dx->datos[i-1][j+1], dy->datos[i-1][j+1]);
	      }

	  } else {

	      /* The X component is biggest,
		 so gradient direction is basically LEFT/RIGHT */

	      xx = fabs(yc)/fabs(xc);
	      yy = 1.0;

	      g2 = norm (dx->datos[i][j+1], dy->datos[i][j+1]);
	      g4 = norm (dx->datos[i][j-1], dy->datos[i][j-1]);

	      if (xc*yc > 0.0) {
		g3 = norm (dx->datos[i-1][j-1], dy->datos[i-1][j-1]);
		g1 = norm (dx->datos[i+1][j+1], dy->datos[i+1][j+1]);
	      } else {
		g1 = norm (dx->datos[i-1][j+1], dy->datos[i-1][j+1]);
		g3 = norm (dx->datos[i+1][j-1], dy->datos[i+1][j-1]);
	      }
	  }

	  /* Compute the interpolated value of the gradient magnitude */

	  if ( (g > (xx*g1 + (yy-xx)*g2)) && (g > (xx*g3 + (yy-xx)*g4)) ) {

	     if (g*MAG_SCALE <= 255)
		PIXEL(mag,i,j) = (BYTE)(g*MAG_SCALE);
	     else
		PIXEL(mag,i,j) = 255;

	  } else {

	     PIXEL(mag,i,j) = 0;

	  }
      }
  }

  return mag;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

/* Parmetros:                                         */
/* -----------                                         */
/* low  ---> Umbral inferior                           */
/* high ---> Umbral superior                           */
/* s    ---> Desviacin estndar de la gaussiana (1.0) */

/* low == high ---> Sin histresis                     */
/* low >  high ---> Seleccin automtica de low y high */

Imagen CannyEdgeDetector (Imagen img, BYTE low, BYTE high, float s)
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  int    i,j;
  Imagen mag;
  Imagen aux;

  /* -------- */
  /* Filtrado */
  /* -------- */

  int      width;

  FloatImg smx,smy,dx,dy;

  float    g[MAX_MASK_SIZE], dg[MAX_MASK_SIZE];

  /* Gaussian and derivative of Gaussian filter mask */
  /* ----------------------------------------------- */

  for (i=0; i<MAX_MASK_SIZE; i++) {

      g[i] = meanGauss ((float)i, s);

      if (g[i] < 0.005) {
	 width = i;
	 break;
      }

      dg[i] = dGauss ((float)i, s);
  }

  WIDTH = width/2;

  /* Smoothing */
  /* --------- */

  /* Convolucin con 'g' en las direcciones X e Y  */

  smx = CreaFloatImg(anchura,altura);
  smy = CreaFloatImg(anchura,altura);

  separable_convolution (img, g, width, smx, smy);

  /* Convolucin con 'dg' en las direcciones X e Y */

  dx = dxy_separable_convolution (smx, dg, width, 1);

  DestruyeFloatImg(&smx);

  dy = dxy_separable_convolution (smy, dg, width, 0);

  DestruyeFloatImg(&smy);

  /* A */  /* ------------------------ */
	   /* Combinacin: norm(dx,dy) */
	   /* ------------------------ */

  // for (i=0; i<anchura; i++)
  //     for (j=0; j<altura; j++)
  //         PIXEL(mag,i,j) =  (BYTE)(MAG_SCALE*norm(dx[i][j],dy[i][j]));


  /* B */ /* ----------------------- */
	  /* Non-maximum suppression */
	  /* ----------------------- */

  mag = nonmax_suppress (dx, dy);

  DestruyeFloatImg(&dx);
  DestruyeFloatImg(&dy);

  /* ---------- */
  /* Histresis */
  /* ---------- */

  if (low!=high) {

     for (i=0; i<WIDTH; i++)
	 for (j=0; j<altura; j++)
	     PIXEL(mag,i,j) = 0x00;

     for (i=anchura-WIDTH; i<anchura; i++)
	 for (j=0; j<altura; j++)
	     PIXEL(mag,i,j) = 0x00;

     for (i=0; i<anchura; i++)
	 for (j=0; j<WIDTH; j++)
	     PIXEL(mag,i,j) = 0x00;

     for (i=0; i<anchura; i++)
	 for (j=altura-WIDTH; j<altura; j++)
	     PIXEL(mag,i,j) = 0x00;


     aux = Hysteresis (mag, high, low);

     DestruyeImagen(&mag);

     return aux;

  } else {

     return mag;

  }

}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*  Shen-Castan Edge Detector: ISEF(infinite symmetric exponential filter) */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

#define OUTLINE 25

float low_thresh=20;
float high_thresh=22;
float b;
float ratio;
int   window_size;
int   do_hysteresis;
int   thinFactor;

// Umbralizacin

FloatImg lap;			// Laplaciana de la imagen
int      nr, nc;		// Nmero de filas y columnas
Imagen   edges;			// Contornos


static void estimate_thresh (float *low, float *hi, int nr, int nc)
{
  float vmax, vmin, scale, x;
  int i,j,k, hist[256], count;

  // Histograma de la laplaciana

  vmin = vmax = fabs(lap->datos[OUTLINE][OUTLINE]);

  for (i=0; i<nr; i++)
      for (j=0; j<nc; j++)
	  if (i>=OUTLINE && i<nr-OUTLINE && j>=OUTLINE && j<nc-OUTLINE) {
	      x = lap->datos[i][j];
	      if (vmin > x) vmin = x;
	      if (vmax < x) vmax = x;
	  }

  for (k=0; k<256; k++)
      hist[k] = 0;

  scale = 256.0/(vmax-vmin + 1);

  for (i=0; i<nr; i++)
      for (j=0; j<nc; j++)
	  if (i>=OUTLINE && i<nr-OUTLINE && j>=OUTLINE && j<nc-OUTLINE) {
	     x = lap->datos[i][j];
	     k = (int)((x - vmin)*scale);
	     hist[k] ++;
	  }

  // El umbral superior debera estar por encima del 80 o del 90% de los pxeles

  k = 255;
  j = (int)(ratio*nr*nc);
  count = hist[255];

  while (count < j) {
	k--;
	if (k<0) break;
	count += hist[k];
  }

  *hi  = (double)k/scale + vmin ;
  *low = (*hi)/2;
}


static int mark_connected (int i, int j, int level)
{
  int notChainEnd;

  if (i >= nr || i < 0 || j >= nc || j < 0)
     return 0;

  if (PIXEL(edges,i,j) != 0)
     return 0;

  if (lap->datos[i][j] == 0)
     return 0;

  if (lap->datos[i][j] > low_thresh)
     PIXEL(edges,i,j) = 1;
  else
     PIXEL(edges,i,j) = 0xFF;

  notChainEnd =0;
  notChainEnd |= mark_connected(i  ,j+1, level+1);
  notChainEnd |= mark_connected(i  ,j-1, level+1);
  notChainEnd |= mark_connected(i+1,j+1, level+1);
  notChainEnd |= mark_connected(i+1,j  , level+1);
  notChainEnd |= mark_connected(i+1,j-1, level+1);
  notChainEnd |= mark_connected(i-1,j-1, level+1);
  notChainEnd |= mark_connected(i-1,j  , level+1);
  notChainEnd |= mark_connected(i-1,j+1, level+1);

  if (notChainEnd && ( level > 0 ) )
     if ( thinFactor > 0 )
	if ( (level%thinFactor) != 0  )
	   PIXEL(edges,i,j) = 0xFF;

  return 1;
}


static Imagen threshold_edges (FloatImg in)
{
  int    i,j;
  WORD   anchura = in->anchura;
  WORD   altura  = in->altura;
  Imagen out = CreaImagen(anchura,altura);

  lap   = in;
  edges = out;
  nr    = anchura;
  nc    = altura;

  estimate_thresh (&low_thresh, &high_thresh, nr, nc);

  if (!do_hysteresis)
     low_thresh = high_thresh;

  for (i=0; i<nr; i++)
      for (j=0; j<nc; j++)
	  PIXEL(edges,i,j) = 0;

  for (i=0; i<nr; i++)
      for (j=0; j<nc; j++)
	  if (i>=OUTLINE && i<nr-OUTLINE && j>=OUTLINE && j<nc-OUTLINE)
	     if (lap->datos[i][j] > high_thresh)
		mark_connected (i,j,0);

  for (i=0; i<nr; i++)
      for (j=0; j<nc; j++)
	  if (PIXEL(edges,i,j)==0xFF)
	     PIXEL(edges,i,j)=0;

  return out;
}


// Zero-crossings de la laplaciana

static int is_candidate_edge (Imagen buff, FloatImg orig, int row, int col)
{
  if ( PIXEL(buff,row,col)==1 && PIXEL(buff,row+1,col)==0 ) {        // positive z-c

     if (orig->datos[row+1][col] - orig->datos[row-1][col] > 0)
	return 1;
     else
	return 0;

  } else if ( PIXEL(buff,row,col)==1 && PIXEL(buff,row,col+1)==0 ) { // positive z-c

     if (orig->datos[row][col+1] - orig->datos[row][col-1] > 0)
	return 1;
     else
	return 0;

  } else if ( PIXEL(buff,row,col)==1 && PIXEL(buff,row-1,col)==0 ) { // negative z-c

     if (orig->datos[row+1][col] - orig->datos[row-1][col] < 0)
	return 1;
     else
	return 0;

  } else if ( PIXEL(buff,row,col)==1 && PIXEL(buff,row,col-1)==0 ) { // negative z-c

     if (orig->datos[row][col+1] - orig->datos[row][col-1] < 0)
	return 1;
     else
	return 0;

  } else return 0;	// not a z-c
}


static float compute_adaptive_gradient (Imagen BLI_buffer, FloatImg orig_buffer, int row, int col)
{
  int i, j;
  float sum_on, sum_off;
  float avg_on, avg_off;
  int   num_on, num_off;

  sum_on = sum_off = 0.0;
  num_on = num_off = 0;

  for (i= (-window_size/2); i<=(window_size/2); i++) {
      for (j=(-window_size/2); j<=(window_size/2); j++) {

	  if (PIXEL(BLI_buffer,row+i,col+j)) {
	     sum_on += orig_buffer->datos[row+i][col+j];
	     num_on ++;
	  } else {
	     sum_off += orig_buffer->datos[row+i][col+j];
	     num_off++;
	  }
      }
  }

  if (num_off>0)
     avg_off = sum_off / num_off;
  else
     avg_off = 0;

  if (num_on>0)
     avg_on = sum_on / num_on;
  else
     avg_on = 0;

  return (avg_off - avg_on);
}


static void locate_zero_crossings (FloatImg orig, FloatImg smoothed,
				   Imagen bli, int anchura, int altura)
{
  int i, j;

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++)
	  if (i<OUTLINE || i>=anchura-OUTLINE || j<OUTLINE || j>=altura-OUTLINE)
	     orig->datos[i][j] = 0.0;
	  else if (is_candidate_edge (bli, smoothed, i, j))
	     orig->datos[i][j] = compute_adaptive_gradient (bli, smoothed, i, j);
	  else
	     orig->datos[i][j] = 0.0;
}


// ISEF (Shen & Castan CVIGP March 1992)

static FloatImg apply_ISEF_vertical (FloatImg x)
{
  WORD     anchura = x->anchura;
  WORD     altura  = x->altura;
  FloatImg y       = CreaFloatImg(anchura,altura);

  int      i,j;
  float    A[2048], B[2048];
  float    b1, b2;


  b1 = (1.0 - b)/(1.0 + b);
  b2 = b*b1;

  for (j=0; j<altura; j++) {

      A[0] = b1 * x->datos[0][j];

      for (i=1; i<anchura; i++)
          A[i] = b1 * x->datos[i][j] + b * A[i-1];

      B[anchura-1] = b2 * x->datos[anchura-1][j];

      for (i=anchura-2; i>=0; i--)
          B[i] = b2 * x->datos[i][j] + b * B[i+1];

      y->datos[anchura-1][j] = A[anchura-1];

      for (i=0; i<anchura-1; i++)
          y->datos[i][j] = A[i] + B[i+1];
  }

  return y;
}


static FloatImg apply_ISEF_horizontal (FloatImg x)
{
  int      i,j,k;
  float    b1, b2;
  float    A[2048], B[2048];
  WORD     anchura = x->anchura;
  WORD     altura  = x->altura;
  FloatImg y       = CreaFloatImg(anchura,altura);

  b1 = (1.0 - b)/(1.0 + b);
  b2 = b*b1;

  for (i=0; i<anchura; i++) {

      A[0] = b1 * x->datos[i][0];

      for (j=1; j<altura; j++)
	  A[j] = b1 * x->datos[i][j] + b * A[j-1];

      B[altura-1] = b2 * x->datos[i][altura-1];

      for (j=altura-2; j>=0; j--)
	  B[j] = b2 * x->datos[i][j] + b * B[j+1];

      y->datos[i][altura-1] = A[altura-1];

      for (j=0; j<altura-1; j++)
	  y->datos[i][j] = A[j] + B[j+1];
  }

  return y;
}

// r: Porcentaje de pxeles por encima del umbral alto        0.99
// s: Factor de suavizado (0<s<1)                             0.9
// w: Tamao de la ventana para calcular el gradiente         7
// t: Contour thinning factor (nmero de pixeles)
// d: Realizar histresis?                                   1

Imagen ShenCastanEdgeDetector (Imagen img, float r, float s, int w, int t, int d)
{
  int      i,j;
  int      width    = OUTLINE+2;
  WORD     anchura  = ANCHURA(img)+width+width;
  WORD     altura   = ALTURA(img)+width+width;
  Imagen   bli,res,out;
  FloatImg buffer,smoothed,aux;

  ratio         = r;
  b             = s;
  window_size   = w;
  thinFactor    = t;
  do_hysteresis = d;

  // img ---> buffer

  buffer   = CreaFloatImg(anchura,altura);

  for (i=0; i<ANCHURA(img); i++)
      for (j=0; j<ALTURA(img); j++)
	  buffer->datos[i+width][j+width] = PIXEL(img,i,j);

  // Suavizado: ISEF

  aux      = apply_ISEF_vertical   (buffer);
  smoothed = apply_ISEF_horizontal (aux);

  DestruyeFloatImg(&aux);

  // bli (band-limited laplacian image)

  bli = CreaImagen(anchura,altura);

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++)
	  if (i>=OUTLINE && i<anchura-OUTLINE && j>=OUTLINE && j<altura-OUTLINE)
	     PIXEL(bli,i,j) = ((smoothed->datos[i][j] - buffer->datos[i][j]) > 0.0);

  // Deteccin de aristas

  locate_zero_crossings (buffer, smoothed, bli, anchura,altura);

  DestruyeFloatImg(&smoothed);
  DestruyeImagen(&bli);

  // Umbralizacin

  res = threshold_edges (buffer);

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++)
	  if (PIXEL(res,i,j) > 0)
	     PIXEL(res,i,j) = 0xFF;
	  else
	     PIXEL(res,i,j) = 0x00;

  DestruyeFloatImg(&buffer);

  // Postprocesamiento

  out = CreaImagen(ANCHURA(img), ALTURA(img));

  for (i=0; i<ANCHURA(out)-1; i++)
      for (j=1; j<ALTURA(out); j++)
	  PIXEL(out,i,j) = PIXEL(res,i+width,j+width);

  for (j=1; j<ALTURA(out); j++)
      PIXEL(out,ANCHURA(out)-1,j) = 0xFF;

  for (i=0; i<ANCHURA(out)-1; i++)
      PIXEL(out,i,0) = 0xFF;

  DestruyeImagen(&res);

  return out;
}

/* ----------------------------------------------------------------------- */
/*  Phase congruency: Local energy                                         */
/* ----------------------------------------------------------------------- */
/*

#define VENTANA 3

float LocalEnergy ( Imagen img, int x, int y )    // sqrt(f^2+h^2)
{
  int   i,px,py;
  float f,h,v,d1,d2;

  // Horizontal

  h = 0;

  for (i=-VENTANA; i<=VENTANA; i++)
      if (i) {
         px = (x+i+ANCHURA(img))%ANCHURA(img);
         h += ((float)PIXEL(img,px,y)/(float)i);
      }

  // Vertical

  v = 0;

  for (i=-VENTANA; i<=VENTANA; i++)
      if (i) {
         py = (y+i+ALTURA(img))%ALTURA(img);
         v += ((float)PIXEL(img,x,py)/(float)i);
      }

  // Diagonal 1

  d1 = 0;

  for (i=-VENTANA; i<=VENTANA; i++)
      if (i) {
         px = (x+i+ANCHURA(img))%ANCHURA(img);
         py = (y+i+ALTURA(img))%ALTURA(img);
         d1 += ((float)PIXEL(img,px,py)/(float)i);
      }

  // Diagonal 2

  d2 = 0;

  for (i=-VENTANA; i<=VENTANA; i++)
      if (i) {
         px = (x+i+ANCHURA(img))%ANCHURA(img);
         py = (y-i+ALTURA(img))%ALTURA(img);
         d2 += ((float)PIXEL(img,px,py)/(float)i);
      }

  // Salida

  f = PIXEL(img,x,y);

  return sqrt( f*f +  h*h  )
       + sqrt( f*f +  v*v  )
       + sqrt( f*f + d1*d1 )
       + sqrt( f*f + d2*d2 );

  // return abs(h) + abs(v) + abs(d1) + abs(d2);
}

/*

Diseo de los filtros
---------------------
   
#include <math.h>
#include <complex.h>
#include <stdio.h>

#define N 6

int main (void)
{
  int i,j;
  int n = N;
  complex x[N],s[N],r[N];
  complex w;
  complex f;
				// Filtro: exp(-x^2/s^2) s=2 * sin ()
  x[0] = 0;
  x[1] = 0.77880078307*sin(1);
  x[2] = 0.36787944117*sin(2);
  x[3] = 0.10539922456*sin(3);
  x[4] = 0.01831563888*sin(4);
  x[5] = 0.00193045413623*sin(5);

  w = exp(complex(0,-2*M_PI/n));

  // DFT

  for (i=0; i<n; i++) {

      s[i] = 0;

      for (j=0; j<4; j++)
	  s[i] += x[j]*pow(w,i*j);

      s[i] /= n;
  }

  // Shift

  f = complex(0,1);

  for (i=0; i<n; i++)
      s[i] *= f;

  // DFT inversa

  for (i=0; i<n; i++) {

      r[i] = 0;

      for (j=0; j<4; j++)
	  r[i] += s[j]*pow(w,i*j);
  }

  // Orden del resultado: 0 N-1 N-2 ... 3 2 1

  return 0;
}

/**/

#define VENTANA 4

float Fodd[VENTANA+1]  = { 0.000000, 0.655338,  0.334512,  0.014874, -0.013861 };
float Feven[VENTANA+1] = { 0.189179, 0.096565, -0.184886, -0.096565, -0.004294 };

float LocalEnergy ( Imagen img, int x, int y )
{
  int   i, pxN, pxP, pyN, pyP;

  float Heven, Veven, D1even, D2even;
  float Hodd,  Vodd,  D1odd,  D2odd;

  Heven = Veven = D1even = D2even = 2*Feven[0]*PIXEL(img,x,y);
  Hodd  = Vodd  = D1odd  = D2odd  = 0;

  for (i=1; i<=VENTANA; i++) {
      pxP = (x+i)%ANCHURA(img);
      pxN = (x-i+ANCHURA(img))%ANCHURA(img);
      pyP = (y+i)%ALTURA(img);
      pyN = (y-i+ALTURA(img))%ALTURA(img);

      // Horizontal

      Heven += Feven[i]*((float)PIXEL(img,pxP,y)+(float)PIXEL(img,pxN,y));
      Hodd  +=  Fodd[i]*((float)PIXEL(img,pxP,y)-(float)PIXEL(img,pxN,y));

      // Vertical

      Veven += Feven[i]*((float)PIXEL(img,x,pyP)+(float)PIXEL(img,x,pyN));
      Vodd  +=  Fodd[i]*((float)PIXEL(img,x,pyP)-(float)PIXEL(img,x,pyN));

      // Diagonal 1

      D1even += Feven[i]*((float)PIXEL(img,pxP,pyP)+(float)PIXEL(img,pxN,pyN));
      D1odd  +=  Fodd[i]*((float)PIXEL(img,pxP,pyP)-(float)PIXEL(img,pxN,pyN));

      // Diagonal 2

      D2even += Feven[i]*((float)PIXEL(img,pxP,pyN)+(float)PIXEL(img,pxN,pyP));
      D2odd  +=  Fodd[i]*((float)PIXEL(img,pxP,pyN)-(float)PIXEL(img,pxN,pyP));
  }

  return   sqrt (  Heven*Heven + Hodd*Hodd  )
         + sqrt (  Veven*Veven + Vodd*Vodd  )
         + sqrt ( D1even*D1even+D1odd*D1odd )
         + sqrt ( D2even*D2even+D2odd*D2odd );
}

/* Detector de contornos */

Imagen PhaseCongruencyEdgeDetector (Imagen img)
{
  WORD     anchura  = ANCHURA(img);
  WORD     altura   = ALTURA(img);
  Imagen   salida;
  FloatImg auxiliar = CreaFloatImg(anchura,altura);
  int      i,j;

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++)
          auxiliar->datos[i][j] = LocalEnergy(img,i,j);

  salida = ImagenFloat(auxiliar);

  DestruyeFloatImg(&auxiliar);

  return salida;
}

/* ------------------------------------------------------------------------- */
