/* ----------------------------------------------------------------------- */
/*    Fernando Berzal Galiano                ETS Ingeniera Informtica    */
/* ----------------------------------------------------------------------- */
/*                       Procesamiento de imgenes                         */
/* ----------------------------------------------------------------------- */
/*                Deteccin de contornos: EDGE DETECTION                   */
/* ----------------------------------------------------------------------- */

#include "IMG.H"

/* ----------------------------------------------------------------------- */
/*  Edge detection (2x2)                                                   */
/* ----------------------------------------------------------------------- */
/*   Corresponde al ajuste de superficies (facet model) cuando N=2x2 y la  */
/*  superficie ajustada sea un plano (Ax+By+C).                            */

Imagen MaskEdgeDetector2x2 ( Imagen img );

/* ----------------------------------------------------------------------- */
/*  Edge detection (3x3)                                                   */
/* ----------------------------------------------------------------------- */

Imagen MaskEdgeDetector3x3 ( Imagen img );

/* -1  0 -1 */
/*  0  4  0 */
/* -1  0 -1 */

Imagen PhillipsMaskEdgeDetector ( Imagen img );

/*  0  1  0 */
/* -1  0  1 */
/*  0 -1  0 */

Imagen FalerMaskEdgeDetector ( Imagen img );

/* ----------------------------------------------------------------------- */
/*  Laplaciana                                                             */
/* ----------------------------------------------------------------------- */

Imagen Laplacian1EdgeDetector ( Imagen img );
Imagen Laplacian2EdgeDetector ( Imagen img );
Imagen Laplacian3EdgeDetector ( Imagen img );

/* ----------------------------------------------------------------------- */
/*  Gradiente                                                              */
/* ----------------------------------------------------------------------- */

Imagen Gradient1EdgeDetector (Imagen x);
Imagen Gradient2EdgeDetector (Imagen x);
Imagen Gradient3EdgeDetector (Imagen x);  /* Gradiente de Roberts */


/* ----------------------------------------------------------------------- */
/*  Sobel                                                                  */
/* ----------------------------------------------------------------------- */

Imagen SobelEdgeDetector ( Imagen img );

/* ----------------------------------------------------------------------- */
/*  Kirsch                                                                 */
/* ----------------------------------------------------------------------- */

Imagen KirschEdgeDetector ( Imagen img );

/* ----------------------------------------------------------------------- */
/*  Prewitt                                                                */
/* ----------------------------------------------------------------------- */

Imagen PrewittEdgeDetector ( Imagen img );

/* ----------------------------------------------------------------------- */
/*  Pseudo-piramidal                                                       */
/* ----------------------------------------------------------------------- */

Imagen PseudoPyramidEdgeDetector ( Imagen origen );

/* ----------------------------------------------------------------------- */
/*  Rango                                                                  */
/* ----------------------------------------------------------------------- */

Imagen RangeEdgeDetector (Imagen img);

/* ----------------------------------------------------------------------- */
/*  Diferencias                                                            */
/* ----------------------------------------------------------------------- */

Imagen DifferenceEdgeDetector (Imagen img);

/* ----------------------------------------------------------------------- */
/*  Homogeneidad                                                           */
/* ----------------------------------------------------------------------- */

Imagen HomogeneityEdgeDetector (Imagen img);

/* ----------------------------------------------------------------------- */
/*  Cracks (relajacin)                                                    */
/* ----------------------------------------------------------------------- */

Imagen CrackEdgeDetector ( Imagen imagen, BYTE n );

/* ----------------------------------------------------------------------- */
/*  Marr/Hildreth                                                          */
/* ----------------------------------------------------------------------- */
/* Parmetro 's' = Desviacin estndar de la distribucin gaussiana (<=3)  */

Imagen MarrHildreth (Imagen img, float s);

/* ----------------------------------------------------------------------- */
/*  Canny                                                                  */
/* ----------------------------------------------------------------------- */
/*                                                                         */
/* Parmetros:                                                             */
/* -----------                                                             */
/* low  ---> Lower threshold (Hysteresis thresholding)                     */
/* high ---> High threshold (Hysteresis thresholding)                      */
/* s    ---> Desviacin estndar de la distribucin gausiana (1.0)         */
/*                                                                         */
/* low == high ---> Sin "Hysteresis thresholding"                          */
/* low >  high ---> Seleccin automtica de low y high                     */

Imagen CannyEdgeDetector (Imagen img, BYTE low, BYTE high, float s);

/* ----------------------------------------------------------------------- */
/*  Shen-Castan: ISEF(infinite symmetric exponential filter)               */
/* ----------------------------------------------------------------------- */

// r: Porcentaje de pixeles por encima del umbral alto        0.99
// s: Factor de suavizado (0<s<1)                             0.9
// w: Tamao de la ventana para calcular el gradiente         7
// t: Contour thinning factor (nmero de pixeles)             1
// d: Realizar hysteresis thresholding?                      1

Imagen ShenCastanEdgeDetector (Imagen im, float r, float s, int w, int t, int d);

/* ----------------------------------------------------------------------- */
/*  Phase congruency: Local energy                                         */
/* ----------------------------------------------------------------------- */

Imagen PhaseCongruencyEdgeDetector (Imagen img);

