/* ----------------------------------------------------------------------- */
/*    Fernando Berzal Galiano                ETS Ingeniera Informtica    */
/* ----------------------------------------------------------------------- */
/*                       Procesamiento de imgenes                         */
/* ----------------------------------------------------------------------- */
/*                               Gradiente                                 */
/* ----------------------------------------------------------------------- */

#include <math.h>

#include "IMG_GRAD.H"
#include "IMG_TYPE.H"

/* --- */

Imagen Gradient1SQRT (Imagen x)
{
  int    i,j;
  double z, dx, dy;

  WORD   anchura = ANCHURA(x);
  WORD   altura  = ALTURA(x);
  Imagen img     = CreaImagen (anchura,altura);

  for (i=anchura-1; i>= 1; i--)
       for (j=altura-1; j>=1; j--) {

	   dx = (int)PIXEL(x,i,j)-(int)PIXEL(x,i-1,j);
	   dy = (int)PIXEL(x,i,j)-(int)PIXEL(x,i,j-1);
	   z = sqrt(dx*dx + dy*dy);

	   if (z > 255.0)
	      z = 255.0;
	   else if (z < 0.0)
	      z = 0.0;

	   PIXEL(img,i,j) = (BYTE)z;
       }

  return img;
}

Imagen Gradient1ABS (Imagen x)
{
  int i,j;
  int z, dx, dy;

  WORD   anchura = ANCHURA(x);
  WORD   altura  = ALTURA(x);
  Imagen img     = CreaImagen (anchura,altura);

  for (i=anchura-1; i>= 1; i--)
       for (j=altura-1; j>=1; j--) {
	   dx = (int)PIXEL(x,i,j)-(int)PIXEL(x,i-1,j);
	   dy = (int)PIXEL(x,i,j)-(int)PIXEL(x,i,j-1);
	   z = abs(dx)+abs(dy);
	   PIXEL(img,i,j) = z/2;
       }

  return img;
}

Imagen Gradient1X (Imagen x)
{
  int i,j;
  int dx;

  WORD   anchura = ANCHURA(x);
  WORD   altura  = ALTURA(x);
  Imagen img     = CreaImagen (anchura,altura);

  for (i=anchura-1; i>= 1; i--)
       for (j=altura-1; j>=1; j--) {
	   dx = (int)PIXEL(x,i,j)-(int)PIXEL(x,i-1,j);
	   PIXEL(img,i,j) = (255+dx)/2;
       }

  return img;
}

Imagen Gradient1Y (Imagen x)
{
  int i,j;
  int dy;

  WORD   anchura = ANCHURA(x);
  WORD   altura  = ALTURA(x);
  Imagen img     = CreaImagen (anchura,altura);

  for (i=anchura-1; i>= 1; i--)
       for (j=altura-1; j>=1; j--) {
	   dy = (int)PIXEL(x,i,j)-(int)PIXEL(x,i,j-1);
	   PIXEL(img,i,j) = (255+dy)/2;
       }

  return img;
}

/* --- */

Imagen Gradient2SQRT (Imagen x)
{
  int    i,j;
  double z, dx, dy;

  WORD   anchura = ANCHURA(x);
  WORD   altura  = ALTURA(x);
  Imagen y       = CreaImagen (anchura,altura);

  for (i=1; i<anchura-1; i++)
      for (j=1; j<altura-1; j++) {

	  dx = (int)PIXEL(x,i,j+1)-(int)PIXEL(x,i,j-1);
	  dy = (int)PIXEL(x,i+1,j)-(int)PIXEL(x,i-1,j);
	  z = sqrt(dx*dx + dy*dy);

	  if (z > 255.0)
	     z = 255.0;
	  else if (z < 0.0)
	     z = 0.0;

	  PIXEL(y,i,j) = (BYTE)z;
      }

  return y;
}

Imagen Gradient2ABS (Imagen x)
{
  int i, j;
  int z, dx, dy;

  WORD   anchura = ANCHURA(x);
  WORD   altura  = ALTURA(x);
  Imagen y       = CreaImagen (anchura,altura);

  for (i=1; i<anchura-1; i++)
      for (j=1; j<altura-1; j++) {
	  dx = (int)PIXEL(x,i,j+1)-(int)PIXEL(x,i,j-1);
	  dy = (int)PIXEL(x,i+1,j)-(int)PIXEL(x,i-1,j);
	  z = abs(dx)+abs(dy);
	  PIXEL(y,i,j) = z/2;
      }

  return y;
}

Imagen Gradient2X (Imagen x)
{
  int i,j;
  int dx;

  WORD   anchura = ANCHURA(x);
  WORD   altura  = ALTURA(x);
  Imagen y       = CreaImagen (anchura,altura);

  for (i=1; i<anchura-1; i++)
      for (j=1; j<altura-1; j++) {
	  dx = (int)PIXEL(x,i+1,j)-(int)PIXEL(x,i-1,j);
	  PIXEL(y,i,j) = (255+dx)/2;
      }

  return y;
}

Imagen Gradient2Y (Imagen x)
{
  int i,j;
  int dy;

  WORD   anchura = ANCHURA(x);
  WORD   altura  = ALTURA(x);
  Imagen y       = CreaImagen (anchura,altura);

  for (i=1; i<anchura-1; i++)
      for (j=1; j<altura-1; j++) {
	  dy = (int)PIXEL(x,i,j+1)-(int)PIXEL(x,i,j-1);
	  PIXEL(y,i,j) = (255+dy)/2;
      }

  return y;
}

/* Gradiente de Roberts */

Imagen GradientRobertsSQRT (Imagen x)
{
  int    i,j;
  double z, dx, dy;

  WORD   anchura = ANCHURA(x);
  WORD   altura  = ALTURA(x);
  Imagen y       = CreaImagen (anchura,altura);

  for (i=0; i<anchura-1; i++)
      for (j=0; j<altura-1; j++) {

	  dx = (int)PIXEL(x,i,j)-(int)PIXEL(x,i+1,j+1);
	  dy = (int)PIXEL(x,i+1,j)-(int)PIXEL(x,i,j+1);
	  z = sqrt(dx*dx + dy*dy);

	  if (z > 255.0)
	     z = 255.0;
	  else if (z < 0.0)
	     z = 0.0;

	  PIXEL(y,i,j) = (BYTE)z;
      }

  return y;
}

Imagen GradientRobertsABS (Imagen x)
{
  int i,j;
  int z, dx, dy;

  WORD   anchura = ANCHURA(x);
  WORD   altura  = ALTURA(x);
  Imagen y       = CreaImagen (anchura,altura);

  for (i=0; i<anchura-1; i++)
      for (j=0; j<altura-1; j++) {
	  dx = (int)PIXEL(x,i,j)-(int)PIXEL(x,i+1,j+1);
	  dy = (int)PIXEL(x,i+1,j)-(int)PIXEL(x,i,j+1);
	  z = abs(dx)+abs(dy);
	  PIXEL(y,i,j) = (BYTE)(z/2);
      }

  return y;
}

Imagen GradientRobertsX (Imagen x)
{
  int i,j;
  int dx;

  WORD   anchura = ANCHURA(x);
  WORD   altura  = ALTURA(x);
  Imagen y       = CreaImagen (anchura,altura);

  for (i=0; i<anchura-1; i++)
      for (j=0; j<altura-1; j++) {
	  dx = (int)PIXEL(x,i,j)-(int)PIXEL(x,i+1,j+1);
	  PIXEL(y,i,j) = (255+dx)/2;
      }

  return y;
}

Imagen GradientRobertsY (Imagen x)
{
  int i,j;
  int dy;

  WORD   anchura = ANCHURA(x);
  WORD   altura  = ALTURA(x);
  Imagen y       = CreaImagen (anchura,altura);

  for (i=0; i<anchura-1; i++)
      for (j=0; j<altura-1; j++) {
	  dy = (int)PIXEL(x,i+1,j)-(int)PIXEL(x,i,j+1);
	  PIXEL(y,i,j) = (255+dy)/2;
  }

  return y;
}

/* ------------------------------------------------------------------------- */
/*  Derivada de la gaussiana                                                 */
/* ------------------------------------------------------------------------- */

#define MAX_MASK 12

/* G(x)=exp((-x^2)/(2*(s^2)))                   */

static float gauss(float x, float sigma)
{
  return (float)exp((double) ((-x*x)/(2*sigma*sigma)));
}

/* G'(x) = -x/(s^2)*exp((-x^2)/(2*(s^2))) = DoG */

static float dGauss (float x, float sigma)
{
  return -x/(sigma*sigma) * gauss(x, sigma);
}

/* Convolucin con la derivada de la gaussiana */

#define DIR_HORIZONTAL 0
#define DIR_VERTICAL   1

static FloatImg ConvolutionDoG ( Imagen img, float *g, int dim, int dir)
{
  int      i,j,k,p1,p2;
  float    x;
  WORD     anchura = ANCHURA(img);
  WORD     altura  = ALTURA(img);
  FloatImg salida  = CreaFloatImg(anchura,altura);

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++) {

	  x = 0.0;                  /*  NB:  G'(-x) = - G'(x) */
                                    /*       G'(0)  = 0       */

	  for (k=1; k<dim; k++) {
	      if (dir == 0) {         // Horizontal
		p1 = (i+k)%anchura;
		p2 = (i-k+anchura)%anchura;
		x += -g[k]*PIXEL(img,p1,j) + g[k]*PIXEL(img,p2,j);
	      } else {                // Vertical
		p1 = (j+k)%altura;
		p2 = (j-k+altura)%altura;
		x += -g[k]*PIXEL(img,i,p1) + g[k]*PIXEL(img,i,p2);
	      }
	  }

	  salida->datos[i][j] = x;
      }

  return salida;
}

/* SQRT(Gx*Gx+Gy*Gy) */

Imagen GradientDoGSQRT (Imagen img, float s, int d)
{
  int      anchura = ANCHURA(img);
  int      altura  = ALTURA(img);
  int      i,j;
  float    dx, dy;
  Imagen   salida;
  FloatImg convX, convY;

  float    dg[MAX_MASK];

  if (d>MAX_MASK)
     d = MAX_MASK;

  // dg[0] = 0 no se usa

  for (i=1; i<d; i++)
      dg[i] = dGauss ((float)i, s);

  convX = ConvolutionDoG (img, dg, d, DIR_HORIZONTAL);
  convY = ConvolutionDoG (img, dg, d, DIR_VERTICAL);

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++) {
          dx = convX->datos[i][j];
          dy = convY->datos[i][j];
          convX->datos[i][j] = sqrt(dx*dx+dy*dy);
  }

  salida = ImagenFloat (convX);

  DestruyeFloatImg(&convX);
  DestruyeFloatImg(&convY);

  return salida;
}

/* |Gx|+|Gy| */

Imagen GradientDoGABS (Imagen img, float s, int d)
{
  int      anchura = ANCHURA(img);
  int      altura  = ALTURA(img);
  int      i,j;
  float    dx, dy;
  Imagen   salida;
  FloatImg convX, convY;

  float    dg[MAX_MASK];

  if (d>MAX_MASK)
     d = MAX_MASK;

  for (i=1; i<d; i++)
      dg[i] = dGauss ((float)i, s);

  convX = ConvolutionDoG (img, dg, d, DIR_HORIZONTAL);
  convY = ConvolutionDoG (img, dg, d, DIR_VERTICAL);

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++) {
          dx = convX->datos[i][j];
          dy = convY->datos[i][j];
          convX->datos[i][j] = abs(dx)+abs(dy);
  }

  salida = ImagenFloat (convX);

  DestruyeFloatImg(&convX);
  DestruyeFloatImg(&convY);

  return salida;
}

/* Gx */

Imagen GradientDoGX (Imagen img, float s, int d)
{
  int      i;
  Imagen   salida;
  FloatImg conv;

  float    dg[MAX_MASK];

  if (d>MAX_MASK)
     d = MAX_MASK;

  for (i=1; i<MAX_MASK; i++)
      dg[i] = dGauss ((float)i, s);

  conv = ConvolutionDoG (img, dg, d, DIR_HORIZONTAL);

  salida = ImagenFloat (conv);

  DestruyeFloatImg(&conv);

  return salida;
}

/* Gy */

Imagen GradientDoGY (Imagen img, float s, int d)
{
  int      i;
  Imagen   salida;
  FloatImg conv;

  float    dg[MAX_MASK];

  if (d>MAX_MASK)
     d = MAX_MASK;

  for (i=1; i<MAX_MASK; i++)
      dg[i] = dGauss ((float)i, s);

  conv = ConvolutionDoG (img, dg, d, DIR_VERTICAL);

  salida = ImagenFloat (conv);

  DestruyeFloatImg(&conv);

  return salida;
}

