/* ----------------------------------------------------------------------- */
/*    Fernando Berzal Galiano                ETS Ingeniera Informtica    */
/* ----------------------------------------------------------------------- */
/*                       Procesamiento de imgenes                         */
/* ----------------------------------------------------------------------- */
/*                          Deteccin de lneas                            */
/* ----------------------------------------------------------------------- */

#include <math.h>

#include "IMG_LINE.H"

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*  r = x cos w + y sin w                                                  */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

FloatImg HoughTransform (Imagen img)
{
  FloatImg Hough;
  int      r, w, i, j;
  int      xc   = ANCHURA(img)/2;
  int      yc   = ALTURA(img)/2;
  int      rmax = (int)sqrt((double)(ANCHURA(img)*ANCHURA(img)+ALTURA(img)*ALTURA(img)))/2.0;
  double   rad[180];

  Hough = CreaFloatImg (180, 2*rmax+1);

  for ( i=0; i<180; i++)
      rad[i] = i*(3.1415926535/180.0);

  for ( i=0; i<ANCHURA(img); i++)
      for ( j=0; j<ALTURA(img); j++)
          if (PIXEL(img,i,j))
	     for (w=0; w<180; w++) {
                 r = (i - xc) * cos(rad[w]) + (j - yc) * sin(rad[w]);
                 Hough->datos[w][rmax+r] ++;
             }
/*

  // Skew (OCR)

  for (i=0; i<180; i++)
      for (j=0; j<2*rmax+1; j++)
          if (Hough->datos[i][j] > tmval) {
             tmval = Hough->datos[i][j];
	     tmax = i;
          }

  skew = tmax;

*/

  return Hough;
}

/* ----------------------------------------------------------------------- */
/*  LineaMAX: Dibuja una lnea segn el algoritmo de Bresenham             */
/*            Slo actualiza un pixel si el nuevo nivel de gris es mayor   */
/* ----------------------------------------------------------------------- */

static void SwapInt ( int *i1, int *i2 )
{
 WORD tmp = *i2;
 *i2 = *i1;
 *i1 = tmp;
}

static void LineaMAX ( Imagen img, int x1, int y1, int x2, int y2, BYTE color)
{
  int d, dx, dy, aincr, bincr, xincr, yincr, x, y;

  if ( abs(x2-x1) < abs(y2-y1) ) {

     if ( y1 > y2 ) {
	SwapInt( &x1, &x2 );
	SwapInt( &y1, &y2 );
     }

     if ( x2 > x1 )
	xincr = 1;
     else
	xincr = -1;

     dy    = y2 - y1;
     dx    = abs( x2-x1 );
     d     = 2 * dx - dy;
     aincr = 2 * (dx - dy);
     bincr = 2 * dx;
     x     = x1;
     y     = y1;


     if ( x>=0 && x<ANCHURA(img) && y>=0 && y<ALTURA(img) )
        if (color>PIXEL(img,x,y))
           PIXEL(img,x,y) = color;

     for (y=y1+1; y<=y2; y++) {
	 if ( d >= 0 ) {
	    x += xincr;
	    d += aincr;
	 } else
	    d += bincr;

         if ( x>=0 && x<ANCHURA(img) && y>=0 && y<ALTURA(img) )
            if (color>PIXEL(img,x,y))
	       PIXEL(img,x,y) = color;
     }

  } else {

     if ( x1 > x2 ) {
	SwapInt( &x1, &x2 );
	SwapInt( &y1, &y2 );
     }

     if ( y2 > y1 )
	yincr = 1;
     else
	yincr = -1;

     dx    = x2 - x1;
     dy    = abs( y2-y1 );
     d     = 2 * dy - dx;
     aincr = 2 * (dy - dx);
     bincr = 2 * dy;
     x     = x1;
     y     = y1;

     if ( x>=0 && x<ANCHURA(img) && y>=0 && y<ALTURA(img) )
        if (color>PIXEL(img,x,y))
           PIXEL(img,x,y) = color;

     for (x=x1+1; x<=x2; x++) {
	 if ( d >= 0 ) {
	    y += yincr;
	    d += aincr;
	 } else
	    d += bincr;

         if ( x>=0 && x<ANCHURA(img) && y>=0 && y<ALTURA(img) )
            if (color>PIXEL(img,x,y))
	       PIXEL (img,x,y) = color;
     }
  }
}

/* ----------------------------------------------------------------------- */

static void LineaHough (Imagen img, int w, int r, int xc, int yc, BYTE color)
{
  int    x1,y1,x2,y2;

  double rad = w * (3.1415926535/180.0);

  if (w>45 && w<135) {
     x1 = 0;
     x2 = ANCHURA(img);
     y1 = (r-(x1-xc)*cos(rad))/sin(rad)+yc;
     y2 = (r-(x2-xc)*cos(rad))/sin(rad)+yc;
  } else {
     y1 = 0;
     y2 = ALTURA(img);
     x1 = (r-(y1-yc)*sin(rad))/cos(rad)+xc;
     x2 = (r-(y2-yc)*sin(rad))/cos(rad)+xc;
  }

  LineaMAX (img, x1, y1, x2, y2, color);
}

/* ----------------------------------------------------------------------- */

Imagen HoughLineDetection (Imagen img)
{
  int      w,r;

  FloatImg Hough;
  Imagen   tmp;
  Imagen   salida  = CreaImagen(ANCHURA(img),ALTURA(img));

  int xc   = ANCHURA(img)/2;
  int yc   = ALTURA(img)/2;
  int rmax = (int)sqrt((double)(ANCHURA(img)*ANCHURA(img)+ALTURA(img)*ALTURA(img)))/2.0;

  Hough = HoughTransform (img);
  tmp   = ImagenFloat (Hough);

  DestruyeFloatImg(&Hough);

  for (w=0; w<180; w++)
      for (r=0; r<ALTURA(tmp); r++)
          if (PIXEL(tmp,w,r))
             LineaHough ( salida, w, r-rmax, xc, yc, PIXEL(tmp,w,r) );

  DestruyeImagen(&tmp);

  return salida;
}

/* ----------------------------------------------------------------------- */

