/* ----------------------------------------------------------------------- */
/*    Fernando Berzal Galiano                ETS Ingeniera Informtica    */
/* ----------------------------------------------------------------------- */
/*                       Procesamiento de imgenes                         */
/* ----------------------------------------------------------------------- */
/*                              Morfologa                                 */
/* ----------------------------------------------------------------------- */

#include "IMG_Morf.H"

/* Comprueba que un pixel queda dentro de la imagen */

int PixelOK (Imagen img, int i, int j)
{
  if ((i<0) || (i>=ANCHURA(img))) return 0;
  if ((j<0) || (j>=ALTURA(img)))  return 0;
  return 1;
}

/* ----------------------------------------------------------------------- */
/*  Structuring elements                                                   */

Imagen SE_3x3_Simple ( BYTE n )  /* vg: n=1 ---> Smooth */
{
  Imagen img = CreaImagen(3,3);

  PIXEL(img,0,0) = n;
  PIXEL(img,0,1) = n;
  PIXEL(img,0,2) = n;
  PIXEL(img,1,0) = n;
  PIXEL(img,1,1) = n;
  PIXEL(img,1,2) = n;
  PIXEL(img,2,0) = n;
  PIXEL(img,2,1) = n;
  PIXEL(img,2,2) = n;

  return img;
}

/* --- */

Imagen SE_3x3 (int t[3][3])
{
  Imagen img = CreaImagen(3,3);

  PIXEL(img,0,0) = t[0][0];
  PIXEL(img,0,1) = t[0][1];
  PIXEL(img,0,2) = t[0][2];
  PIXEL(img,1,0) = t[1][0];
  PIXEL(img,1,1) = t[1][1];
  PIXEL(img,1,2) = t[1][2];
  PIXEL(img,2,0) = t[2][0];
  PIXEL(img,2,1) = t[2][1];
  PIXEL(img,2,2) = t[2][2];

  return img;
}


Imagen SE_3x3_Soft (void)
{
  Imagen img     = CreaImagen(3,3);
  int    t[3][3] = { { 1, 3, 1}, { 3,16, 3}, { 1, 3, 1} };

  PIXEL(img,0,0) = t[0][0];
  PIXEL(img,0,1) = t[0][1];
  PIXEL(img,0,2) = t[0][2];
  PIXEL(img,1,0) = t[1][0];
  PIXEL(img,1,1) = t[1][1];
  PIXEL(img,1,2) = t[1][2];
  PIXEL(img,2,0) = t[2][0];
  PIXEL(img,2,1) = t[2][1];
  PIXEL(img,2,2) = t[2][2];

  return img;
}

Imagen SE_3x3_MexicanHat (void)
{
  Imagen img     = CreaImagen(3,3);
  int    t[3][3] = { { 1, 2, 1}, { 2, 4, 2}, { 1, 2, 1} };

  PIXEL(img,0,0) = t[0][0];
  PIXEL(img,0,1) = t[0][1];
  PIXEL(img,0,2) = t[0][2];
  PIXEL(img,1,0) = t[1][0];
  PIXEL(img,1,1) = t[1][1];
  PIXEL(img,1,2) = t[1][2];
  PIXEL(img,2,0) = t[2][0];
  PIXEL(img,2,1) = t[2][1];
  PIXEL(img,2,2) = t[2][2];

  return img;
}


/* --- */

Imagen SE_Sphere ( BYTE radio )
{
  WORD   x, dx, y, dy;
  WORD   dim = 2*radio+1;
  Imagen img = CreaImagen(dim,dim);

  for (x=0; x<dim; x++)
      for (y=0; y<dim; y++) {

	  if (x<radio)
	     dx = radio - x;
	  else
	     dx = x - radio;

	  if (y<radio)
	     dy = radio - y;
	  else
	     dy = y - radio;

	  PIXEL(img,x,y) = (dx>dy)? radio-dx: radio-dy;
      }

  return img;
}

/* ----------------------------------------------------------------------- */

Imagen Dilation (Imagen a, Imagen se, WORD origenX, WORD origenY)
{
  int    i,j,ii,jj,k,n;

  WORD   anchura   = ANCHURA(a);
  WORD   altura    = ALTURA(a);
  WORD   anchuraSE = ANCHURA(se);
  WORD   alturaSE  = ALTURA(se);

  Imagen b = CreaImagen (anchura, altura);

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++) {

	  k = -10000;

	  for (ii=0; ii<anchuraSE; ii++)
	      for (jj=0; jj<alturaSE; jj++) {

		  if (!PixelOK(a,i+ii-origenX,j+jj-origenY))
		     continue;

		  n = (int)PIXEL(se,ii,jj) + (int)PIXEL(a,i+ii-origenX,j+jj-origenY);

		  if (n>k) k = n;
	      }

	  if (k > 0xFF)
	     PIXEL(b,i,j) = 0xFF;
	  else if (k < 0)
	     PIXEL(b,i,j) = 0x00;
	  else
	     PIXEL(b,i,j) = k;
      }

  return b;
}

/* ----------------------------------------------------------------------- */

Imagen Erosion (Imagen a, Imagen se, WORD origenX, WORD origenY)
{
  int i,j,ii,jj,k,n;

  WORD   anchura   = ANCHURA(a);
  WORD   altura    = ALTURA(a);
  WORD   anchuraSE = ANCHURA(se);
  WORD   alturaSE  = ALTURA(se);

  Imagen b = CreaImagen (anchura, altura);


  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++) {

	  k = 10000;

	  for (ii=0; ii<anchuraSE; ii++)
	      for (jj=0; jj<alturaSE; jj++) {

		  if (!PixelOK(a,i+ii-origenX, j+jj-origenY))
		     continue;

		  n = PIXEL(a,i+ii-origenX,j+jj-origenY) - PIXEL(se,ii,jj);

		  if (n < k) k = n;
	      }

	  if (k > 0xFF)
	     PIXEL(b,i,j) = 0xFF;
	  else if (k<0x00)
	     PIXEL(b,i,j) = 0x00;
	  else
	     PIXEL(b,i,j) = k;
      }

  return b;
}

/* ----------------------------------------------------------------------- */

Imagen Gradient (Imagen a, Imagen se, WORD origenX, WORD origenY)
{
  int    i, j, k;
  WORD   anchura = ANCHURA(a);
  WORD   altura  = ALTURA(a);
  Imagen b = CreaImagen (anchura,altura);
  Imagen d = Dilation (a, se, origenX, origenY);
  Imagen e = Erosion (a, se, origenX, origenY);

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++) {

	  k = ((int)PIXEL(d,i,j) - (int)PIXEL(e,i,j));

	  if (k<0x00)
	     k = 0x00;
	  else if (k>0xFF)
	     k = 0xFF;

	  PIXEL(b,i,j) = k;
      }

  DestruyeImagen(&d);
  DestruyeImagen(&e);

  return b;
}

/* ----------------------------------------------------------------------- */

Imagen Closing (Imagen a, Imagen se, WORD origenX, WORD origenY)
{
  Imagen b = Dilation (a, se, origenX, origenY);
  Imagen c = Erosion (b, se, origenX, origenY);

  DestruyeImagen(&b);

  return c;
}
/* ----------------------------------------------------------------------- */

Imagen Opening (Imagen a, Imagen se, WORD origenX, WORD origenY)
{
  Imagen b = Erosion  (a, se, origenX, origenY);
  Imagen c = Dilation (b, se, origenX, origenY);

  DestruyeImagen(&b);

  return c;
}

/* ----------------------------------------------------------------------- */

Imagen MorphologicalSmooth (Imagen a, Imagen se, WORD origenX, WORD origenY)
{
  Imagen b = Opening (a, se, origenX, origenY);
  Imagen c = Closing (b, se, origenX, origenY);

  DestruyeImagen(&b);

  return c;
}

/* ----------------------------------------------------------------------- */

Imagen InteriorOutlining (Imagen a, Imagen se, WORD origenX, WORD origenY)
{
  int    i, j, k;
  WORD   anchura = ANCHURA(a);
  WORD   altura  = ALTURA(a);
  Imagen b = CreaImagen (anchura,altura);
  Imagen e = Erosion (a, se, origenX, origenY);

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++) {

	  k = ((int)PIXEL(a,i,j) - (int)PIXEL(e,i,j));

	  if (k<0x00)
	     k = 0x00;
	  else if (k>0xFF)
	     k = 0xFF;

	  PIXEL(b,i,j) = k;
      }

  DestruyeImagen(&e);

  return b;
}

/* ----------------------------------------------------------------------- */

Imagen ExteriorOutlining (Imagen a, Imagen se, WORD origenX, WORD origenY)
{
  int    i, j, k;
  WORD   anchura = ANCHURA(a);
  WORD   altura  = ALTURA(a);
  Imagen b = CreaImagen (anchura,altura);
  Imagen d = Dilation (a, se, origenX, origenY);

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++) {

	  k = ((int)PIXEL(d,i,j) - (int)PIXEL(a,i,j));

	  if (k<0x00)
	     k = 0x00;
	  else if (k>0xFF)
	     k = 0xFF;

	  PIXEL(b,i,j) = k;
      }

  DestruyeImagen(&d);

  return b;
}

/* ----------------------------------------------------------------------- */
/*  Ejemplo

#include "FB_IMAG.C"
#include "FBG_CRT.H"
#include "FBG_KBD.H"
#include "FBG_GRAF.C"

#define RADIO 3

void main (void)
{
  Imagen img, se, i2, i3, i4;
  int    x;

  img = CargaImagen("ajedrez.grx");
  se  = SE_Sphere(RADIO);

  ModoSVGA();
  PaletaGris();

  if (img) MuestraImagen (img, 1, 1 );

  i2 = Dilation(img,se,RADIO,RADIO);

  if (i2) MuestraImagen (i2, 401, 1);

  DestruyeImagen (&i2);

  i3 = MorphologicalSmooth(img,se,RADIO,RADIO);

  if (i3) MuestraImagen (i3, 1, 301);

  DestruyeImagen (&i3);

  i4 = Erosion(img,se,RADIO,RADIO);

  if (i4) MuestraImagen (i4, 401, 301);

  DestruyeImagen (&i4);

  GetChar();

  DestruyeImagen (&img);

  ModoTexto();
}

/**/
