/* ----------------------------------------------------------------------- */
/*    Fernando Berzal Galiano                ETS Ingeniera Informtica    */
/* ----------------------------------------------------------------------- */
/*                       Procesamiento de imgenes                         */
/* ----------------------------------------------------------------------- */

#include <math.h>

#include "IMG_Proc.H"

#include "IMG_TYPE.H"

#define MIN(a,b) (((a)<(b))?(a):(b))

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

BYTE Rango (Imagen img, BYTE *min, BYTE *max)
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  WORD   x,y;

  *max = 0;
  *min = 0xFF;

  for (x=0; x<anchura; x++)
      for (y=0; y<altura; y++) {

	  if (PIXEL(img,x,y)<(*min))
	     *min = PIXEL(img,x,y);

	  if (PIXEL(img,x,y)>(*max))
	     *max = PIXEL(img,x,y);
      }

  return (*max)-(*min);
}

/* --- */

BYTE MaxPixel (Imagen img)
{
  WORD   anchura  = ANCHURA(img);
  WORD   altura   = ALTURA(img);
  WORD   x,y;
  BYTE   max = 0;

  for (x=0; x<anchura; x++)
      for (y=0; y<altura; y++)
	  if (PIXEL(img,x,y)>max) max=PIXEL(img,x,y);

  return max;
}

/* --- */

BYTE MinPixel (Imagen img)
{
  WORD   anchura  = ANCHURA(img);
  WORD   altura   = ALTURA(img);
  WORD   x,y;
  BYTE   min = 0xFF;

  for (x=0; x<anchura; x++)
      for (y=0; y<altura; y++)
	  if (PIXEL(img,x,y)<min) min=PIXEL(img,x,y);

  return min;
}

/* ----------------------------------------------------------------------- */
/*  Negativar una imagen (GRAYLEVEL)                                       */
/* ----------------------------------------------------------------------- */

Imagen Negativar (Imagen img)
{
  WORD   anchura  = ANCHURA(img);
  WORD   altura   = ALTURA(img);
  WORD   x,y;
  Imagen negativo = CreaImagen(anchura,altura);

  if (negativo)
     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL (negativo,x,y) = 0xFF-PIXEL(img,x,y);

  return negativo;
}

/* ----------------------------------------------------------------------- */
/*  Aclarar una imagen (GRAYLEVEL)                                         */
/* ----------------------------------------------------------------------- */

Imagen Aclarar (Imagen img, BYTE n)
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  WORD   x,y;
  Imagen aclarada = CreaImagen(anchura,altura);

  if (aclarada)
     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(aclarada,x,y) = ( n >= (0xFF-PIXEL(img,x,y)) ) ?
				   0xFF : PIXEL(img,x,y) + n;

  return aclarada;
}

/* ----------------------------------------------------------------------- */
/*  Oscurecer una imagen (GRAYLEVEL)                                       */
/* ----------------------------------------------------------------------- */

Imagen Oscurecer (Imagen img, BYTE n)
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  WORD   x,y;
  Imagen oscurecida = CreaImagen (anchura,altura);

  if (oscurecida)
     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(oscurecida,x,y) = ( PIXEL(img,x,y) > n ) ?
				     PIXEL(img,x,y) - n : 0x00;

  return oscurecida;
}

/* ----------------------------------------------------------------------- */
/*  Contrastar una imagen (GRAYLEVEL)                                      */
/* ----------------------------------------------------------------------- */

Imagen Contrastar (Imagen img)
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  WORD   x,y,i;
  Imagen contrast;
  BYTE   min,max,rango;
  BYTE   val[256];

  rango = Rango (img, &min, &max);

  if (max!=min && rango<0xFF) {

     contrast = CreaImagen (anchura,altura);

     if (contrast) {

	for (i=min; i<=max; i++)
	    val[i] = ((float)(i-min)/(float)rango)*255.0;

	for (x=0; x<anchura; x++)
	    for (y=0; y<altura; y++)
		PIXEL(contrast,x,y) = val[PIXEL(img,x,y)];
     }

     return contrast;

  } else {

     return img;

  }
}

/* ----------------------------------------------------------------------- */
/*  Ajuste del contraste: STRETCHING                                       */
/* ----------------------------------------------------------------------- */
/*  Polilnea: (0,0)(x,y)(u,v)(255,255)                                    */

Imagen Stretch (Imagen img, BYTE x, BYTE y, BYTE u, BYTE v)
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  WORD   i,j;
  Imagen contrast = CreaImagen (anchura,altura);
  BYTE   val[256];

  if (contrast) {

     for (i=0; i<x; i++)
         val[i] = ((float)i/(float)x)*y;

     val[x] = y;

     for (i=x+1; i<u; i++)
         val[i] = y + ((float)(i-x)/(float)(u-x))*(v-y);

     val[u] = v;

     for (i=u+1; i<256; i++)
         val[i] = v + ((float)(i-u)/(float)(255-u))*(255-v);

     for (i=0; i<anchura; i++)
         for (j=0; j<altura; j++)
             PIXEL(contrast,i,j) = val[PIXEL(img,i,j)];
  }

  return contrast;
}

/* ----------------------------------------------------------------------- */
/*  Ecualizacin del histograma (GRAYLEVEL)                                */
/* ----------------------------------------------------------------------- */

Imagen Equalization ( Imagen img )
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  float  pixels  = (float)((long)anchura*(long)altura)/256.0;
  WORD   x,y;

  BYTE   gris[256];
  QUAD   nivel[256];

  Imagen salida  = CreaImagen (anchura,altura);


  if (salida) {

     for (x=0; x<256; x++)
	 nivel[x] = 0;

     for (x=0; x<anchura; x++)                               /* Histograma */
	 for (y=0; y<altura; y++)
	     nivel[PIXEL(img,x,y)]++;

     for (x=1; x<256; x++)                         /* Histograma acumulado */
	 nivel[x] += nivel[x-1];


     for (x=0; x<256; x++)        /* Transformacin de la escala de grises */
	 gris[x] = MIN(nivel[x]/pixels,255);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(salida,x,y) = gris[PIXEL(img,x,y)];
  }

  return salida;
}

/* ----------------------------------------------------------------------- */
/*  Logaritmo                                                              */
/* ----------------------------------------------------------------------- */

Imagen LogImagen (Imagen img)
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  WORD   i,x,y;
  BYTE   nivel[256];
  Imagen logIMG  = CreaImagen(anchura,altura);

  if (logIMG) {

     for (i=0; i<256; i++)
         nivel[i] = 255.0 * log ( (exp(1)-1)*i/255.0 + 1 );

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL(logIMG,x,y) = nivel[PIXEL(img,x,y)];
  }

  return logIMG;
}

/* ----------------------------------------------------------------------- */
/*  Exponencial                                                            */
/* ----------------------------------------------------------------------- */

Imagen ExpImagen (Imagen img)
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  WORD   i,x,y;
  BYTE   nivel[256];
  Imagen expIMG  = CreaImagen(anchura,altura);

  if (expIMG) {

     for (i=0; i<256; i++)
         nivel[i] = 255.0/(exp(1)-1) * ( exp(i/255.0) - 1 );

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     PIXEL (expIMG,x,y) = nivel[PIXEL(img,x,y)];
  }

  return expIMG;
}

/* ----------------------------------------------------------------------- */
/*  GDA: Gradation Removal by averaging (GRAYLEVEL)                        */
/* ----------------------------------------------------------------------- */

Imagen GDA (Imagen img, BYTE X_rec, BYTE Y_rec)
{
  WORD anchura = ANCHURA(img);
  WORD altura  = ALTURA(img);

  QUAD  Rec[64][64];
  QUAD  pixels[64][64];
  float media;

  float XPrec, YPrec;

  BYTE color;
  WORD x,y,i,j;
  int  dif;

  Imagen gda = CreaImagen (anchura,altura);

  if (gda) {

     if (X_rec>64) X_rec=64;
     if (Y_rec>64) Y_rec=64;

     XPrec = (float)anchura/(float)X_rec;
     YPrec = (float)altura/(float)Y_rec;

     for (x=0; x<X_rec; x++)
	 for (y=0; y<Y_rec; y++) {
	     Rec[x][y] = 0;
             pixels[x][y] = 0;
         }

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++) {
             i = (int)(x/XPrec);
             j = (int)(y/YPrec);
	     Rec [i][j] += PIXEL(img,x,y);
             pixels[i][j]++;
         }

     media = 0;

     for (x=0; x<X_rec; x++)
	 for (y=0; y<Y_rec; y++) {
	     Rec[x][y] /= pixels[x][y];
	     media += Rec[x][y]*pixels[x][y];
	 }

     media /= (anchura*altura);

     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++) {
	     color = PIXEL(img,x,y);
	     dif   = media - Rec[(int)(x/XPrec)][(int)(y/YPrec)];

	     if (dif>0)
		color = (dif>0xFF-color)? 0xFF: color+dif;
	     else
		color = (-dif>color)? 0: color+dif;

	     PIXEL(gda,x,y) = color;
	 }
  }

  return gda;
}

/* ----------------------------------------------------------------------- */
/*  Median Filter                                                          */
/* ----------------------------------------------------------------------- */

BYTE Mediana ( BYTE colores[9] )
{
  BYTE i,j,tmp;

  for (i=0; i<9; i++)
      for (j=i+1; j<9; j++)
	  if (colores[i]>colores[j]) {
	     tmp        = colores[i];
	     colores[i] = colores[j];
	     colores[j] = tmp;
	  }

  return colores[4];
}

/* --- */

Imagen MedianFilter ( Imagen img )
{
  WORD anchura = ANCHURA(img) - 2;
  WORD altura  = ALTURA(img)  - 2;
  WORD x,y;
  BYTE colores[9];

  Imagen filt = CreaImagen (anchura, altura);

  if (filt)
     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++) {
	     colores[0] = PIXEL(img,x  ,y  );
	     colores[1] = PIXEL(img,x+1,y  );
	     colores[2] = PIXEL(img,x+2,y  );
	     colores[3] = PIXEL(img,x  ,y+1);
	     colores[4] = PIXEL(img,x+1,y+1);
	     colores[5] = PIXEL(img,x+2,y+1);
	     colores[6] = PIXEL(img,x  ,y+2);
	     colores[7] = PIXEL(img,x+1,y+2);
	     colores[8] = PIXEL(img,x+2,y+2);
	     PIXEL(filt,x,y) = Mediana(colores);
	 }

  return filt;
}

/* ----------------------------------------------------------------------- */
/*  Modal Filter                                                           */
/* ----------------------------------------------------------------------- */

BYTE Moda ( BYTE color[9] )
{
  BYTE f[9];
  BYTE c[9];
  BYTE moda;
  BYTE i,j;
  WORD suma;

  /* Frecuencias relativas */

  for (i=0; i<9; i++) {
      f[i] = 1;
      for (j=i+1; j<9; j++)
	  if (color[i]==color[j])
	     f[i] ++;
  }

  /* Moda */

  c[0] = 0;
  moda = 1;

  for (i=1; i<9; i++) {
      if (f[i]>f[c[0]]) {
	 c[0] = i;
	 moda = 1;
      } else if (f[i]==f[c[0]]) {
	 c[moda] = i;
	 moda ++ ;
      }
  }

  /* Media de las modas */

  suma = color[c[0]];

  for (i=1; i<moda; i++)
      suma += color[c[i]];

  return suma/moda;
}

/* --- */

Imagen ModalFilter ( Imagen img )
{
  WORD anchura = ANCHURA(img) - 2;
  WORD altura  = ALTURA(img)  - 2;
  WORD x,y;
  BYTE colores[9];

  Imagen filt = CreaImagen (anchura, altura);

  if (filt)
     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++) {
	     colores[0] = PIXEL(img,x  ,y  );
	     colores[1] = PIXEL(img,x+1,y  );
	     colores[2] = PIXEL(img,x+2,y  );
	     colores[3] = PIXEL(img,x  ,y+1);
	     colores[4] = PIXEL(img,x+1,y+1);
	     colores[5] = PIXEL(img,x+2,y+1);
	     colores[6] = PIXEL(img,x  ,y+2);
	     colores[7] = PIXEL(img,x+1,y+2);
	     colores[8] = PIXEL(img,x+2,y+2);
	     PIXEL(filt,x,y) = Moda(colores);
	 }

  return filt;
}

/* ----------------------------------------------------------------------- */
/*  Averaging Filter                                                       */
/* ----------------------------------------------------------------------- */

int Media ( Imagen img, WORD x, WORD y )
{
  int i,j;
  int sum = 0;

  for (i=x-1; i<=x+1; i++)
      for (j=y-1; j<=y+1; j++) 
	  sum += PIXEL(img,i,j);

  return (sum/9);
}

/* --- */

Imagen AveragingFilter ( Imagen img )
{
  WORD anchura = ANCHURA(img) - 2;
  WORD altura  = ALTURA(img)  - 2;
  WORD x,y;

  Imagen filt = CreaImagen (anchura, altura);

  if (filt)
     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++) 
	     PIXEL(filt,x,y) = Media(img,x+1,y+1);

  return filt;
}

/* ----------------------------------------------------------------------- */
/*  Closest Averaging                                                      */
/* ----------------------------------------------------------------------- */

BYTE Media7 ( BYTE colores[9] )
{
  BYTE i,j,tmp;
  WORD sum;

  for (i=0; i<9; i++)                    /* Ordenacin */
      for (j=i+1; j<9; j++)
	  if (colores[i]>colores[j]) {
	     tmp        = colores[i];
	     colores[i] = colores[j];
	     colores[j] = tmp;
	  }

  sum = colores[1];                      /* Media de los 7 centrales */

  for (i=2; i<8; i++)
      sum += colores[i];

  return (sum/7);
}

/* --- */

Imagen ClosestAveragingFilter ( Imagen img )
{
  WORD anchura = ANCHURA(img) - 2;
  WORD altura  = ALTURA(img)  - 2;
  WORD x,y;
  BYTE colores[9];

  Imagen filt = CreaImagen (anchura, altura);

  if (filt)
     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++) {
	     colores[0] = PIXEL(img,x  ,y  );
	     colores[1] = PIXEL(img,x+1,y  );
	     colores[2] = PIXEL(img,x+2,y  );
	     colores[3] = PIXEL(img,x  ,y+1);
	     colores[4] = PIXEL(img,x+1,y+1);
	     colores[5] = PIXEL(img,x+2,y+1);
	     colores[6] = PIXEL(img,x  ,y+2);
	     colores[7] = PIXEL(img,x+1,y+2);
	     colores[8] = PIXEL(img,x+2,y+2);
	     PIXEL(filt,x,y) = Media7(colores);
	 }

  return filt;
}

/* ----------------------------------------------------------------------- */
/*  Gaussiana                                                              */
/* ----------------------------------------------------------------------- */
/*  s: Desviacin estndar de la gaussiana                                 */
/*  d: Tamao del filtro ---> Ventana (2d+1)*(2d+1)                        */

#define MAX_GAUSS 12

static float gauss(float x, float sigma)
{
  return (float)exp((double) ((-x*x)/(2*sigma*sigma)));
}

Imagen GaussianFilter ( Imagen img, float s, int d)
{
  WORD     anchura = ANCHURA(img);
  WORD     altura  = ALTURA(img);
  int      i,j,k;
  int      p1,p2;
  float    c;
  float    g[MAX_GAUSS];
  Imagen   salida;
  FloatImg convX, convY;

  if (d>MAX_GAUSS)
     d = MAX_GAUSS;

  for (i=0; i<d; i++)
      g[i] = gauss ((float)i, s);

  convX = CreaFloatImg(anchura,altura);
  convY = CreaFloatImg(anchura,altura);

  /* X */

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++) {

	  c = g[0] * PIXEL(img,i,j);

	  for (k=1; k<d; k++) {                  /* G(-x)=G(x) */
	      p1 = (i+k)%anchura;
	      p2 = (i-k+anchura)%anchura;
	      c += g[k]*PIXEL(img,p1,j) + g[k]*PIXEL(img,p2,j);
	  }

	  convX->datos[i][j] = c;
      }

  /* Y */

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++) {

	  c = g[0] * convX->datos[i][j];

	  for (k=1; k<d; k++) {                  /* G(-x)=G(x) */
	      p1 = (j+k)%altura;
	      p2 = (j-k+altura)%altura;
	      c += g[k]*(convX->datos[i][p1]) + g[k]*(convX->datos[i][p2]);
	  }

	  convY->datos[i][j] = c;
      }

  /* Salida */

  salida = ImagenFloat (convY);

  DestruyeFloatImg(&convX);
  DestruyeFloatImg(&convY);

  return salida;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

Imagen Suavizado ( Imagen origen, float k)
{
  WORD   anchura = ANCHURA(origen);
  WORD   altura  = ALTURA(origen);
  WORD   x,y;
  QUAD   suma = 0;
  BYTE   media, color;

  Imagen destino = CreaImagen(anchura,altura);

  for (x=0; x<anchura; x++)
      for (y=0; y<altura; y++)
	  suma += PIXEL(origen,x,y);

  media = suma / ((QUAD)anchura*(QUAD)altura);

  for (x=0; x<anchura; x++)
      for (y=0; y<altura; y++) {
	  color = PIXEL(origen,x,y);

	  if (color>media)
	     PIXEL(destino,x,y)=color-(BYTE)((color-media)*k);
	  else if (color<media)
	     PIXEL(destino,x,y)=color+(BYTE)((media-color)*k);
	  else
	     PIXEL(destino,x,y)=color;
      }

  return destino;
}

/* ----------------------------------------------------------------------- */

Imagen UnsharpMask (Imagen origen, float s, float k)
{
  WORD   anchura = ANCHURA (origen);
  WORD   altura  = ALTURA  (origen);
  Imagen soft    = Suavizado ( origen, s );
  Imagen destino = CreaImagen(anchura,altura);

  WORD   x,y;
  int    co,cm,cd;

  for (x=0; x<anchura; x++)
      for (y=0; y<altura; y++) {
	  co = PIXEL(origen,x,y);
	  cm = PIXEL(soft,x,y);
	  cd = co+k*(co-cm);

	  if (cd>=0x00 && cd<=0xFF)
	     PIXEL(destino,x,y) = cd;
	  else if (cd<0)
	     PIXEL(destino,x,y) = 0x00;
	  else
	     PIXEL(destino,x,y) = 0xFF ;
      }

  DestruyeImagen(&soft);

  return destino;
}

/* ----------------------------------------------------------------------- */

Imagen AfiladoLaplaciana ( Imagen origen, float k )
{
  WORD   anchura = ANCHURA(origen);
  WORD   altura  = ALTURA(origen);
  WORD   x, y;
  int    color;
  int    Laplacian  [3][3] = { {-2, 1,-2}, { 1, 4, 1}, {-2, 1,-2} };
  Imagen laplac  = Convolution ( origen, Laplacian, 1 );
  Imagen destino = CreaImagen (anchura, altura);
 
  for (x=1; x<anchura-1; x++)
      for (y=1; y<altura-1; y++) {
	 color = PIXEL(origen,x,y) + k*PIXEL(laplac,x-1,y-1);

	 if (color>0x00 && color<0xFF)
	    PIXEL(destino,x,y) = color;
	 else if ( color<=0x00 )
	    PIXEL(destino,x,y) = 0x00;
	 else
	    PIXEL(destino,x,y) = 0xFF;
      }

  DestruyeImagen (&laplac);

  return destino;
}

/* ----------------------------------------------------------------------- */

Imagen AfiladoDV ( Imagen origen, BYTE w, float k )                 /* AGC */
{
  WORD   anchura = ANCHURA(origen);
  WORD   altura  = ALTURA(origen);
  Imagen destino = CreaImagen(anchura,altura);
  WORD   x,y,n,m;
  BYTE   W = (2*w+1)*(2*w+1);
  float  media,dv;

  for (x=w; x<anchura-w; x++)
      for (y=w; y<altura-w; y++) {

	  media = 0;

	  for (n=x-w; n<=x+w; n++)
	      for (m=y-w; m<=y+w; m++)
		  media += PIXEL(origen,n,m);

	  media /= W;
	  dv     = 0;

	  for (n=x-w; n<=x+w; n++)
	      for (m=y-w; m<=y+w; m++)
		  dv += (((float)PIXEL(origen,n,m)-media)*((float)PIXEL(origen,n,m)-media));

	  dv /= W;
	  dv *= k;
          dv += PIXEL(origen,x,y);

	  if (dv>=0 && dv<256)
	     PIXEL (destino,x,y) = dv;
	  else if (dv<0)
	     PIXEL (destino,x,y) = 0x00;
	  else
	     PIXEL (destino,x,y) = 0xFF;
  }

  return destino;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

Imagen GradienteSqrt (Imagen img)
{
  WORD   anchura = ANCHURA(img)-1;
  WORD   altura  = ALTURA(img)-1;
  Imagen destino = CreaImagen(anchura,altura);
  WORD   x,y;
  int    a,b;
  float  r;

  for (x=0; x<anchura; x++)
      for (y=0; y<altura; y++) {
	  a = (int)PIXEL(img,x,y)-(int)PIXEL(img,x+1,y);
	  b = (int)PIXEL(img,x,y)-(int)PIXEL(img,x,y+1);
	  r = a*a+b*b;

	  if (r>0)
	     PIXEL(destino,x,y) = (BYTE)(sqrt(r)/M_SQRT2);
	  else
	     PIXEL(destino,x,y) = 0;
      }

  return destino;    
}


Imagen GradienteAbs (Imagen img)
{
  WORD   anchura = ANCHURA(img)-1;
  WORD   altura  = ALTURA(img)-1;
  Imagen destino = CreaImagen(anchura,altura);
  WORD   x,y;
  int    a,b;

  for (x=0; x<anchura; x++)
      for (y=0; y<altura; y++) {
	  a = (int)PIXEL(img,x,y)-(int)PIXEL(img,x+1,y);
	  if (a<0) a=-a;
	  b = (int)PIXEL(img,x,y)-(int)PIXEL(img,x,y+1);
	  if (b<0) b=-b;
	  PIXEL(destino,x,y) = (BYTE)((a+b)/2);
      }

  return destino;    
}

Imagen GradienteRobertsSqrt (Imagen img)
{
  WORD   anchura = ANCHURA(img)-1;
  WORD   altura  = ALTURA(img)-1;
  Imagen destino = CreaImagen(anchura,altura);
  WORD   x,y;
  int    a,b;
  float  r;

  for (x=0; x<anchura; x++)
      for (y=0; y<altura; y++) {
	  a = (int)PIXEL(img,x,y)-(int)PIXEL(img,x+1,y+1);
	  b = (int)PIXEL(img,x+1,y)-(int)PIXEL(img,x,y+1);
	  r = a*a+b*b;

	  if (r>0)
	     PIXEL(destino,x,y) = (BYTE)(sqrt(r)/M_SQRT2);
	  else
	     PIXEL(destino,x,y) = 0;
      }

  return destino;    
}


Imagen GradienteRobertsAbs (Imagen img)
{
  WORD   anchura = ANCHURA(img)-1;
  WORD   altura  = ALTURA(img)-1;
  Imagen destino = CreaImagen(anchura,altura);
  WORD   x,y;
  int    a,b;

  for (x=0; x<anchura; x++)
      for (y=0; y<altura; y++) {
	  a = (int)PIXEL(img,x,y)-(int)PIXEL(img,x+1,y+1);
	  if (a<0) a=-a;
	  b = (int)PIXEL(img,x+1,y)-(int)PIXEL(img,x,y+1);
	  if (b<0) b=-b;
	  PIXEL(destino,x,y) = (BYTE)((a+b)/2);
      }

  return destino;    
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

Imagen SuavizadoPseudoAVG (Imagen img)
{
  int t[3][3] = { { 1, 1, 1}, { 1, 2, 1}, { 1, 1, 1} };

  return Convolution(img,t,10);
}

Imagen SuavizadoMexicanHat (Imagen img)
{
  int t[3][3] = { { 1, 2, 1}, { 2, 4, 2}, { 1, 2, 1} };

  return Convolution(img,t,16); 
}

Imagen SuavizadoSOFT (Imagen img)
{
  int t[3][3] = { { 1, 3, 1}, { 3,16, 3}, { 1, 3, 1} };

  return Convolution(img,t,32);
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

Imagen Realzado4entorno (Imagen img)
{
  int t[3][3] = { { 0,-1, 0}, {-1, 5,-1}, { 0,-1, 0} };

  return Convolution(img,t,1);
}

Imagen Realzado8entorno (Imagen img)
{
  int t[3][3] = { {-1,-1,-1}, {-1, 9,-1}, {-1,-1,-1} };

  return Convolution(img,t,1);
}

Imagen RealzadoDistEuclid (Imagen img)
{
  // int t[3][3] = { { 1,-2, 1}, {-2, 5,-2}, { 1,-2, 1} };

  int t[3][3] = { {-2, 1,-2}, { 1, 5, 1}, {-2, 1,-2} };

  return Convolution(img,t,1);
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
