/* ----------------------------------------------------------------------- */
/*    Fernando Berzal Galiano                ETS Ingeniera Informtica    */
/* ----------------------------------------------------------------------- */
/*                       Procesamiento de imgenes                         */
/* ----------------------------------------------------------------------- */

#include <stdlib.h>
#include <math.h>

#include "IMG_RGB.H"

/* ----------------------------------------------------------------------- */
/*  Convierte una imagen a niveles de gris eliminando su paleta            */
/* ----------------------------------------------------------------------- */

void ImagenGRAYLEVEL (Imagen img)
{
  // Criterio: |(r,g,b)| = |(gris,gris,gris)| --> gris = sqrt((r*r+g*g+b*b)/3)

  BYTE  nivel[256];
  int   c,x,y;
  float r,g,b;

  if (img->Paleta) {

     for (c=0; c<256; c++) {
	 r = PALETA(img,c).R;
	 g = PALETA(img,c).G;
	 b = PALETA(img,c).B;
	 nivel[c] = (BYTE) sqrt ((r*r+g*g+b*b)/3.0);
     }


     for (x=0; x<ANCHURA(img); x++)
	 for (y=0; y<ALTURA(img); y++)
	     PIXEL(img,x,y) = nivel[PIXEL(img,x,y)];

     free (img->Paleta);
     img->Paleta = NULL;
  }
}

/* --- */

void ImagenGRAYLEVELMedia (Imagen img)
{
  // Criterio: gris = (r+g+b)/3

  BYTE nivel[256];
  int  c,x,y;

  if (img->Paleta) {

     for (c=0; c<256; c++)
	 nivel[c] = ( (PALETA(img,c).R)
		    + (PALETA(img,c).G)
		    + (PALETA(img,c).B) ) / 3;

     for (x=0; x<ANCHURA(img); x++)
	 for (y=0; y<ALTURA(img); y++)
	     PIXEL(img,x,y) = nivel[PIXEL(img,x,y)];

     free (img->Paleta);
     img->Paleta = NULL;
  }
}

/* --- */

void ImagenGRAYLEVELMax (Imagen img)
{
  // Criterio: gris = Mx { R, G, B }

  BYTE nivel[256];
  int  c,x,y;
  int  max;

  if (img->Paleta) {

     for (c=0; c<256; c++) {
	 max = PALETA(img,c).R;
	 if (PALETA(img,c).G>max) max = PALETA(img,c).G;
	 if (PALETA(img,c).B>max) max = PALETA(img,c).B;
	 nivel[c] = max;
     }

     for (x=0; x<ANCHURA(img); x++)
	 for (y=0; y<ALTURA(img); y++)
	     PIXEL(img,x,y) = nivel[PIXEL(img,x,y)];

     free (img->Paleta);
     img->Paleta = NULL;
  }
}

/* --- */

void ImagenGRAYLEVELMin (Imagen img)
{
  // Criterio: gris = Min { R, G, B }

  BYTE nivel[256];
  int  c,x,y;
  int  min;

  if (img->Paleta) {

     for (c=0; c<256; c++) {
	 min = PALETA(img,c).R;
	 if (PALETA(img,c).G<min) min = PALETA(img,c).G;
	 if (PALETA(img,c).B<min) min = PALETA(img,c).B;
	 nivel[c] = min;
     }

     for (x=0; x<ANCHURA(img); x++)
	 for (y=0; y<ALTURA(img); y++)
	     PIXEL(img,x,y) = nivel[PIXEL(img,x,y)];

     free (img->Paleta);
     img->Paleta = NULL;
  }
}

/* --- */

void ImagenLuminanciaCIE (Imagen img)
{
  // Criterio: Y = 0.177*R + 0.813*G + 0.011*B (CIE)

  BYTE nivel[256];
  int  c,x,y;

  if (img->Paleta) {

     for (c=0; c<256; c++) 
	 nivel[c] = 0.177*PALETA(img,c).R
		  + 0.813*PALETA(img,c).G
		  + 0.011*PALETA(img,c).B;

     for (x=0; x<ANCHURA(img); x++)
	 for (y=0; y<ALTURA(img); y++)
	     PIXEL(img,x,y) = nivel[PIXEL(img,x,y)];

     free (img->Paleta);
     img->Paleta = NULL;
  }
}

/* --- */

void ImagenLuminanciaNTSC (Imagen img)
{
  // Criterio: Y = 0.299*R + 0.587*G + 0.114*B (CIE)

  BYTE nivel[256];
  int  c,x,y;

  if (img->Paleta) {

     for (c=0; c<256; c++) 
	 nivel[c] = 0.299*PALETA(img,c).R
		  + 0.587*PALETA(img,c).G
		  + 0.114*PALETA(img,c).B;

     for (x=0; x<ANCHURA(img); x++)
	 for (y=0; y<ALTURA(img); y++)
	     PIXEL(img,x,y) = nivel[PIXEL(img,x,y)];

     free (img->Paleta);
     img->Paleta = NULL;
  }
}

/* ----------------------------------------------------------------------- */
/*  Obtencin a partir de una imagen multibanda de un plano determinado    */
/* ----------------------------------------------------------------------- */

void ImagenBandaR (Imagen img)
{
  BYTE  nivel[256];
  int   c,x,y;

  if (img->Paleta) {

     for (c=0; c<256; c++) 
	 nivel[c] = PALETA(img,c).R;


     for (x=0; x<ANCHURA(img); x++)
	 for (y=0; y<ALTURA(img); y++)
	     PIXEL(img,x,y) = nivel[PIXEL(img,x,y)];

     free (img->Paleta);
     img->Paleta = NULL;
  }
}

/* --- */

void ImagenBandaG (Imagen img)
{
  BYTE  nivel[256];
  int   c,x,y;

  if (img->Paleta) {

     for (c=0; c<256; c++) 
	 nivel[c] = PALETA(img,c).G;


     for (x=0; x<ANCHURA(img); x++)
	 for (y=0; y<ALTURA(img); y++)
	     PIXEL(img,x,y) = nivel[PIXEL(img,x,y)];

     free (img->Paleta);
     img->Paleta = NULL;
  }
}

/* --- */

void ImagenBandaB (Imagen img)
{
  BYTE  nivel[256];
  int   c,x,y;

  if (img->Paleta) {

     for (c=0; c<256; c++) 
	 nivel[c] = PALETA(img,c).B;


     for (x=0; x<ANCHURA(img); x++)
	 for (y=0; y<ALTURA(img); y++)
	     PIXEL(img,x,y) = nivel[PIXEL(img,x,y)];

     free (img->Paleta);
     img->Paleta = NULL;
  }
}


