/* ----------------------------------------------------------------------- */
/*    Fernando Berzal Galiano                ETS Ingeniera Informtica    */
/* ----------------------------------------------------------------------- */
/*                       Procesamiento de imgenes                         */
/* ----------------------------------------------------------------------- */
/*                             Segmentacin                                */
/* ----------------------------------------------------------------------- */

#include "IMG.H"


/* ----------------------------------------------------------------------- */
/*  Thresholding (GRAYLEVEL)                                               */
/* ----------------------------------------------------------------------- */
/*  umbral, min, max: 0 .. 255                                             */

Imagen Thresholding ( Imagen img, BYTE umbral, BYTE min, BYTE max );

/* ----------------------------------------------------------------------- */
/*  Hysteresis Thresholding                                                */
/* ----------------------------------------------------------------------- */
/*  high, low: 0 .. 255                                                    */

Imagen Hysteresis ( Imagen mag, BYTE high, BYTE low);

/* ----------------------------------------------------------------------- */
/*  Mean Grey Level Thresholding                                           */
/* ----------------------------------------------------------------------- */

Imagen MeanThresholding (Imagen img);

/* ----------------------------------------------------------------------- */
/*  Black Percentage Thresholding (p-tile method)                          */
/* ----------------------------------------------------------------------- */
/*  pct: Porcentaje de pxeles negros (%)                                  */

Imagen BlackPercentageThresholding (Imagen img, float pct);

/* ----------------------------------------------------------------------- */
/*  Two Peaks Thresholding                                                 */
/* ----------------------------------------------------------------------- */

Imagen TwoPeaksThresholding (Imagen img);

/* ----------------------------------------------------------------------- */
/*  Laplacian Thresholding (using edge pixels)                             */
/* ----------------------------------------------------------------------- */
/*  PCT (%): 85.0                                                          */

Imagen LaplacianThresholding (Imagen img, float PCT);

/* ----------------------------------------------------------------------- */
/*  Iterative Selection Thresholding                                       */
/* ----------------------------------------------------------------------- */

Imagen IterativeSelectionThresholding (Imagen img);

Imagen FastIterSelectionThresholding  (Imagen img);

/* ----------------------------------------------------------------------- */
/*  Otsu Thresholding: Otsu's method of 'grey level histograms' (GLH)      */
/* ----------------------------------------------------------------------- */
/* Threshold selection using grey level histograms. SMC-9 No 1 Jan 1979    */

Imagen OtsuThresholding (Imagen img);

/* ----------------------------------------------------------------------- */
/* Pun thresholding method using entropy                                   */
/* ----------------------------------------------------------------------- */

Imagen PunThresholding (Imagen img);

/* ----------------------------------------------------------------------- */
/* Kapur thresholding method using entropy                                 */
/* ----------------------------------------------------------------------- */

Imagen KapurThresholding (Imagen img);

/* ----------------------------------------------------------------------- */
/* Johannsen thresholding method using entropy                             */
/* ----------------------------------------------------------------------- */

Imagen JohannsenThresholding (Imagen img);

/* ----------------------------------------------------------------------- */
/*  Fuzzy Thresholding                                                     */
/* ----------------------------------------------------------------------- */

#define FUZZY_ENTROPY 1
#define FUZZY_YAGER   2

Imagen FuzzyThresholding (Imagen img, int method);

/* ----------------------------------------------------------------------- */
/* Minimum error thresholding                                              */
/* ----------------------------------------------------------------------- */

Imagen MinimumErrorThresholding (Imagen img);

/* ----------------------------------------------------------------------- */
/*  Relaxation Thresholding (Rosenfeld & Kak) - very very very slow        */
/* ----------------------------------------------------------------------- */

Imagen RelaxationThresholding (Imagen origen);

/* ----------------------------------------------------------------------- */
/* Digital Desk - Adaptive thresholding (Wellner, 1993)                    */
/* ----------------------------------------------------------------------- */
/* Ideal para textos (OCR)                                                 */

Imagen DigitalDeskThresholding (Imagen origen);



/* ----------------------------------------------------------------------- */
/*  Split (GRAYLEVEL)                                                      */
/* ----------------------------------------------------------------------- */
/*  min, max: 0 .. 255                                                     */

Imagen Split ( Imagen img, BYTE min, BYTE max );

/* ----------------------------------------------------------------------- */
/*  Region Splitting (GRAYLEVEL)                                           */
/* ----------------------------------------------------------------------- */

Imagen RegionSplit (Imagen origen, BYTE tonos);


/* ----------------------------------------------------------------------- */
/*  Texture Segmentation                                                   */
/* ----------------------------------------------------------------------- */

    // alg   Operador aplicado       TEXTURE_*
    // dir   Direccin               DIR_*
    // pd    Distancia de pixel      1
    // ws    Tamao de la ventana    6 ---> Ventana 13*13

#define TEXTURE_AVERAGE_GREY_LEVEL      0
#define TEXTURE_STANDARD_DEVIATION      1
#define TEXTURE_MAXIMUM_PROBABILITY     2
#define TEXTURE_ENERGY                  3
#define TEXTURE_CONTRAST                4
#define TEXTURE_HOMOGENEITY             5
#define TEXTURE_ENTROPY                 6

#define DIR_HORIZONTAL                  0
#define DIR_45                          1
#define DIR_VERTICAL                    2
#define DIR_135                         3

Imagen TextureSegmentation (Imagen img, int alg, int ws, int dir, int pd);
