/* ----------------------------------------------------------------------- */
/*    Fernando Berzal Galiano                ETS Ingeniera Informtica    */
/* ----------------------------------------------------------------------- */
/*                       Procesamiento de imgenes                         */
/* ----------------------------------------------------------------------- */
/*                             Segmentacin                                */
/* ----------------------------------------------------------------------- */

#include "IMG_SEGM.H"

#include <stdlib.h>
#include <math.h>

/* ----------------------------------------------------------------------- */
/*  Thresholding (GRAYLEVEL)                                               */
/* ----------------------------------------------------------------------- */

Imagen Thresholding ( Imagen img, BYTE umbral, BYTE min, BYTE max )
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  WORD   x, y;
  Imagen salida  = CreaImagen(anchura,altura);

  if (salida)
     for (x=0; x<anchura; x++)
	 for (y=0; y<altura; y++)
	     if (PIXEL(img,x,y)<umbral)
		PIXEL (salida,x,y) = min;
	     else
		PIXEL (salida,x,y) = max;

  return salida;
}

/* ----------------------------------------------------------------------- */
/*  Hysteresis Thresholding                                                */
/* ----------------------------------------------------------------------- */

#define PROPORCION 0.025 /* Proporcin de pxeles incluidos automticamente */

static void estimate_threshold (Imagen mag, BYTE *hi, BYTE *low)
{
  int  i,j,k;
  long hist[256], count;

  /* Histograma */

  for (k=0; k<256; k++) hist[k] = 0;

  for (i=0; i<ANCHURA(mag); i++)
      for (j=0; j<ALTURA(mag); j++)
	  hist[PIXEL(mag,i,j)]++;

  j = (int)(PROPORCION*ANCHURA(mag)*ALTURA(mag));

  k     = 0xFF;
  count = hist[0xFF];

  while (count<j && k>0) {
	k--;
	count += hist[k];
  }

  *hi = k;

  i=0;
  while (hist[i]==0) i++;

  *low = (*hi+i)/2.0;
}

/* --- */

static int ValidCoord (Imagen im, int i, int j)
{
  if ((i<0) || (i>=ANCHURA(im))) return 0;
  if ((j<0) || (j>=ALTURA(im))) return 0;

  return 1;
}

/* Recorre recursivamente los pixels (>low) a partir del pixel (i,j) */

static int trace ( int i, int j, BYTE low, Imagen img, Imagen mag)
{
  int  n,m;
  char flag = 0;

  if (PIXEL(img,i,j) == 0x00) {

     PIXEL(img,i,j) = 0xFF;

     for (n=-1; n<=1 && !flag; n++) {
	 for (m=-1; m<=1 && !flag; m++) {

	     if (i==0 && m==0) continue;

	     if (ValidCoord(mag, i+n, j+m) && PIXEL(mag,i+n,j+m) >= low)
		if (trace(i+n, j+m, low, img, mag))
		    flag=1;
	 }
     }

     return 1;

  } else

     return 0;
}

/* ----------------------------------------------------------------------- */
/*   HYSTERESIS: Partiendo de pxeles con un valor mayor que HIGH, trazar  */
/* una cadena conectada de pxeles con un valor mayor que LOW.             */

Imagen Hysteresis ( Imagen mag, BYTE high, BYTE low)
{
  int    i,j;
  WORD   anchura = ANCHURA(mag);
  WORD   altura  = ALTURA(mag);
  Imagen img     = CreaImagen(anchura,altura);

  if (high<low)
     estimate_threshold (mag, &high, &low);

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++)
	    if (PIXEL(mag,i,j) >= high)
		trace (i, j, low, img, mag);

  return img;
}

/* ----------------------------------------------------------------------- */
/*  Mean Grey Level Thresholding                                           */
/* ----------------------------------------------------------------------- */

Imagen MeanThresholding (Imagen img)
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);

  long   i, j;
  float  mean = 0;

  Imagen salida = CreaImagen(anchura,altura);

  /* Media */

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++)
	  mean += PIXEL(img,i,j);

  mean /= (float)((long)anchura*(long)altura);

  /* Umbralizacin */

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++)
	  if (PIXEL(img,i,j) < mean)
	     PIXEL(salida,i,j) = 0x00;
	  else
	     PIXEL(salida,i,j) = 0xFF;

  return salida;
}

/* ----------------------------------------------------------------------- */
/*  Black Percentage Thresholding (p-tile method)                          */
/* ----------------------------------------------------------------------- */
/*  pct: Porcentaje de pxeles negros (%)                                  */

Imagen BlackPercentageThresholding (Imagen img, float pct)
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  WORD   i, j;
  QUAD   hist[256];
  QUAD   count;
  BYTE   t;
  QUAD   n = (long) anchura * (long) altura;
  QUAD   m = (pct/100.0)*n;

  /* Histograma */

  for (i=0; i<256; i++)
      hist[i] = 0;

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++)
	  hist[PIXEL(img,i,j)] ++;

  /* Umbral */

  t     = 0;
  count = 0;

  for (i=0; i<256 && count<m; i++) {
      count += hist[i];

      if (count >= m)
	 t = i;
  }

  return Thresholding (img,t,0x00,0xFF);
}

/* ----------------------------------------------------------------------- */
/*  Two Peaks Thresholding                                                 */
/* ----------------------------------------------------------------------- */

Imagen TwoPeaksThresholding (Imagen img)
{
  WORD anchura = ANCHURA(img);
  WORD altura  = ALTURA(img);

  int  i,j,k;
  QUAD hist[256];
  BYTE t;

  /* Histograma */

  for (i=0; i<256; i++)
      hist[i] = 0;

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++)
	  hist[PIXEL(img,i,j)] ++;

  /* Primer pico del histograma */

  j = 0;

  for (i=0; i<256; i++)
      if (hist[i] > hist[j])
	 j = i;

  /* Segundo pico del histograma */

  k = 0;

  for (i=0; i<256; i++)
      if (i>0 && hist[i-1]<=hist[i] && i<256 && hist[i+1]<=hist[i])
	 if ((k-j)*(k-j)*hist[k] < (i-j)*(i-j)*hist[i])
	    k = i;

  /* Umbral */

  t = j;

  if (j<k) {

     for (i=j; i<k; i++)
	 if (hist[i] < hist[t])
	    t = i;

  } else {

     for (i=k; i<j; i++)
	 if (hist[i] < hist[t])
	    t = i;

  }

  return Thresholding (img,t,0x00,0xFF);
}

/* ----------------------------------------------------------------------- */
/*  Laplacian Thresholding                                                 */
/* ----------------------------------------------------------------------- */
/*  PCT (%): 85.0                                                          */

static int hi_pct (QUAD *hist, int NH, long n, float pct)
{
  int  i;
  QUAD j = 0;
  QUAD m = (pct/100.0) * n;

  for (i=0; i<NH; i++) {

      j += hist[i];

      if (j>=m)
	 return i;
  }

  return 0;
}

/* Clculo del umbral */

static BYTE peaks (Imagen img, float **lap, WORD lval)
{
  WORD anchura = ANCHURA(img);
  WORD altura  = ALTURA(img);
  QUAD hist[256];
  WORD i,j,k;
  BYTE t;

  /* Histograma */

  for (i=0; i<256; i++)
      hist[i] = 0;

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++)
	  if (lap[i][j] >= lval)
	      hist[PIXEL(img,i,j)]++;

  /* Primer pico */

  j = 0;

  for (i=0; i<256; i++)
      if (hist[i] > hist[j])
	 j = i;

  /* Segundo pico */

  k = 0;

  for (i=0; i<256; i++)
      if (i>0 && hist[i-1]<=hist[i] && i<256 && hist[i+1]<=hist[i])
	 if ((k-j)*(k-j)*hist[k] < (i-j)*(i-j)*hist[i])
	    k = i;

  /* Umbral */

  t = j;

  if (j<k) {

     for (i=j; i<k; i++)
	 if (hist[i] < hist[t])
	    t = i;

  } else {

     for (i=k; i<j; i++)
	 if (hist[i] < hist[t])
	    t = i;
  }

  return t;
}

/* ----------------------------------------------------------------------- */

Imagen LaplacianThresholding (Imagen img, float PCT)
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);

  QUAD   hist[8*256];

  float **Lap;
  int     i, j, x, y, t, v;

  /* Laplaciana de 'img' */

  Lap = (float **)calloc ( anchura, sizeof (float *) );

  for (i=0; i<anchura; i++)
      Lap[i] = (float *) calloc ( altura, sizeof (float)  );


  for (x=1; x<anchura-1; x++)
      for (y=1; y<altura-1; y++) {

	  Lap[x][y] = 0;

	  for (i= -1; i<=1; i++)
	      for (j= -1; j<=1; j++)
		  if (i!=0 || j!=0)
		     Lap[x][y] += PIXEL(img,x+i,y+j);

	  Lap[x][y] -= 8*(int)PIXEL(img,x,y);

	  if (Lap[x][y]<=0)
	     Lap[x][y] = 0.0;
	  else
	     Lap[x][y] /= 8.0;
  }

  /* %PCT ms elevado de la laplaciana */

  for (i=0; i<8*256; i++)
      hist[i] = 0;

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++)
	  hist[(int)(Lap[i][j])] ++;

  v = hi_pct (hist, 8*256, (long)anchura*(long)altura, PCT);

  /* Umbral */

  t = peaks (img, Lap, v);

  /* Liberacin de memoria: 'Lap' */

  for (i=0; i<anchura; i++)
      free(Lap[i]);

  free(Lap);

  /* Umbralizacin */

  return Thresholding (img,t,0x00,0xFF);
}

/* ----------------------------------------------------------------------- */
/*  Iterative Selection Thresholding                                       */
/* ----------------------------------------------------------------------- */

Imagen IterativeSelectionThresholding (Imagen img)
{
  WORD  anchura = ANCHURA(img);
  WORD  altura  = ALTURA(img);
  QUAD  n = (long)anchura * (long)altura;

  float tt, tb, to, t2;
  WORD  i, j;
  float no, nb;

  tb = 0.0;
  to = 0.0;
  no = 0;
  nb = 0;

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++)
	   to += PIXEL(img,i,j);

  tt = to / (float)n ;

  while (n) {

    no = 0;
    nb = 0;
    tb = 0.0;
    to = 0.0;

    for (i=0; i<anchura; i++)
	for (j=0; j<altura; j++)
	    if ( PIXEL(img,i,j) >= tt ) {
	       to += PIXEL(img,i,j);
	       no ++;
	    } else {
	       tb += PIXEL(img,i,j);
	       nb ++;
	    }

    if (no == 0) no = 1;
    if (nb == 0) nb = 1;

    t2 = (tb/nb + to/no )/2.0;

    if (t2==tt) n=0;

    tt = t2;
  }

  return Thresholding ( img, (BYTE)tt, 0x00, 0xFF);
}

/* --- */

Imagen FastIterSelectionThresholding (Imagen img)
{
  WORD  anchura = ANCHURA(img);
  WORD  altura  = ALTURA(img);
  QUAD  n       = (long)anchura*(long)altura;

  WORD i, j;
  long told, tt, a, b, c, d;
  QUAD hist[256];

  /* Histograma */

  for (i=0; i<256; i++)
      hist[i] = 0;

  tt = 0;

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++) {
	  hist[PIXEL(img,i,j)] ++;
	  tt += PIXEL(img,i,j);
      }

  tt = tt/(float)n;

  do {
      told = tt;

      a = 0;
      b = 0;

      for (i=0; i<=told; i++) {
	  a += i*hist[i];
	  b += hist[i];
      }

      b += b;

      c = 0;
      d = 0;

      for (i=told+1; i<256; i++) {
	  c += i*hist[i];
	  d += hist[i];
      }

      d += d;

      if (b==0) b = 1;
      if (d==0) d = 1;

      tt = a/b + c/d;

  } while (tt != told);

  return Thresholding ( img, (BYTE)tt, 0x00, 0xFF);
}

/* ----------------------------------------------------------------------- */
/*  Otsu Thresholding: Otsu's method of 'grey level histograms'            */
/* ----------------------------------------------------------------------- */
/* Threshold selection using grey level histograms. SMC-9 No 1 Jan 1979    */

static float w (float *p, int k)
{
  int   i;
  float x=0.0;

  for (i=1; i<=k; i++)
      x += p[i];

  return x;
}

/* --- */

static float u (float *p, int k)
{
  int   i;
  float x=0.0;

  for (i=1; i<=k; i++)
      x += (float)i*p[i];

  return x;
}

/* --- */

static float nu (float *p, int k, float ut, float vt)
{
  float x, y;

  y = w(p,k);
  x = ut*y - u(p,k);
  x = x*x;
  y = y*(1.0-y);

  if (y>0)
     x = x/y;
  else
     x = 0.0;

  return x/vt;
}

/* --- */

Imagen OtsuThresholding (Imagen img)
{
  WORD anchura = ANCHURA(img);
  WORD altura  = ALTURA(img);
  QUAD n       = (long)anchura*(long)altura;

  int   i,j,k,t;
  float y,z,p[256+1],ut,vt;

  /* Inicializacin  */

  for (i=0; i<=256; i++)
      p[i] = 0.0;

  /* Histograma */

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++)
	  p [ PIXEL(img,i,j) + 1 ] ++;

  /* Normalizacin */

  for (i=1; i<=256; i++)
      p[i] /= (float)n;


  ut = u(p,256);              		/* Media global    */
  vt = 0.0;            			/* Varianza global */

  for (i=1; i<=256; i++)
      vt += (i-ut)*(i-ut)*p[i];

  j = -1; k = -1;

  for (i=1; i<=256; i++) {
      if ((j<0) && (p[i] > 0.0))
	 j = i;       			/* Primer ndice */
      if (p[i] > 0.0)
	 k = i;                      	/* ltimo ndice */
  }

  z = -1.0;
  t = 0;

  for (i=j; i<=k; i++) {

      y = nu (p, i, ut, vt);

      if (y>=z) {
	 z = y;
	 t = i;
      }
  }

  return Thresholding(img,t,0x00,0xFF);
}

/* ----------------------------------------------------------------------- */
/* Pun thresholding method using entropy                                   */
/* ----------------------------------------------------------------------- */

static float entropy (float *h, int a)
{
  if (h[a] > 0.0)
     return -(h[a] * (float)log((double)h[a]));
  else
     return 0.0;
}

/* --- */

static float flog (float x)
{
  if (x <= 0.0)
     return 0.0;
  else
     return (float)log((double)x);
}

/* --- */

static float maxtot (float *h, int i)
{
  float x;
  int   j;

  x = h[0];

  for (j=1; j<=i; j++)
      if (x < h[j])
	 x = h[j];

  return x;
}

/* --- */

static float maxfromt (float *h, int i)
{
  int   j;
  float x;

  x = h[i+1];

  for (j=i+2; j<256; j++)
      if (x < h[j])
	 x = h[j];

  return x;
}

/* --- */

Imagen PunThresholding (Imagen img)
{
  WORD anchura = ANCHURA(img);
  WORD altura  = ALTURA(img);
  QUAD n       = (long)anchura*(long)altura;

  int   i, j, t;
  float Ht[256], HT, Pt[256], x, F[256], y, z, hist[256], to, from;

  /* Histograma */

  for (i=0; i<256; i++)
      hist[i] = 0.0;

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++)
	  hist[PIXEL(img,i,j)]++;

  for (i=0; i<256; i++)
      hist[i] /= (float)n;

  /* Factores */

  HT = Ht[0] = entropy (hist, 0);
  Pt[0] = hist[0];

  for (i=1; i<256; i++) {
      Pt[i] = Pt[i-1] + hist[i];
      x = entropy(hist, i);
      Ht[i] = Ht[i-1] + x;
      HT += x;
  }

  /* Funcin que se debe maximizar */

  t = 0;

  for (i=0; i<256; i++) {
      to   = maxtot(hist,i);
      from = maxfromt(hist, i);

      if (to > 0.0 && from > 0.0) {
	 x = (Ht[i]/HT)* flog(Pt[i])/flog(to);
	 y = 1.0 - (Ht[i]/HT);
	 z = flog(1 - Pt[i])/flog(from);
      } else
	 x = y = z = 0.0;

      F[i] = x + y*z;

      if (i>0 && F[i] > F[t])
	 t = i;
  }

  return Thresholding (img, t, 0x00, 0xFF);
}

/* ----------------------------------------------------------------------- */
/* Kapur thresholding method using entropy                                 */
/* ----------------------------------------------------------------------- */

static float entropia (float *h, int a, float p)
{
  if (h[a] > 0.0 && p>0.0)
     return -(h[a]/p * (float)log((double)(h[a])/p));
  else
     return 0.0;
}

/* --- */

Imagen KapurThresholding (Imagen img)
{
  WORD  anchura = ANCHURA(img);
  WORD  altura  = ALTURA(img);
  QUAD  n       = (long)anchura*(long)altura;

  int   i, j, t;
  float Hb, Hw, Pt[256], hist[256], F[256];

  /* Histograma */

  for (i=0; i<256; i++)
      hist[i] = 0.0;

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++)
	  hist[PIXEL(img,i,j)]++;

  for (i=0; i<256; i++)
      hist[i] /= (float)n;

  /* Factores */

  Pt[0] = hist[0];

  for (i=1; i<256; i++)
      Pt[i] = Pt[i-1] + hist[i];

  /* Funcin que se debe maximizar */

  t = 0;

  for (i=0; i<256; i++) {

      Hb = Hw = 0.0;

      for (j=0; j<256; j++)
	  if (j<=i)
	      Hb += entropia (hist, j, Pt[i]);
	  else
	      Hw += entropia (hist, j, 1.0-Pt[i]);

      F[i] = Hb+Hw;

      if (i>0 && F[i] > F[t])
	 t = i;
  }

  return Thresholding (img, t, 0x00, 0xFF);
}

/* ----------------------------------------------------------------------- */
/* Johannsen thresholding method using entropy                             */
/* ----------------------------------------------------------------------- */

static float entrop (float h)
{
  if (h > 0.0)
     return (-h * (float)log((double)(h)));
  else
     return 0.0;
}

/* --- */

static float floatlog (float x)
{
  if (x > 0.0)
     return (float)log((double)x);
  else
     return 0.0;
}

/* --- */

Imagen JohannsenThresholding (Imagen img)
{
  WORD  anchura = ANCHURA(img);
  WORD  altura  = ALTURA(img);
  QUAD  n       = (long)anchura*(long)altura;

  int   i, j, t;
  float Sb, Sw, Pt[256], hist[256], F[256], Pq[256];

  /* Histograma */

  for (i=0; i<256; i++)
      hist[i] = 0.0;

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++)
	  hist[PIXEL(img,i,j)]++;

  for (i=0; i<256; i++)
      hist[i] /= (float)n;

  /* Factores */

  Pt[0] = hist[0];
  Pq[0] = 1.0 - Pt[0];

  for (i=1; i<256; i++) {
      Pt[i] = Pt[i-1] + hist[i];
      Pq[i] = 1.0 - Pt[i-1];
  }

  /* Funcin que se debe minimizar */

  t = -1;

  for (i=0; i<256; i++) {
      if (hist[i]>0) {
	  Sb = floatlog(Pt[i])+(1.0/Pt[i])*(entrop(hist[i])+entrop(Pt[i-1]));
	  Sw = floatlog(Pq[i])+(1.0/Pq[i])*(entrop(hist[i])+entrop(Pq[i+1]));
	  F[i] = Sb+Sw;

	  if ((t<0) || (F[i] < F[t]))
	     t = i;
      }
  }

  return Thresholding (img, t, 0x00, 0xFF);
}

/* ----------------------------------------------------------------------- */
/*  Fuzzy Thresholding                                                     */
/* ----------------------------------------------------------------------- */

#define FUZZY_ENTROPY 1
#define FUZZY_YAGER   2

/* --- */

static double Ux (int g, int u0, int u1, int t)
{
  double ux, x;

  if (g <= t) {
     x = 1.0 + ((double)abs(g - u0))/255.0;
     ux = 1.0/x;
  } else {
     x = 1.0 + ((double)abs(g - u1))/255.0;
     ux = 1.0/x;
  }

  return ux;
}

/* --- */

static double Shannon (double x)
{
  if (x > 0.0 && x < 1.0)
     return (double)(-x*log((double)x) - (1.0-x)*log((double)(1.0-x)));
  else
     return 0.0;
}

/* --- */

static double Yager (int u0, int u1, int t)
{
  int i;
  double x, sum=0.0;

  for (i=0; i<256; i++) {
      x = Ux(i, u0, u1, t);
      x = x*(1.0-x);
      sum += x*x;
  }

  x = (double)sqrt((double)sum);

  return x;
}

/* --- */

static double fuzzy (QUAD *hist, int u0, int u1, int t, int method)
{
  int    i;
  double H=0;

  if (method == FUZZY_ENTROPY) {

     for (i=0; i<256; i++)
	 H += Shannon(Ux(i,u0,u1,t))*hist[i];

     return H;

  } else {

     return Yager (u0, u1, t);

  }
}

/* --- */

Imagen FuzzyThresholding (Imagen img, int method)
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);

  QUAD   hist[256], S[256], Sbar[256], W[256], Wbar[256];
  QUAD   sum, minsum;

  double F[256];
  double MN   = (double)anchura*(double)altura;
  double maxv = 0.0;
  double delta;

  int    i, j, start, end, u0, u1, t, tbest=-1;

  /* Histograma */

  for (i=0; i<256; i++)
      hist[i] = 0;

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++)
	  hist[PIXEL(img,i,j)] ++;

  /* Histograma acumulado */

  S[0] = hist[0];
  W[0] = 0;

  for (i=1; i<256; i++){
      S[i] = S[i-1] + hist[i];
      W[i] = i*hist[i] + W[i-1];
  }

  /* Histograma acumulado inverso */

  Sbar[255] = 0;
  Wbar[255] = 0;

  for (i=254; i>=0; i--) {
      Sbar[i] = Sbar[i+1] + hist[i+1];
      Wbar[i] = Wbar[i+1] + (i+1)*hist[i+1];
  }


  for (t=1; t<255; t++)	{

      if ( hist[t]!=0.0 && S[t]!=0.0 && Sbar[t]!=0.0) {

	 u0 = (int)(W[t]/S[t] + 0.5);
	 u1 = (int)(Wbar[t]/Sbar[t] + 0.5);

	 F[t] = fuzzy (hist, u0, u1, t, method)/MN;

	 if (F[t] > maxv)
	    maxv = F[t];

	 if (tbest < 0)
	    tbest = t;
	 else if (F[t] < F[tbest])
	    tbest = t;
      }
  }

  /* Encontrar el mejor umbral dentro de un rango */

  delta = F[tbest] + (maxv-F[tbest])*0.05;        /* 5% */

  start = (int)(tbest - delta);

  if (start <= 0)
     start = 1;

  end = (int)(tbest + delta);

  if (end>=255)
     end = 254;

  minsum = 1000000000;

  for (i=start; i<=end; i++) {

      sum = hist[i-1] + hist[i] + hist[i+1];

      if (sum < minsum) {
	 t = i;
	 minsum = sum;
      }
  }

  /* Umbralizacin */

  return Thresholding (img,t,0x00,0xFF);
}

/* ----------------------------------------------------------------------- */
/* Minimum error thresholding                                              */
/* ----------------------------------------------------------------------- */

static QUAD h[256];

static float METlog (float x)
{
  if (x > 0.0)
     return (float)log((double)x);
  else
     return 0.0;
}

static float P1 (int t)
{
  int  i;
  QUAD sum = 0;

  for (i=0; i<=t; i++)
      sum += h[i];

  return (float)sum;
}

static float P2 (int t)
{
  int  i;
  QUAD sum = 0;

  for (i=t+1; i<256; i++)
      sum += h[i];

  return (float)sum;
}

static float u1 (int t)
{
  int   i;
  float sum=0.0, p;

  p = P1(t);

  if (p <= 0.0) {

     return 0.0;

  } else {

     for (i=0; i<=t; i++)
	 sum += h[i]*i;

     return sum/p;
  }

}

static float u2 (int t)
{
  int   i;
  float sum=0.0, p;

  p = P2(t);

  if (p <= 0.0) {

     return 0.0;

  } else {

     for (i=t+1; i<256; i++)
	 sum += h[i]*i;

     return sum/p;
  }

}

static float s1 (int t)
{
  int   i;
  float sum=0.0, p, u, x;

  p = P1(t);

  if (p<=0.0) {

     return 0.0;

  } else {

     u = u1(t);

     for (i=0; i<=t; i++) {
	 x = (i-u)*(i-u);
	 sum += x*h[i];
     }

     return sum/p;
  }
}

static float s2 (int t)
{
  int   i;
  float sum=0.0, p, u, x;

  p = P2(t);

  if (p<=0.0) {

     return 0.0;

  } else {

     u = u2(t);

     for (i=t+1; i<256; i++) {
	 x = (i-u)*(i-u);
	 sum += x*h[i];
     }

     return sum/p;
  }
}

static float J(int t)
{
  float a = P1(t);
  float b = s1(t);
  float c = P2(t);
  float d = s2(t);

  return 1.0 + 2.0*(a*METlog(b) + c*METlog(d)) - 2.0*(a*METlog(a) + c*METlog(c));
}

/* --- */

Imagen MinimumErrorThresholding (Imagen img)
{
  int   i,j,t;
  float F[256];

  WORD anchura = ANCHURA(img);
  WORD altura  = ALTURA(img);

  for (i=0; i<256; i++)
      h[i] = 0;

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++)
	  h[PIXEL(img,i,j)] ++;  // Histograma

  t = 0;

  for (i=1; i<256; i++) {
      F[i] = J(i);

      if (F[i] < F[t])
	 t = i;
  }

  return Thresholding (img,t,0x00,0xFF);
}

/* ----------------------------------------------------------------------- */
/*  Relaxation Thresholding (Rosenfeld & Kak) - very slow                  */
/* ----------------------------------------------------------------------- */

/* Media */

static float mean (Imagen img)
{
  WORD anchura = ANCHURA(img);
  WORD altura  = ALTURA(img);

  int   i, j;
  float sum = 0;

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++)
	    sum += PIXEL(img,i,j);

  return sum/(float)((long)anchura*(long)altura);
}

/* Mnimo y mximo */

static void minmax (Imagen img, float *gmin, float *gmax)
{
  WORD anchura = ANCHURA(img);
  WORD altura  = ALTURA(img);
  int  i,j;

  *gmin = *gmax = PIXEL(img,0,0);

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++)
	  if (*gmin > PIXEL(img,i,j))
	     *gmin = PIXEL(img,i,j);
	  else if (*gmax < PIXEL(img,i,j))
	     *gmax = PIXEL(img,i,j);
}

// void meanminmax (Imagen img, int r, int c, float *mean, float *xmin, float *xmax)
// {
//   int  i,j;
//   long sum = 0, k=0;
//
//   *xmin = *xmax = PIXEL(img,r,c);
//
//   for (i=r-10; i<=r+10; i++)
//       for (j=c-10; j<=c+10; j++)
// 	    if (range(img,i,j)) {
// 	       if (*xmin > PIXEL(img,i,j))
// 		  *xmin = PIXEL(img,i,j);
// 	       else if (*xmax < PIXEL(img,i,j))
// 		  *xmax = PIXEL(img,i,j);
//
// 	       sum += PIXEL(img,i,j);
// 	       k++;
// 	    }
//
//   *mean = (float)sum/(float)k;
// }

/* p: Probabilidad de que un pixel sea negro  */
/* q: Probabilidad de que un pixel sea blanco */

static void assign_class (Imagen img, float **p, float **q)
{
  WORD anchura = ANCHURA(img);
  WORD altura  = ALTURA(img);

  int   i,j;
  float ud2, y, z, u, lm2;

  /* Rosenfeld & Smith class assignment */

  u = mean (img);
  minmax (img, &z, &y);
  ud2 = 2.0*(u-z);
  lm2 = 2.0*(y-u);

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++)
	  if (PIXEL(img,i,j) <= u) {
	     p[i][j] = 0.5 + (u-(int)PIXEL(img,i,j))/ud2;
	     q[i][j] = 1.0 - p[i][j];
	  } else {
	     q[i][j] = 0.5 + ((int)PIXEL(img,i,j)-u)/lm2;
	     p[i][j] = 1.0 - q[i][j];
	  }

  // Parker class assignment (using small regions)
  // ---------------------------------------------
  //
  // for (i=0; i<anchura; i++)             /* Initially, all are 50% */
  //     for (j=0; j<altura; j++)
  //         p[i][j] = q[i][j] = 0.5;
  //
  // for (i=0; i<anchura; i++)             /* Mean of local area */
  //	 for (j=0; j<altura; j++) {
  //	     meanminmax (im, i, j, &u, &z, &y);
  //	     ud2 = 2.0*(u-z);
  //	     lm2 = 2.0*(y-u);
  //	     if (PIXEL(img,i,j) <= u) {
  //		p[i][j] = 0.5 + (u-(int)PIXEL(img,i,j))/ud2;
  //		q[i][j] = 1.0 - p[i][j];
  //	     } else {
  //		q[i][j] = 0.5 + ((int)PIXEL(img,i,j)-u)/lm2;
  //		p[i][j] = 1.0 - q[i][j];
  //	     }
  //	  }
}

/* --- */

static float R (int l1, int l2)
{
  if (l1 == l2)
     return 0.9;
  else
     return -0.9;
}

/* --- */

static float Q(float **p, float **q, int r, int c, int clase)
{
  int   i,j;
  float sum = 0.0;

  for (i=r-1; i<=r+1; i++)
      for (j=c-1; j<=c+1; j++)
	  if (i!=r || j!=c)
	     sum += R(clase, 0)*p[i][j] + R(clase, 1)*q[i][j];

  return sum/8.0;
}

/* --- */

static float update ( WORD anchura, WORD altura,
                      float **p, float **q, float **pp, float **qq)
{
  float z, qw, pk, qb;
  int   i, j;

  for (i=1; i<anchura-1; i++)
      for (j=1; j<altura-1; j++) {
	  qb = (1.0 + Q(p, q, i, j, 0));
	  qw = (1.0 + Q(p, q, i, j, 1));
	  pk = p[i][j]*qb + q[i][j]*qw;

	  if (pk != 0.0) {
	     pp[i][j] = p[i][j]*qb/pk;
	     qq[i][j] = q[i][j]*qw/pk;
	  }
      }

  z = 0.0;

  for (i=1; i<anchura-1; i++)
      for (j=1; j<altura-1; j++) {
	  z += fabs(p[i][j]-pp[i][j]) + fabs(q[i][j]-qq[i][j]);
	  p[i][j] = pp[i][j];
	  q[i][j] = qq[i][j];
	  qq[i][j] = pp[i][j] = 0.0;
      }

  return z;
}

/* Threshold */

static void threshRK (Imagen img, float **p, float **q)
{
  WORD anchura = ANCHURA(img);
  WORD altura  = ALTURA(img);

  int i,j;

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++)
	  if (p[i][j] > q[i][j])
	     PIXEL(img,i,j) = 0x00;
	  else
	     PIXEL(img,i,j) = 0xFF;
}

/* --- */

static float **f2d_alloc (int nr, int nc)
{
  int     i ;
  float **x = (float **)calloc ( nr, sizeof (float *) );

  for (i=0; i<nr; i++)
      x[i] = (float *) calloc ( nc, sizeof (float)  );

  return x;
}

static void f2d_free (float **x, int nr)
{
  int i;

  for (i=0; i<nr; i++)
      free (x[i]);

  free (x);
}

#define MAX_RT_ITER 10

Imagen RelaxationThresholding (Imagen origen)
{
  WORD   anchura = ANCHURA(origen);
  WORD   altura  = ALTURA(origen);
  Imagen img     = CreaImagen(anchura,altura);

  float res = 0.0, minres = 10000.0;
  float **p, **q, **pp, **qq;
  int   iter = 0, count = 0;

  p  = f2d_alloc (anchura, altura);
  q  = f2d_alloc (anchura, altura);
  pp = f2d_alloc (anchura, altura);
  qq = f2d_alloc (anchura, altura);

  assign_class (origen, p, q);   /* Initial assignment of pixel classes */

  // threshRK (img, p, q);

  do {

    res = update (anchura, altura, p, q, pp, qq);

    iter ++;

    if (res < minres) {
       minres = res;
       count = 1;
    } else if (fabs(res-minres) < 0.0001) {
       count++;
    }

    // threshRK (img, p, q);

  } while ( iter<MAX_RT_ITER && res>1.0 && count<3 ) ;

  threshRK (img, p, q);

  f2d_free(p, anchura);
  f2d_free(q, anchura);
  f2d_free(pp,anchura);
  f2d_free(qq,anchura);

  return img;
}

/* ----------------------------------------------------------------------- */
/* Digital Desk - Adaptive thresholding (Wellner, 1993)                    */
/* ----------------------------------------------------------------------- */
/* Ideal para textos (OCR)                                                 */

Imagen DigitalDeskThresholding (Imagen origen)
{
  WORD   anchura = ANCHURA(origen);
  WORD   altura  = ALTURA(origen);
  Imagen img     = CreaImagen(anchura,altura);

  float  pct     = 15.0;      /* Make smaller to darken the image          */
  float  Navg    = 8.0;       /* Fraction of a row in the average (ie 1/8) */

  int    row, col, inc;
  long   i,j;
  float  mean;

  long   N   = (long)anchura * (long)altura;
  float  s   = (int)(float)(anchura/Navg);
  float  sum = ((float)127)*s;

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++)
	  PIXEL(img,i,j) = PIXEL(origen,i,j);

  /* 'Boustrophedon' scanning */

  row = col = 0;
  inc = 1;

  for (i=0; i<N; i++) {

      if (col>=anchura) {
	 col = anchura-1;
	 row ++;
	 inc = -1;
      } else if (col<0) {
	 col = 0;
	 row ++;
	 inc = 1;
      }

      /* Estimate the mean of the last NC/8 pixels. */

      sum  = sum - sum/s + PIXEL(img,col,row);
      mean = sum/s;

      if (PIXEL(img,col,row) < mean*(100-pct)/100.0)
	 PIXEL(img,col,row) = 0x00;
      else
	 PIXEL(img,col,row) = 0xFF;

      col += inc;
  }

  return img;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

/* ----------------------------------------------------------------------- */
/*  Split                                                                  */
/* ----------------------------------------------------------------------- */

Imagen Split ( Imagen img, BYTE min, BYTE max )
{
  WORD  anchura = ANCHURA(img);
  WORD  altura  = ALTURA(img);
  QUAD  pixels  = (long)anchura*(long)altura;

  QUAD  gris[256];
  QUAD  nivel[256];
  float media[256];

  WORD  x,y;
  BYTE  umbral;
  float valor, aux, tmp;

  /* Histograma */

  for (x=0; x<256; x++)
      gris[x] = 0;

  for (x=0; x<anchura; x++)
      for (y=0; y<altura; y++)
	   gris[PIXEL(img,x,y)]++;

  /* Histograma acumulado */

  nivel[0] = gris[0];

  for (x=1; x<256; x++)
      nivel[x] = nivel[x-1] + gris[x];

  /* Umbral = max { t(x)/(P-t(x)) * (m(g)-m(255))^2 } - 1 */
  /*   t(x) = nivel[x]                                    */
  /*   P    = Pxeles de la imagen                        */
  /*   m(g) = sum(x*gris[x])/nivel[x]                     */

  media[0] = 0;
  valor    = 0;

  for (x=1; x<256; x++)
      media[x] = media[x-1] + x*gris[x];

  for (x=1; x<256; x++)
      media[x] = (nivel[x]>0)? media[x] / nivel[x]: 0;


  for (x=0; x<256 && pixels-nivel[x]>0; x++) {
      tmp = media[x]-media[255];
      tmp *= tmp;

      aux = (nivel[x]*tmp)/(pixels-nivel[x]);

      if ((x==0) || (aux>valor)) {
	 umbral = x;
	 valor  = aux;
      }
  }

  if (umbral>0)
     umbral--;

  return Thresholding ( img, umbral, min, max );
}

/* ----------------------------------------------------------------------- */
/*  Region Splitting (GRAYLEVEL)                                           */
/* ----------------------------------------------------------------------- */

Imagen RegionSplit (Imagen origen, BYTE tonos)
{
  WORD anchura = ANCHURA(origen);
  WORD altura  = ALTURA(origen);
  WORD x,y;
  BYTE color;
  QUAD gris[256];
  BYTE colores[256];
  BYTE nuevo[256];

  Imagen destino = CreaImagen(anchura,altura);

  /* Histograma */

  for (x=0; x<256; x++)
      gris[x] = 0;

  for (x=1; x<anchura; x++)
      for (y=0; y<altura;y++)
	   gris[PIXEL(origen,x,y)]++;

  /* Picos del histograma */

  for (x=0; x<256; x++)
      colores[x] = x;

  for (x=0; x<256; x++)
      for (y=x+1; y<256; y++)
	  if (gris[colores[x]]<gris[colores[y]]) {
	     color      = colores[x];
	     colores[x] = colores[y];
	     colores[y] = color;
	  }

  for (x=256-tonos; x<256; x++)
      for (y=x+1; y<256; y++) {
	  if (colores[x]>colores[y]) {
	     color      = colores[x];
	     colores[x] = colores[y];
	     colores[y] = color;
	  }
      }

  /* Histograma --> Frecuencia acumulada */

  for (x=1; x<256; x++)
      gris[x] += gris[x-1];

  /* Asignacin de colores */

  for (x=0; x<=colores[256-tonos]; x++)
      nuevo [x] = 0;

  color = 256-tonos;

  for (x=colores[256-tonos]; x<colores[255]; x++) {

      if (x==colores[color+1]) {
	 color++;
	 nuevo[x] = (tonos+color-256)*(256.0/tonos);
      } else {
	 if (gris[x]-gris[color] > gris[color+1]-gris[color])
	    nuevo[x] = (tonos+color+1-256)*(256/tonos);
	 else
	    nuevo[x] = (tonos+color  -256)*(256/tonos);
      }
  }

  for (x=colores[255]; x<256; x++)
      nuevo [x] = 0xFF;

  /* Dibujo final */

  for (x=0; x<anchura; x++)
      for (y=0; y<altura; y++)
	  PIXEL(destino,x,y) = nuevo[PIXEL(origen,x,y)];

  return destino;
}


/* ----------------------------------------------------------------------- */
/*  Texture Segmentation                                                   */
/* ----------------------------------------------------------------------- */

    // alg   Operador aplicado       TEXTURE_*
    // dir   Direccin               DIR_*
    // pd    Distancia de pixel      1
    // ws    Tamao de la ventana    6 ---> Ventana 13*13

#define TEXTURE_AVERAGE_GREY_LEVEL      0
#define TEXTURE_STANDARD_DEVIATION      1
#define TEXTURE_MAXIMUM_PROBABILITY     2
#define TEXTURE_ENERGY                  3
#define TEXTURE_CONTRAST                4
#define TEXTURE_HOMOGENEITY             5
#define TEXTURE_ENTROPY                 6

#define DIR_HORIZONTAL                  0
#define DIR_45                          1
#define DIR_VERTICAL                    2
#define DIR_135                         3

/* --- */

typedef struct hstruct {
			 float Sum[2*256], Diff[2*256];
			 float Mean;
		       } HISTOGRAM;

/* Mtricas */

static float TSstddev (HISTOGRAM *h)
{
  float s1, s2, meanmean, var;
  int   i,j;

  s1 = s2 = 0.0;

  meanmean = h->Mean + h->Mean;

  for (i=0; i<2*256; i++) {
      s1 += (i-meanmean)*(i-meanmean)*(h->Sum[i]);
      j = i-255;
      s2 += j*j*(h->Diff[i]);
  }

  var = (s1 + s2)/2.0;

  return (float)sqrt((double)var);
}

static float TSp_max (HISTOGRAM *h)
{
  int   i;
  float y = 0.0;

  for (i=0; i<2*256; i++)
      if (h->Sum[i] > y) y = h->Sum[i];

  return y;
}

static float TSenergy (HISTOGRAM *h)
{
  int   i;
  float y=0.0, z = 0.0;

  for (i=0; i<2*256; i++) {
      y += (h->Sum[i])*(h->Sum[i]);
      z += (h->Diff[i])*(h->Diff[i]);
  }

  return y*z;
}

static float TScontrast (HISTOGRAM *h)
{
  int   i;
  float y=0.0;

  for (i= -255; i<=255; i++)
      y += i*i*(h->Diff[i+256]);

  return y;
}

static float TShomo (HISTOGRAM *h)
{
  int   i;
  float y = 0.0;

  for (i= -255; i<=255; i++)
      y += 1.0/(1.0 + i*i) * (h->Diff[i+256]);

  return y;
}

static float TSentropy (HISTOGRAM *h)
{
  int   i;
  float h1=0.0, h2=0.0;

  for (i=0; i<2*256; i++)
      if (h->Sum[i] > 0)
	 h1 += (h->Sum[i])*log((double)h->Sum[i]);

  for (i=-255; i<=255; i++)
      if (h->Diff[i+256] > 0)
	 h2 += (h->Diff[i+256])*log((double)h->Diff[i+256]);

  return (-h1-h2);
}

/* GLCM: Sum and difference histograms, and the mean */

static void TSglcm (Imagen img, int d, int dir,
                    int top, int bottom, int left, int right, HISTOGRAM *h)
{
  int   p1, p2, i, j;
  float k;
  float sum;

  k   = 0;
  sum = 0.0;

  for (i=0; i<2*256; i++) {
      h->Sum[i]  = 0;
      h->Diff[i] = 0;
  }

  for (i=top; i<bottom; i++)
      for (j=left; j<right; j++) {

	    p1 = PIXEL(img,i,j);

	    if ((j+d<right) && (dir==0)) 		      // Horizontal
		p2 = PIXEL(img,i,j+d);
	    else if ((i+d<bottom) && (dir==2)) 		      // Vertical
		p2 = PIXEL(img,i+d,j);
	    else if ((i+d<bottom) && (j-d>=left) && (dir==1)) // 45
		p2 = PIXEL(img,i+d,j-d);
	    else if ((i+d<bottom) && (j+d<right) && (dir==3)) // 135
		p2 = PIXEL(img,i+d,j+d);
	    else
		continue;

	    k++;
	    h->Sum[p1+p2]++;
	    h->Diff[256+(p1-p2)]++;
	  }

  for (i=0; i<2*256; i++) {
      if (k>0) {
         h->Sum[i] /= k;
	 h->Diff[i] /= k;
      }
      sum += h->Sum[i]*i;
  }

  h->Mean = sum/2.0;

}

/* --- */

static void measure_window (Imagen img, int op, float **res, int ws, int dir, int pd)
{
  int       top, bottom, left, right;
  int       wtop, wbottom, wleft, wright;
  int       i, j;
  HISTOGRAM h;

  top    = ws;
  bottom = ALTURA(img)-ws-1;
  left   = ws;
  right  = ANCHURA(img)-ws-1;

  for (i=left; i<=right; i++) {
      for (j=top; j<=bottom; j++) {

	  wtop    = i-ws;
	  wbottom = i+ws;
	  wleft   = j-ws;
	  wright  = j+ws;

	  TSglcm (img, pd, dir, wtop, wbottom, wleft, wright, &h);

	  switch (op) {

	    case TEXTURE_AVERAGE_GREY_LEVEL:  res[i][j] = h.Mean;
					      break;
	    case TEXTURE_STANDARD_DEVIATION:  res[i][j] = TSstddev(&h);
					      break;
	    case TEXTURE_MAXIMUM_PROBABILITY: res[i][j] = TSp_max(&h);
					      break;
	    case TEXTURE_ENERGY:              res[i][j] = TSenergy(&h);
					      break;
	    case TEXTURE_CONTRAST:            res[i][j] = TScontrast(&h);
					      break;
	    case TEXTURE_HOMOGENEITY:         res[i][j] = TShomo(&h);
					      break;
	    case TEXTURE_ENTROPY:             res[i][j] = TSentropy(&h);
					      break;
	  }
      }
  }
}

/* --- */

static Imagen rescale (float **res, int nr, int nc, int ws)
{
  int    i, j;
  float  xmin, xmax, xd;
  Imagen img = CreaImagen(nr-2*ws,nc-2*ws);

  xmin = xmax = res[ws][ws];

  for (i=ws; i<nr-ws; i++)
      for (j=ws; j<nc-ws; j++) {
	  if (res[i][j] > xmax) xmax = res[i][j];
	  if (res[i][j] < xmin) xmin = res[i][j];
      }

  xd = xmax - xmin;

  if (xd>0)
     for (i=ws; i<nr-ws; i++)
         for (j=ws; j<nc-ws; j++)
	     PIXEL(img,i-ws,j-ws) = (BYTE) ( ( (res[i][j]-xmin)/xd ) * 255.0 );

  return img;
}

/* --- */

Imagen TextureSegmentation (Imagen img, int alg, int ws, int dir, int pd)
{
  float **res ;
  int     i;
  Imagen  salida;

  /* Reserva de memoria */

  res = (float **)calloc ( ANCHURA(img), sizeof (float *) );

  for (i=ws; i<ANCHURA(img)-ws; i++)
      res[i] = (float *) calloc ( ALTURA(img), sizeof (float)  );

  /* Applicacin de la mtrica escogida */

  measure_window (img, alg, res, ws, dir, pd);

  salida = rescale (res, ANCHURA(img), ALTURA(img), ws);

  /* Liberacin de memoria */

  for (i=ws; i<ANCHURA(img)-ws; i++)
      free(res[i]);

  free(res);

  return salida;
}

