/* ----------------------------------------------------------------------- */
/*    Fernando Berzal Galiano                ETS Ingeniera Informtica    */
/* ----------------------------------------------------------------------- */
/*                       Procesamiento de imgenes                         */
/* ----------------------------------------------------------------------- */
/*                         Introduccin de ruido                           */
/* ----------------------------------------------------------------------- */

#include "IMG_SNR.H"

/* ----------------------------------------------------------------------- */
/*                    Generacin de nmeros aleatorios                     */
/* ----------------------------------------------------------------------- */

#include <math.h>
#include <stdlib.h>

/* ----------------------------- */
/* ----------------------------- */
/* Generador congruencial lineal */
/* ----------------------------- */
/* ----------------------------- */
/*  Mdulo 2^29 = 536870912     */
/*  Multiplicador 5             */
/*  Incremento 3^18 = 387420489 */
/* ----------------------------- */
/* Perodo: 536870912            */
/* ----------------------------- */

unsigned long anterior;


void SemillaGCL (unsigned long numero)
{
  anterior = numero&0x1FFFFFFF;
}


unsigned long GCL (void)
{
  anterior = (5*anterior+387420489)&0x1FFFFFFF;

  /* Valor mximo = 5*anterior+387420489 = 3071775044  */
  /*  No se produce overflow:  ULONG_MAX = 4294967295 */

  /* El mdulo % se sustituye por una operacin AND    */

  return anterior;
}



/* ----------------------------- */
/* ----------------------------- */
/* Generador uniforme u(0,1)     */
/* ----------------------------- */
/* ----------------------------- */
/* Perodo: 536870912            */
/* ----------------------------- */

/* A partir de un generador congruencial lineal: u = GCL / 536870912 */

unsigned long Un;


void SemillaUniforme (unsigned long numero)
{
  Un = numero&0x1FFFFFFF;
}


double Uniforme (void)
{
  /* ----------------------------- */
  /* Generador congruencial lineal */
  /* ----------------------------- */

  Un = (5*Un+387420489)&0x1FFFFFFF;

  /* ----------------------------- */
  /* Generador uniforme u(0,1)     */
  /* ----------------------------- */

  return Un/536870912.0;
}


/* --------------------- */
/* --------------------- */
/* Generador exponencial */
/* --------------------- */
/* --------------------- */

double        media;

unsigned long EXPn;


void SemillaExponencial (unsigned long semilla, double media_generador)
{
  EXPn  = semilla&0x1FFFFFFF;
  media = media_generador;
}


double Exponencial (void)
{
  /* ----------------------------- */
  /* Generador congruencial lineal */
  /* ----------------------------- */

  EXPn = (5*EXPn+387420489)&0x1FFFFFFF;

  /* ----------------------------- */
  /* Generador uniforme u(0,1)     */
  /* ----------------------------- */
  /* Mtodo de inversin:          */
  /*  exp = -media * ln(1-u)       */
  /* ----------------------------- */

  return -media*log(1-EXPn/536870912.0);
}

/* ---------------- */
/* ---------------- */
/* Generador normal */
/* ---------------- */
/* ---------------- */

double mediaNORMAL;
double sdNORMAL;

void GeneradorNormal (unsigned long semilla, double media_generador, double sd_generador)
{
  SemillaUniforme(semilla);

  mediaNORMAL = media_generador;
  sdNORMAL    = sd_generador;
}

double Normal (void)
{
  static double a = 0, w2 = 0;
  double        u1, u2, b, v, x, w1;
  int           s, t;

  if (a == 1) {
     a = 0;
     return mediaNORMAL + w2*sdNORMAL;
  }

  b = 2;

  while (b>1) {

    a  = 1;
    u1 = Uniforme();
    u1 = u1 + u1;

    if (u1 > 1.0) {
       s = 0;
       u1 -= 1.0;
    } else
       s = 1;

    u2 = Uniforme();
    u2 = u2+u2;

    if (u2 > 1.0) {
       t = 0;
       u2 -= 1.0;
    } else
       t = 1;

    b = u1*u1 + u2*u2;
  }

  x = Uniforme();

  if (x!=0)
     v = -log(x);
  else
     v = 0;

  w1 = u1*sqrt(2*v/b);
  w2 = u2*sqrt(2.0*v/b);

  if (s==0)
     w1 = -w1;

  if (t == 0)
     w2 = -w2;

  return mediaNORMAL + w1*sdNORMAL;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

Imagen RuidoAditivoUniforme  (Imagen img, float p, int max)
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  Imagen salida  = CreaImagen(anchura,altura);
  int    i,j;
  double x;

  randomize();
  SemillaUniforme ( rand() );

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++) {

          if (Uniforme()<p) {

	     x = PIXEL(img,i,j) + 2*max*(Uniforme()-0.5);

             if (x>255)
                PIXEL(salida,i,j) = 255;
             else if (x<0)
                PIXEL(salida,i,j) = 0;
             else
                PIXEL(salida,i,j) = x;

          } else {
             PIXEL(salida,i,j) = PIXEL(img,i,j);
          }
      }

  return salida;
}

/* ----------------------------------------------------------------------- */

Imagen RuidoAditivoNormal    (Imagen img, float p, float media, float sd)
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  Imagen salida  = CreaImagen(anchura,altura);
  int    i,j;
  double x;

  randomize();
  GeneradorNormal ( rand(), media, sd );

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++) {

          if (Uniforme()<p) {

	     x = PIXEL(img,i,j) + Normal();

             if (x>255)
                PIXEL(salida,i,j) = 255;
             else if (x<0)
                PIXEL(salida,i,j) = 0;
             else
                PIXEL(salida,i,j) = x;

          } else {
             PIXEL(salida,i,j) = PIXEL(img,i,j);
          }
      }

  return salida;
}

/* ----------------------------------------------------------------------- */

Imagen RuidoAditivoExponencial (Imagen img, float p, float media)
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  Imagen salida  = CreaImagen(anchura,altura);
  int    i,j;
  double x;

  randomize();
  SemillaExponencial ( rand(), media );

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++) {

          if (Uniforme()<p) {

             if (Uniforme()<0.5)
	        x = PIXEL(img,i,j) + Exponencial();
             else
                x = PIXEL(img,i,j) - Exponencial();

             if (x>255)
                PIXEL(salida,i,j) = 255;
             else if (x<0)
                PIXEL(salida,i,j) = 0;
             else
                PIXEL(salida,i,j) = x;

          } else {
             PIXEL(salida,i,j) = PIXEL(img,i,j);
          }
      }

  return salida;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

Imagen RuidoMultiplicativoUniforme  (Imagen img, float p, int max)
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  Imagen salida  = CreaImagen(anchura,altura);
  int    i,j;
  double x;

  randomize();
  SemillaUniforme ( rand() );

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++) {

          if (Uniforme()<p) {

	     x = PIXEL(img,i,j)*max*Uniforme();

             if (x>255)
                PIXEL(salida,i,j) = 255;
             else if (x<0)
                PIXEL(salida,i,j) = 0;
             else
                PIXEL(salida,i,j) = x;

          } else {
             PIXEL(salida,i,j) = PIXEL(img,i,j);
          }
      }

  return salida;
}

/* ----------------------------------------------------------------------- */

Imagen RuidoMultiplicativoNormal (Imagen img, float p, float media, float sd)
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  Imagen salida  = CreaImagen(anchura,altura);
  int    i,j;
  double x;

  randomize();
  GeneradorNormal ( rand(), media, sd );

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++) {

          if (Uniforme()<p) {

	     x = Normal();

             if (x<0)
                x = 0;
             else
                x *= PIXEL(img,i,j);

             if (x>255)
                PIXEL(salida,i,j) = 255;
             else if (x<0)
                PIXEL(salida,i,j) = 0;
             else
                PIXEL(salida,i,j) = x;

          } else {
             PIXEL(salida,i,j) = PIXEL(img,i,j);
          }
      }

  return salida;
}

/* ----------------------------------------------------------------------- */

Imagen RuidoMultiplicativoExponencial (Imagen img, float p, float media)
{
  WORD   anchura = ANCHURA(img);
  WORD   altura  = ALTURA(img);
  Imagen salida  = CreaImagen(anchura,altura);
  int    i,j;
  double x;

  randomize();
  SemillaExponencial ( rand(), media );

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++) {

          if (Uniforme()<p) {

             x = PIXEL(img,i,j)*Exponencial();

             if (x>255)
                PIXEL(salida,i,j) = 255;
             else if (x<0)
                PIXEL(salida,i,j) = 0;
             else
                PIXEL(salida,i,j) = x;

          } else {
             PIXEL(salida,i,j) = PIXEL(img,i,j);
          }
      }

  return salida;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */


