/* ----------------------------------------------------------------------- */
/*    Fernando Berzal Galiano                ETS Ingeniera Informtica    */
/* ----------------------------------------------------------------------- */
/*                       Procesamiento de imgenes                         */
/* ----------------------------------------------------------------------- */

#include "IMG_TYPE.H"

#include <string.h>
#include <stdlib.h>
#include <math.h>

/* ----------------------------------------------------------------------- */

static float cnorm (Complex c)   /* |c|^2 */
{
  return c.real*c.real + c.imag*c.imag;
}


/* --- */

static float cmod (Complex c)
{
  float norm = c.real*c.real + c.imag*c.imag;

  if (norm>0)
     return sqrt(norm);
  else
     return 0;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*  ComplexImg                                                             */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

void NormalizarComplexImg (ComplexImg x)
{
  int       i,j;
  float     xmax = 0.0;
  Complex **d    = x->datos;

  for (i=0; i<(x->anchura); i++)
      for (j=0; j<(x->altura); j++) {
          if (cnorm(d[i][j]) > xmax)
	     xmax = cnorm(d[i][j]);
      }

  if (xmax>0)
     xmax = sqrt(xmax);

  for (i=0; i<(x->anchura); i++)
      for (j=0; j<(x->altura); j++) {
	     d[i][j].real /= xmax;
	     d[i][j].imag /= xmax;
      }
}

/* ----------------------------------------------------------------------- */

ComplexImg CreaComplexImg (Imagen img)
{
  int        i,j;
  ComplexImg x = (InfoComplexImg *) malloc (sizeof(InfoComplexImg));

  x->altura  = ALTURA(img);
  x->anchura = ANCHURA(img);
  x->datos   = (Complex **) malloc (ANCHURA(img) * sizeof(Complex *) );

  for (i=0; i<ANCHURA(img); i++)
      x->datos[i] = (Complex *) malloc (sizeof(Complex)*ALTURA(img));

  for (i=0; i<ANCHURA(img); i++)
      for (j=0; j<ALTURA(img); j++) {
	  x->datos[i][j].real = PIXEL(img,i,j);
	  x->datos[i][j].imag = 0.0;
      }

  return x;
}

/* ----------------------------------------------------------------------- */

ComplexImg CreaComplex (int dimX, int dimY)
{
  int        i,j;
  ComplexImg x = (InfoComplexImg *) malloc (sizeof(InfoComplexImg));

  x->altura  = dimY;
  x->anchura = dimX;
  x->datos   = (Complex **) malloc (dimX * sizeof(Complex *) );

  for (i=0; i<dimX; i++)
      x->datos[i] = (Complex *) malloc (sizeof(Complex)*dimY);

  for (i=0; i<dimX; i++)
      for (j=0; j<dimY; j++) {
	  x->datos[i][j].real = 0.0;
	  x->datos[i][j].imag = 0.0;
      }

  return x;
}

/* ----------------------------------------------------------------------- */

ComplexImg CopiaComplexImg (ComplexImg y)
{
  int        i,j;
  ComplexImg x = (InfoComplexImg *) malloc(sizeof(InfoComplexImg));

  x->anchura = y->anchura;
  x->altura  = y->altura;
  x->datos   = (Complex **) malloc ((y->anchura) * sizeof(Complex *) );

  for (i=0; i<(y->anchura); i++)
      x->datos[i] = (Complex *) malloc(sizeof(Complex)*(y->altura));

  for (i=0; i<(y->anchura); i++)
      for (j=0; j<(y->altura); j++) {
	  x->datos[i][j].real = y->datos[i][j].real;
	  x->datos[i][j].imag = y->datos[i][j].imag;
      }

  return x;
}

/* ----------------------------------------------------------------------- */

void DestruyeComplexImg (ComplexImg *x)
{
  int i;

  if (*x) {

     for (i=0; i<((*x)->anchura); i++)
	 free((*x)->datos[i]);

     free ((*x)->datos);
     free (*x);
     *x = NULL;
  }
}

/* ----------------------------------------------------------------------- */
/*  Conversin ComplexImg ---> Imagen                                         */
/* ----------------------------------------------------------------------- */

Imagen ImagenComplex ( ComplexImg image, WORD anchura, WORD altura)
{
  float  xmax, xmin, x, xdif;
  int    i,j,imin,jmin;
  Imagen output = CreaImagen(anchura,altura);

  /* IMAGE: complex --> real */

  imin = (image->anchura - anchura)/2;
  jmin = (image->altura  - altura) /2;


  xmax = -1.0e20;
  xmin = -xmax;

  for (i=imin; i<imin+anchura; i++) {
      for (j=jmin; j<jmin+altura; j++) {

	  x = cmod (image->datos[i][j]);

	  if (x > xmax)
	     xmax = x;

	  if (x < xmin)
	     xmin = x;
      }
  }

  xdif = xmax-xmin;

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++) {
	  x = cmod(image->datos[i+imin][j+jmin]);
	  x = (x-xmin)/xdif * 255.0;
	  PIXEL(output,i,j) = (BYTE)x;
      }

  return output;
}

/* --- */

Imagen ImagenLogComplex ( ComplexImg image, WORD anchura, WORD altura)
{
  float  xmax, xmin, x, xdif;
  int    i,j,imin,jmin;
  Imagen output = CreaImagen(anchura,altura);

  /* IMAGE: complex --> real */

  imin = (image->anchura - anchura)/2;
  jmin = (image->altura  - altura) /2;


  xmax = -1.0e20;
  xmin = -xmax;

  for (i=imin; i<imin+anchura; i++) {
      for (j=jmin; j<jmin+altura; j++) {

	  x = cmod (image->datos[i][j]);
	   
	  if (x > xmax)
	     xmax = x;

	  if (x < xmin)
	     xmin = x;
      }
  }

  xmin = log(xmin+1);
  xmax = log(xmax+1);
  xdif = xmax-xmin;

  for (i=0; i<anchura; i++)
      for (j=0; j<altura; j++) {
	  x = cmod(image->datos[i+imin][j+jmin]);
          x = log(x+1);
	  x = (x-xmin)/xdif * 255.0;
	  PIXEL(output,i,j) = (BYTE)x;
      }

  return output;
}


/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*  FloatImg                                                               */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

FloatImg CreaFloatImg (WORD anchura, WORD altura)
{
  int      i,j;
  FloatImg x = (InfoFloatImg *) malloc (sizeof(InfoFloatImg));

  if (x) {
     x->altura  = altura;
     x->anchura = anchura;
     x->datos   = (float **) malloc (anchura * sizeof(float *) );

     if (x->datos) {

	for (i=0; i<anchura; i++) {
	    x->datos[i] = (float *) malloc (sizeof(float)*altura);

            if (x->datos[i]) {

               for (j=0; j<altura; j++)
                   x->datos[i][j] = 0;

            } else {

               for (j=0; j<i; j++)
                   free (x->datos[j]);

               free (x);
               x = NULL;
            }
	}

     } else {
        free (x);
        x = NULL;
     }
  }

  return x;
}

/* ----------------------------------------------------------------------- */

void DestruyeFloatImg (FloatImg *x)
{
  int i;

  if (*x) {

     for (i=0; i<((*x)->anchura); i++)
	 free((*x)->datos[i]);

     free ((*x)->datos);
     free (*x);
     *x = NULL;
  }
}

/* ----------------------------------------------------------------------- */
/*  Conversin FloatImg ---> Imagen                                        */
/* ----------------------------------------------------------------------- */

Imagen ImagenFloat ( FloatImg imagen)
{
  int    anchura = imagen->anchura;
  int    altura  = imagen->altura;
  float  xmax, xmin, x, xdif;
  int    i,j;
  Imagen output = CreaImagen(anchura,altura);

  xmax = -1.0e20;
  xmin = -xmax;

  for (i=0; i<anchura; i++) {
      for (j=0; j<altura; j++) {
          x = imagen->datos[i][j];
	  if (x > xmax)
	     xmax = x;
	  if (x < xmin)
	     xmin = x;
      }
  }

  xdif = xmax-xmin;

  if (xdif)
     for (i=0; i<anchura; i++)
         for (j=0; j<altura; j++) {
	     x = imagen->datos[i][j];
	     x = (x-xmin)/xdif * 255.0;
	     PIXEL(output,i,j) = (BYTE)x;
         }

  return output;
}

/* ----------------------------------------------------------------------- */
